/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#include "watchdog.h"
#include <boost/thread/condition_variable.hpp>
#include <boost/thread/future.hpp>
#include <thread>
#include "util.h"

namespace ds {
namespace asio {

struct Watchdog::Impl {
    boost::asio::io_service &m_ioService;
    boost::chrono::microseconds m_failureLatency;
    volatile bool m_shutdown;
    boost::mutex m_mutex;
    boost::condition_variable m_conditionVariable;
    std::thread m_thread;

    Impl(boost::asio::io_service &ioService, boost::chrono::microseconds failureLatency)
            : m_ioService(ioService)
            , m_failureLatency(failureLatency)
            , m_shutdown(false)
            , m_thread([this](){ run(); }) {
        auto failureLatencyMs = boost::chrono::duration_cast<boost::chrono::milliseconds>(m_failureLatency).count();
        VDSM_IENTER("failureLatencyMs:%lld", failureLatencyMs);
    }
    ~Impl() {
        VDSM_IENTER("");
        m_shutdown = true;
        m_conditionVariable.notify_one();
        m_thread.join();
    }

    void run() {
        VDSM_ENTER("");
        boost::unique_lock<boost::mutex> lock(m_mutex);
        boost::chrono::steady_clock clock;
        while (true) {
            auto timePoint = clock.now();
            bool pong = false;
            VDSM_INFO("Watchdog ping");
            m_ioService.dispatch([&]() {
                VDSM_INFO("Watchdog pong");
                pong = true;
            });
            m_conditionVariable.wait_for(lock, m_failureLatency, [&](){
                auto duration = clock.now() - timePoint;
                auto durationMs = boost::chrono::duration_cast<boost::chrono::milliseconds>(duration).count();
                VDSM_INFO("Watchdog wait for m_shutdown:%d durationMs:%lld", m_shutdown, durationMs);
                return m_shutdown || (duration >= m_failureLatency);
            });
            if (m_shutdown) {
                break;
            }
            if (!pong) {
                VDSM_ERROR("Watchdog failure");
                abort();
            }
        }
    }
};
Watchdog::Watchdog(boost::asio::io_service &ioService, boost::chrono::microseconds failureLatency)
        : m_impl(new Impl(ioService, failureLatency)) {}
Watchdog::~Watchdog() = default;

} // namespace sqlite3
} // namespace ds
