/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#include "sqlite3.h"
#include <boost/lexical_cast.hpp>
#include <sqlite3.h>
#include <errno.h>
#include <string.h>

namespace ds {
namespace sqlite3 {

const char* Database::MEMORY_FILE_NAME = ":memory:";

void Database::Deleter::operator()(::sqlite3* ptr) {
    sqlite3_close_v2(ptr);
}

Database::Database(const char* fileName, Mode mode) {
    int flags = SQLITE_OPEN_NOMUTEX;
    switch (mode) {
        case Mode::ReadWrite:
            flags |= SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE;
            break;
        case Mode::ReadOnly:
            flags |= SQLITE_OPEN_READONLY;
            break;
    }

    ::sqlite3* ptr = NULL;
    int ret = sqlite3_open_v2(fileName, &ptr, flags, NULL);
    m_ptr.reset(ptr); // call sqlite3_close_v2 in every case
    if (ret != SQLITE_OK) {
        throw std::runtime_error(std::string("Could not open database ") + fileName + ": " +
                sqlite3_errmsg(get()));
    }
}

void Database::execRaw(const char* sql) {
    char *errmsg = NULL;
    int ret;

    ret = sqlite3_exec(get(), sql, NULL, NULL, &errmsg);
    if (ret != SQLITE_OK) {
        std::string msg = errmsg;
        sqlite3_free(errmsg);
        throw std::runtime_error(msg);
    }
}

struct BackupDeleter {
    void operator()(::sqlite3_backup *x) { sqlite3_backup_finish(x); }
};
typedef std::unique_ptr< ::sqlite3_backup, BackupDeleter> BackupPtr;

void Database::backupTo(::sqlite3 *destination) {
    BackupPtr backup(sqlite3_backup_init(destination, "main", get(), "main"));
    sqlite3_backup_step(backup.get(), -1);
}

void Statement::Deleter::operator()(::sqlite3_stmt* ptr) {
    sqlite3_finalize(ptr);
}

Statement::Statement(Database& db, const char *sql) {
    sqlite3_stmt* ptr = NULL;
    const char *rem = NULL;
    int ret = sqlite3_prepare_v2(db.get(), sql, -1, &ptr, &rem);
    m_ptr.reset(ptr); // call sqlite3_finalize in every case
    if (ret != SQLITE_OK) {
        throw std::runtime_error(std::string(__func__) + ": " + sqlite3_errstr(ret));
    }

    if (rem && *rem != '\0') {
        throw std::runtime_error(std::string(__func__) + ": " + sqlite3_errstr(ret));
    }
}

bool Statement::getBusy() const {
    return sqlite3_stmt_busy(m_ptr.get());
}

void Statement::assumeBusy(bool expectedBusy) const {
    auto realBusy = getBusy();
    if (realBusy != expectedBusy) {
        throw std::runtime_error(std::string(__func__)
                + " realBusy:" + boost::lexical_cast<std::string>(realBusy)
                + " != expectedBusy:" + boost::lexical_cast<std::string>(expectedBusy));
    }
}

void Statement::reset() {
    int ret = sqlite3_reset(get());
    if (ret != SQLITE_OK) {
        throw std::runtime_error(std::string(__func__) + ": " + sqlite3_errstr(ret));
    }
}

void Statement::BindDeleter::operator()(::sqlite3_stmt* ptr) {
    sqlite3_clear_bindings(ptr);
}

Statement::StepResult Statement::step() {
    int ret = sqlite3_step(get());
    if (ret == SQLITE_DONE) {
        return StepResult::Done;
    }
    if (ret == SQLITE_ROW) {
        return StepResult::Row;
    }
    reset(); // make sure to reset busy state.
    throw std::runtime_error(std::string(__func__) + ": " + sqlite3_errstr(ret));
}

template <typename F>
void Statement::bindAtImpl(const F &f) {
    auto ret = f();
    if (ret != SQLITE_OK) {
        throw std::runtime_error(std::string(__func__) + ": " + sqlite3_errstr(ret));
    }
}

void Statement::bindAt(int index, Nullptr) {
    bindAtImpl([&](){ return sqlite3_bind_null(get(), index); });
}

void Statement::bindAt(int index, int value) {
    bindAtImpl([&](){ return sqlite3_bind_int(get(), index, value); });
}

void Statement::bindAt(int index, const char *value) {
    bindAtImpl([&](){ return sqlite3_bind_text(get(), index, value, -1, SQLITE_STATIC); });
}

void Statement::bindAt(int index, const std::string &value) {
    bindAtImpl([&](){ return sqlite3_bind_text(get(), index, value.c_str(), value.size(), SQLITE_STATIC); });
}

void Statement::bindBlobAt(int index, const uint8_t *value, int size) {
    bindAtImpl([&](){ return sqlite3_bind_blob(get(), index, value, size, SQLITE_STATIC); });
}

template <>
const char* Statement::getColumn<const char*>(int i) const {
    return reinterpret_cast<const char *>(sqlite3_column_text(m_ptr.get(), i));
}

template <>
std::string Statement::getColumn<std::string>(int i) const {
    return std::string(getColumn<const char*>(i));
}

template <>
int Statement::getColumn<int>(int i) const {
    return sqlite3_column_int(m_ptr.get(), i);
}

Transaction::Transaction(Database& database) : m_database(database), m_state(State::Active) {
    database.execRaw("BEGIN");
}

Transaction::~Transaction() {
    switch (m_state) {
        case State::Active:
            try {
                m_database.execRaw("ROLLBACK");
            } catch (const std::exception&) {
                // Do not throw an exception in the destructor
            }
            break;
        case State::Committed:
            break;
    }
}

void Transaction::commit() {
    switch (m_state) {
        case State::Active:
            m_database.execRaw("COMMIT");
            m_state = State::Committed;
            break;
        case State::Committed:
            throw std::runtime_error("Transaction already committed");
    }
}

} // namespace sqlite3
} // namespace ds
