/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.


    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"

#include "datamodel.h"

#include <utlist.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sqlite3.h>

#include "registers.h"
#include "dsm.h"
#include "bsp.h"
#include "heating.h"
#include "selectA.h"
#include "selectB.h"
#include "mmi.h"
#include "bitset8.h"
#include "button.h"
#include "version.h"
#include "crc16_ccitt.h"
#include "hash.h"
#include "util.h"
#include "device-api-proto.h"
#include "vdc/vdc-api-proto.h"
#include "datamodel.sql.h"

Q_DEFINE_THIS_MODULE(datamodel)

#define SQL_ASSERT(test_) \
    do { \
        if (test_) { \
        } \
        else { \
            printf("SQL Assertion failed in datamodel at line %d:\n" \
                   "error code %d, error message %s\n", \
                   __LINE__, sqlite3_errcode(me->pDb), sqlite3_errmsg(me->pDb)); \
            abort(); \
        } \
    } while (FALSE)

#define LOCK_ACCESS   ;
#define UNLOCK_ACCESS ;

enum DatamodelConstants {
    DEFAULT_HARDWARE_REVISION       = 0,
    DEFAULT_APARTMENT_ID            = 0,
    DEFAULT_WARN_TEMPERATURE        = 90,
    DEFAULT_ERROR_TEMPERATURE       = 150,
    DEFAULT_AUTO_PING_INTERVAL      = 20000,                /* 20 seconds */
    DEFAULT_ONE_TIME_PING_INTERVAL  = 60000,                /* 60 seconds */
    SAVE_DEVICE_STATE_TIMEOUT       = BSP_TICKS(3600000),   /* 60 minutes */
    DEFAULT_DSM_INFO_FLAGS          = (MMI_SHOW_COLOR_IN_READY_STATE | ONE_TIME_OFF_24H_PING_MODE),
    DEFAULT_UPSTREAM_FREQUNECY      = UPSTREAM_FREQUENCY_15_20,
};

typedef enum {
    ONLY_ACTIVE = 1,
    ONLY_INACTIVE = 2,
    ALL = 3
} DatamodelDeviceModifier;

#define Datamodel_Changed()                         do { LOCK_ACCESS; ++me->changeCount; UNLOCK_ACCESS; } while (0)

#define PostFIFO_ErrorOK(activeObject_)             QActive_postFIFO(activeObject_, (QEvent *)&l_errorOkEvt)



typedef struct {
    uint8_t stopTimeouts[MAX_ZONES][LAST_STANDARD_GROUP];                       /**< cached stop timeout for each zone/standard group */
    uint8_t activeGroupStopTimeout[MAX_ZONES][(LAST_STANDARD_GROUP + 7) / 8];   /**< group needs a stop timeout bit-field */
    uint8_t userGroupStopTimeouts[NUM_USER_GROUP_WITH_HSM];                     /**< cached stop timeout for user groups with HSM */
    uint8_t activeUserGroupStopTimeout[(NUM_USER_GROUP_WITH_HSM + 7) / 8];      /**< user group needs a stop timeout bit-field */
} DatamodelStopTimeouts;

/* Active object class -----------------------------------------------------*/
typedef struct DatamodelTag {
    QActive super;
    QTimeEvt opTimeEvt;                         /**< flash operation timer */
    QEQueue deferredQueue;                      /**< deferred queue */
    QActive * requester;                        /**< requester active object */
    uint32_t hash;                              /**< configuration hash */
    uint32_t changeCount;                       /**< configuration change count */
    uint32_t hwRevision;                        /**< cached hardware revision */
    uint16_t lastValidDeviceId;                 /**< last valid device ID */
    uint8_t validDevices[(MAX_DEVICE+7)/8];     /**< valid devices bit-field */
    uint8_t activeDevices[(MAX_DEVICE+7)/8];    /**< active devices bit-field */
    uint16_t deviceStateChangeCount;            /**< device state change count */
    uint16_t deviceStatePage;                   /**< device state page */
    uint8_t localZoneId;                        /**< local zone id for initialization of zone info */
    uint8_t groupId;                            /**< group id for initialization of zone info */
    uint8_t buttonId;                           /**< button id for initialization of stop timers */
    bool dsmConfigValid;                        /**< dSM configuration block valid */
    dsuid_t dsmid;                              /**< current dSUID */
    DatamodelResDsmProperties dsmProperties;    /**< cached dSM properties (response event) */
    uint8_t numberOfZones;                      /**< number of zones */
    DatamodelZoneInfo zoneInfos[MAX_ZONE];      /**< cached zone informations */

    uint8_t validZoneInfos[(MAX_ZONE+7)/8];     /**< valid zone information bit-field (bits MIN_ZONE to MAX_ZONE) */
    DatamodelStopTimeouts stopTimeouts;         /**< cached stop timeouts for standard and user groups with HSM */

    /* sqlite backend */
    sqlite3 *pDb;
    bool databaseOpen;                          /**< flag for database status (open or closed) */
} Datamodel;

/* Local objects -----------------------------------------------------------*/
static Datamodel l_datamodel;           /**< the single instance of the data model active object */
static QEvent const * l_deferredQueueSto[128];
static const DatamodelDeviceModifier l_dsmApiZoneDevCountModifiers[] = { ALL, ONLY_ACTIVE, ONLY_INACTIVE, ALL };

static const DatamodelErrorEvt l_errorOkEvt = { { DATAMODEL_RES_ERROR_SIG, 0 }, ERROR_OK };
static const QEvent l_flash_dma_done = { FLASH_DMA_DONE_SIG, 0 };

/* states */
static QState Datamodel_initial(Datamodel *me, QEvent const *e);
static QState Datamodel_read_dsm_configuration(Datamodel *me, QEvent const *e);
static QState Datamodel_read_dsm_info(Datamodel *me, QEvent const *e);
static QState Datamodel_parse_zone(Datamodel *me, QEvent const *e);
static QState Datamodel_parse_device(Datamodel *me, QEvent const *e);
static void Datamodel_fill_device_from_SQL(DatamodelDevice *device, sqlite3_stmt *statement);
static uint64_t groups_array_to_int(uint8_t *groups);
static void groups_int_to_array(uint64_t groups, uint8_t *array);
static QState Datamodel_parse_device_state(Datamodel *me, QEvent const *e);
static uint16_t Datamodel_parse_groups_next_batch(Datamodel *me);
static QState Datamodel_parse_groups(Datamodel *me, QEvent const *e);
static QState Datamodel_parse_clusters(Datamodel *me, QEvent const *e);

static QState Datamodel_idle(Datamodel *me, QEvent const *e);

static QState Datamodel_busy(Datamodel *me, QEvent const *e);
static void Datamodel_get_device_dsuid_entry(Datamodel *me, const char* dsuid, uint8_t privateId);
static bool Datamodel_get_device_by_index_from_vdc(Datamodel *me, const dsuid_t  *vdcdsuid,  uint8_t modifier, uint16_t deviceIndex, uint16_t *deviceId);
static void Datamodel_get_device_id_entry(Datamodel *me, uint16_t deviceId, uint8_t privateId);
static bool Datamodel_get_device_entry_by_index(Datamodel *me, uint8_t localZoneId, uint8_t deviceIndex, DatamodelDeviceModifier modifier, const dsuid_t  *vdcdsuid, DatamodelDevice *output_device);
static uint16_t Datamodel_get_zone_usage(Datamodel *me, uint8_t localZoneId);
static void Datamodel_add_zone_to_db(Datamodel *me, uint8_t localZoneId, uint16_t zoneId);
static void Datamodel_publish_zone_added(uint8_t localZoneId, uint16_t zoneId);
static void Datamodel_remove_zone_from_db(Datamodel *me, uint8_t localZoneId);
static void Datamodel_group_membership_modify(Datamodel *me, uint16_t deviceId, uint8_t groupId, uint8_t modifier, bool sendDownstream);
static uint16_t Datamodel_get_zone_device_count(Datamodel *me, uint8_t localZoneId, DatamodelDeviceModifier deviceModifier, const dsuid_t *dsuid);
static bool Datamodel_dsm_entry_exists(Datamodel *me, const char *dsmId);
static void Datamodel_get_dsm_info_name(Datamodel *me, const dsuid_t *dsmid);
static void Datamodel_get_group_info(Datamodel *me, uint8_t localZoneId, uint8_t groupId, const dsuid_t *vdc_dsuid);
static void Datamodel_get_scene_info(Datamodel *me, uint8_t localZoneId, uint8_t groupId, uint8_t sceneId);
static void Datamodel_set_dsm_apartment_id(Datamodel *me, uint8_t apartmentId);
static void Datamodel_set_dsm_temperature_limits(Datamodel *me, uint8_t warnTemperature, uint8_t errorTemperature);
static void Datamodel_set_dsm_ping_intervals(Datamodel *me, uint16_t autoPingInterval, uint16_t oneTimePingInterval);
static void Datamodel_set_dsm_flags(Datamodel *me, uint8_t flags);
static void Datamodel_set_dsm_upstream_frequency(Datamodel *me, uint8_t upstreamFrequency);
static void Datamodel_set_dsm_name(Datamodel *me, const uint8_t *name, const dsuid_t *dsmid);
static void Datamodel_get_dsm_properties(Datamodel *me, const dsuid_t *dsmid);
static void Datamodel_set_device_name(Datamodel *me, uint16_t deviceId, const char *name);
static void Datamodel_set_group_name(Datamodel *me, uint8_t localZoneId, uint8_t groupId, const uint8_t *name);
static void Datamodel_set_group_stop_timeout(Datamodel *me, uint8_t localZoneId, uint8_t groupId, uint8_t stopTimeout);
static void Datamodel_set_group_color(Datamodel *me, uint8_t localZoneId, uint8_t groupId, uint8_t color);
static void Datamodel_add_device(Datamodel *me, DatamodelDevice *device);
static void Datamodel_ping_device_state(Datamodel *me, uint16_t deviceId, uint16_t nextDeviceId, bool active, bool oneTimePingDone);
static void Datamodel_set_device_active(Datamodel *me, uint16_t deviceId, bool active);
static void Datamodel_zone_modify_zoneId(Datamodel *me, uint8_t localZoneId, uint16_t zoneId);
static void Datamodel_zone_modify_name(Datamodel *me, uint8_t localZoneId, const uint8_t *name, dsuid_t const *vdcdsuid);
static void Datamodel_zone_group_modify_remove(Datamodel *me, uint8_t localZoneId, uint8_t groupId);
static void Datamodel_device_set_zone(Datamodel *me, uint16_t deviceId, uint8_t localZoneId);
static void Datamodel_device_set_button_group(Datamodel *me, uint16_t deviceId, uint8_t groupId);
static void Datamodel_device_set_button_set_local_priority(Datamodel *me, uint16_t deviceId, bool newSetLocalPrioFlag);
static void Datamodel_device_set_button_set_no_coming_home(Datamodel *me, uint16_t deviceId, bool noComingHome);
static void Datamodel_device_set_button_set_apartment(Datamodel *me, uint16_t deviceId, bool apartment);
static void Datamodel_device_set_button_set_output_channel(Datamodel *me, uint16_t deviceId, uint8_t channelId, bool sendDownstream);
static void Datamodel_device_set_button_set_global_group(Datamodel *me, uint16_t deviceId, uint8_t group, bool sendDownstream);
static void Datamodel_device_set_ltmode(Datamodel *me, uint16_t deviceId, uint8_t LTMODERegister);
static void Datamodel_device_set_output_load(Datamodel *me, uint16_t deviceId, uint8_t MODERegister);
static void Datamodel_device_set_button_id_and_group(Datamodel *me, uint16_t deviceId, uint8_t buttonIdAndGroupRegister);
static void Datamodel_device_set_locked(Datamodel *me, uint16_t deviceId, bool locked);
static void Datamodel_remove_inactive_device(Datamodel *me);
static void Datamodel_remove_device(Datamodel *me, uint16_t deviceId);
static void Datamodel_factory_reset(Datamodel *me);

static bool Datamodel_GetZoneInfoPerVdcByIndex(Datamodel *me, uint8_t zoneIndex, const dsuid_t *vdc_dsuid, DatamodelZoneInfo **zoneInfo);

static void Datamodel_send_device_accessibility(Datamodel *me, uint16_t deviceId, bool on);

static DatamodelZoneInfo * Datamodel_CreateGetZoneInfoByDsuid(Datamodel *me, const dsuid_t *vdcdsuid);
static void Datamodel_add_default_zone_for_vdc(Datamodel *me, const dsuid_t *dsmid);

static QState Datamodel_stopped(Datamodel *me, QEvent const *e);
static QState Datamodel_stopped_top(Datamodel *me, QEvent const *e);

/* static helper functions */
static void Datamodel_InitZoneInfo(DatamodelZoneInfo * zoneInfo);
static DatamodelZoneInfo * Datamodel_GetZoneInfoByZoneId(uint16_t zoneId);
static uint8_t Datamodel_CreateZone(Datamodel *me, uint16_t zoneId);
static void Datamodel_RemoveZone(Datamodel *me, uint16_t zoneId);
static uint8_t Datamodel_GetGroupIdByGroupIndex(DatamodelZoneInfo const * zoneInfo, uint8_t index);
static uint8_t Datamodel_GetGroupCountOfZone(DatamodelZoneInfo const * zoneInfo);
static void Datamodel_SendClusterInfoResponse(Datamodel *me, DatamodelCluster const * cluster, uint16_t numOfDevices);
static uint8_t Datamodel_ZoneCount(Datamodel *me, const dsuid_t *vdc_dsuid);
static uint8_t Datamodel_GetUnusedLocalZoneId(void);
static void Datamodel_SetLocalZoneIdUsed(uint8_t localId);
static void Datamodel_SetLocalZoneIdUnused(uint8_t localId);
static bool Datamodel_GetDefaultLocalPrioritySetting(uint16_t fid);
static void Datamodel_InitDevice(DatamodelDevice * device,
                                 const char *dsuid,
                                 uint16_t deviceId,
                                 uint16_t vid,
                                 uint16_t fid,
                                 uint16_t ver,
                                 uint8_t locked,
                                 const char *vdcdsuid,
                                 uint16_t zoneId);
static void Datamodel_InitGroup(Datamodel *me, uint8_t localZoneId, uint8_t groupId, uint8_t stateMachineGroupId, const uint8_t * name);
static void Datamodel_InitZone(DatamodelZone * zone, uint8_t localZoneId, uint16_t zoneId, uint8_t zoneTagId);
static bool Datamodel_GetLocalZoneId(uint16_t zoneId, uint8_t *localZoneId);
static void Datamodel_PostError(QActive *activeObj, uint8_t error);
static void Datamodel_PostLongError(QActive *activeObj, uint8_t error, uint8_t privateId);
static void Datamodel_SendZoneInfo(QActive *activeObj, DatamodelZoneInfo *zoneInfo);
static void Datamodel_NotifyChange(void);
static void Datamodel_NotifyUserGroup(uint8_t groupId, uint8_t stateMachineGroupId);

static bool Datamodel_SelectRandomDeviceId(Datamodel *me, uint16_t *deviceId);
static void Datamodel_UpdateCachedStopTimeout(Datamodel *me, uint8_t localZondId, uint8_t groupId, uint8_t stopTimeout);
static void Datamodel_ResetGroupStopTimeoutActive(Datamodel *me);
static void Datamodel_SetGroupActiveStopTimeout(Datamodel *me, DatamodelDevice const *device);
static void Datamodel_ClearGroupActiveStopTimeout(Datamodel *me, uint8_t localZoneId, uint8_t groupId);
static bool Datamodel_IsZoneSensor(Datamodel *me, uint16_t deviceId, uint8_t sensorIndex, dsuid_t *dsuid, uint8_t *type, bool *sensorPrecision);
static bool Datamodel_GetZoneSensorZoneId(Datamodel *me, dsuid_t *sensorDsuid, uint16_t *zoneId, uint8_t *localZoneId);
static void Datamodel_UpdateHash(Datamodel *me);
static void Datamodel_InitCluster(DatamodelCluster *cluster, uint8_t clusterId);
static void Datamodel_convert_group_to_cluster(Datamodel *me);
static uint16_t Datamodel_getClusterDeviceCount(const dsuid_t *vdc_dsuid, uint8_t clusterId);
static uint16_t Datamodel_getGroupDeviceCount(const dsuid_t *vdc_dsuid, uint8_t groupId, uint8_t localZoneId);

static uint8_t Datamodel_setClusterName(uint8_t groupId, const uint8_t* name);
static uint8_t Datamodel_setClusterStopTimeout(uint8_t groupId, uint8_t stopTimeout);
static uint8_t Datamodel_setClusterRemoteStopTimeout(uint8_t groupId, uint8_t remoteStopTimeout, uint8_t color);
static uint8_t Datamodel_setLocationProtection(uint8_t groupId, uint16_t location, uint16_t floor, uint16_t protectionClass);
static uint8_t Datamodel_setConfigurationLock(uint8_t groupId, bool lock);
static uint8_t Datamodel_setSceneLock(uint8_t groupId, const uint8_t *sceneLock, uint8_t sceneLockSize);
static uint8_t Datamodel_get_or_create_localZone_by_zoneId(Datamodel *me, uint16_t zoneId);

/* Global-scope objects ----------------------------------------------------*/
QActive * const AO_Datamodel = (QActive *)&l_datamodel;      /* "opaque" AO pointer */

/*..........................................................................*/
uint8_t Datamodel_DeviceButtonZone(DatamodelDevice const * device)
{
    return device->localZoneId;
}

/*..........................................................................*/
uint8_t Datamodel_DeviceButtonActiveGroup(DatamodelDevice const * device)
{
    /* initialize group ID with value saved in button ID and group configuration */
    uint8_t groupId = Datamodel_DeviceButtonGroup(device);
    /* is user group applicable? */
    uint8_t userGroupId = device->buttonUserGroup;

    if (userGroupId != DATAMODEL_GROUP_ID_NOT_APPLICABLE) {
        groupId = userGroupId;
    }

    if (groupId == 0) {
        return FidGetClass(device->fid);
    }

    return groupId;
}

/*..........................................................................*/
uint8_t Datamodel_DeviceButtonId(DatamodelDevice const * device)
{
    return (device->buttonIdAndGroup & 0x0f);
}

/*..........................................................................*/
uint8_t Datamodel_DeviceButtonGroup(DatamodelDevice const * device)
{
    return (device->buttonIdAndGroup & 0xf0) >> 4;
}

/*..........................................................................*/
uint8_t Datamodel_ButtonId_ButtonGroupId(uint8_t buttonId)
{
    switch (buttonId) {
    case BUTTON_GROUP_1:
    case BUTTON_GROUP_5: {
        /* button group 1 */
        return BUTTON_GROUP_1;
    }
    case BUTTON_GROUP_2:
    case BUTTON_GROUP_6: {
        /* button group 2 */
        return BUTTON_GROUP_2;
    }
    case BUTTON_GROUP_3:
    case BUTTON_GROUP_7: {
        /* button group 3 */
        return BUTTON_GROUP_3;
    }
    case BUTTON_GROUP_4:
    case BUTTON_GROUP_8: {
        /* button group 4 */
        return BUTTON_GROUP_4;
    }
    default: {
        /* all others are main group */
        break;
    }
    }
    return 0;
}

/*..........................................................................*/
void Datamodel_updateDeviceVdc(uint16_t deviceId, char *vdcdsuid, uint16_t fid)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel updateDeviceVdc Database closed!\n");
        UNLOCK_ACCESS
        return;
    }

    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
               "UPDATE devices SET vdcdsuid=?, fid=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 1, vdcdsuid, -1, SQLITE_STATIC) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, fid) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 3, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
}

/*..........................................................................*/
bool Datamodel_DsmPropertiesProgrammingModeDisabled()
{
    Datamodel *me = &l_datamodel;
    return (me->dsmProperties.flags & PROGRAMMING_MODE_MASK) == PROGRAMMING_MODE_DISABLED;
}

/*..........................................................................*/
bool Datamodel_DsmPropertiesRoomDimmingDisabled()
{
    Datamodel *me = &l_datamodel;
    return (me->dsmProperties.flags & ROOM_DIMMING_MASK) == ROOM_DIMMING_DISABLED;
}

/*..........................................................................*/
bool Datamodel_IsDeviceValid(uint16_t deviceId)
{
    Datamodel *me = &l_datamodel;
    if (IsBitSet8Set(me->validDevices, deviceId)) {
        return TRUE;
    } else {
        return FALSE;
    }
}

/*..........................................................................*/
dsuid_t Datamodel_GetDsmid()
{
    return l_datamodel.dsmid;
}

/*..........................................................................*/
bool Datamodel_IsCluster(uint8_t groupId)
{
    if ((groupId >= FIRST_CLUSTER) && (groupId <= LAST_CLUSTER)) {
        return TRUE;
    }
    return FALSE;
}

/**
 * Data model AO constructor.
 */
void Datamodel_ctor(dsuid_t id, const char* databasePath)
{
    /* set initial state */
    Datamodel *me = &l_datamodel;
    QActive_ctor(&me->super, (QStateHandler)&Datamodel_initial);
    /* initialize deferred queue */
    QEQueue_init(&me->deferredQueue, l_deferredQueueSto, Q_DIM(l_deferredQueueSto), "deferred", &me->super);
    /* initialize time events */
    QTimeEvt_ctor(&me->opTimeEvt, FLASH_OP_TIMEOUT_SIG);
    /* initialize extended state variables */
    me->hash = 0;
    me->changeCount = 0;
    me->lastValidDeviceId = MIN_DEVICE-1;
    /* set all devices to active (default) */
    memset(me->activeDevices, 0xFF, sizeof(me->activeDevices));
    me->dsmConfigValid = FALSE;
    /* initialize dSM properties */
    me->dsmProperties.super.sig = DATAMODEL_RES_DSM_PROPERTIES_SIG;
    me->dsmProperties.super.dynamic_ = 0;
    me->dsmProperties.apartmentId = DEFAULT_APARTMENT_ID;
    me->dsmProperties.maxTemperature = 0;
    me->dsmProperties.currentTemperature = 0;
    me->dsmProperties.warnTemperature = DEFAULT_WARN_TEMPERATURE;
    me->dsmProperties.errorTemperature = DEFAULT_ERROR_TEMPERATURE;
    /* initialize default flags */
    me->dsmProperties.flags = DEFAULT_DSM_INFO_FLAGS;
    /* initialize default ping intervals */
    me->dsmProperties.autoPingInterval = DEFAULT_AUTO_PING_INTERVAL;
    me->dsmProperties.oneTimePingInterval = DEFAULT_ONE_TIME_PING_INTERVAL;
    /* initialize upstream frequency */
    me->dsmProperties.upstreamFrequency = DEFAULT_UPSTREAM_FREQUNECY;
    /* initialize stop timeouts to 0 and 120 for group 2 */
    memset(&me->stopTimeouts, 0, sizeof(me->stopTimeouts));
    uint8_t i;
    for (i = 0; i < MAX_ZONES; ++i) {
        me->stopTimeouts.stopTimeouts[i][SHADE_GROUP] = 120;
    }
    /* check last signal in range */
    Q_ASSERT(LAST_SIG <= (sizeof(QSignal) == 2 ? 0xffff : 0xff));

    memcpy(me->dsmid.id, id.id, sizeof(dsuid_t));

    if (access(databasePath, R_OK | W_OK) != 0 && errno != ENOENT)
    {
        vdsm_report(LOG_ERR, "Error accessing database file %s: %s\n", databasePath,
                strerror(errno));
        exit(-1);
    }

    SQL_ASSERT(sqlite3_open_v2(databasePath, &me->pDb,
            SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE | SQLITE_OPEN_FULLMUTEX,
            NULL) == SQLITE_OK);
    {
        me->databaseOpen = TRUE;
        sqlite3_stmt *statement;
        int rval;
        uint8_t const min_dbversion = 14; // oldest DB version supported
        uint8_t const dbversion = 19;
        uint8_t current_dbversion = 0; // none

        // Locking not needed. Access before startup
        rval = sqlite3_prepare_v2(me->pDb, "SELECT key, value FROM db_config WHERE key=?;", -1, &statement, 0);
        if (rval == SQLITE_OK) {
            /* database already exists */
            SQL_ASSERT(sqlite3_bind_text(statement, 1, "db_version", 10, SQLITE_STATIC) == SQLITE_OK);
            if (sqlite3_step(statement) == SQLITE_ROW) {
                current_dbversion = sqlite3_column_int(statement, 1);
            } else {
                vdsm_report(LOG_ERR, "Error db version not found %s\n", databasePath);
                sqlite3_finalize(statement);
                exit(-1);
            }
            /* no support for DB versions below min_dbversion (incompatible alpha-phase legacy) */
            if (current_dbversion < min_dbversion) {
                vdsm_report(LOG_INFO, "Found too old database version %d -> deleting and starting fresh at version %d\n", current_dbversion, min_dbversion);
                current_dbversion = 0;
            }
            sqlite3_finalize(statement);
        }
        if (current_dbversion == 0) {
            /* no DB or DB version we can't migrate - delete and re-create minimal version we support */
            sqlite3_close(me->pDb);
            /* delete */
            unlink(databasePath);
            /* re-create */
            SQL_ASSERT(sqlite3_open_v2(databasePath, &me->pDb,
                    SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE | SQLITE_OPEN_FULLMUTEX,
                    NULL) == SQLITE_OK);
            /* create initial schema, starting point is min_dbversion */
            char *errmesg;

            vdsm_report(LOG_INFO, "Creating database version %d, db file %s\n", min_dbversion, databasePath);
            rval = sqlite3_exec(me->pDb, CREATE_DB_V14, NULL, NULL, &errmesg);
            if (rval != SQLITE_OK) {
                vdsm_report(LOG_ERR, "Error creating database version %d, db file %s: %s\n", min_dbversion, databasePath, errmesg);
                sqlite3_free(errmesg);
                sqlite3_finalize(statement);
                exit(-1);
            }
            current_dbversion = min_dbversion;
        }
        /* enable foreign_keys feature */
        vdsm_report(LOG_DEBUG, "Enable foreign keys feature\n");
        SQL_ASSERT(sqlite3_exec(me->pDb, "PRAGMA foreign_keys = 1;", NULL, NULL, NULL) == SQLITE_OK);
        /* update database if neccessary */
        for (; current_dbversion < dbversion; ++current_dbversion) {
            vdsm_report(LOG_INFO, "Found database version %d, db file %s\n", current_dbversion, databasePath);
            switch (current_dbversion) {
            case 14:
                /* convert db to version 15 */
                SQL_ASSERT(sqlite3_exec(me->pDb, CONVERT_TO_DB_V15, NULL, NULL, NULL) == SQLITE_OK);
                break;
            case 15:
                /* convert db to version 16 */
                SQL_ASSERT(sqlite3_exec(me->pDb, CONVERT_TO_DB_V16, NULL, NULL, NULL) == SQLITE_OK);
                break;
            case 16:
                /* convert db to version 17 */
                SQL_ASSERT(sqlite3_exec(me->pDb, CONVERT_TO_DB_V17, NULL, NULL, NULL) == SQLITE_OK);
                break;
            case 17:
                /* convert db to version 18 */
                SQL_ASSERT(sqlite3_exec(me->pDb, CONVERT_TO_DB_V18, NULL, NULL, NULL) == SQLITE_OK);
                break;
            case 18:
                /* convert db to version 19 */
                SQL_ASSERT(sqlite3_exec(me->pDb, CONVERT_TO_DB_V19, NULL, NULL, NULL) == SQLITE_OK);
                break;
            default:
                vdsm_report(LOG_ERR, "Database version %d unknown\n", dbversion);
            }
        }

        if (dbversion == current_dbversion) {
            vdsm_report(LOG_INFO, "Using database version %d, db file %s\n", dbversion, databasePath);
        } else {
            vdsm_report(LOG_ERR, "Wrong database version %d, db file %s\n", dbversion, databasePath);
            exit(-1);
        }
    }
}

/*..........................................................................*/
static QState Datamodel_initial(Datamodel *me, QEvent const *e)
{
    (void)e;        /* suppress the compiler warning about unused parameter */
    /* QS dictionary entries */
    QS_OBJ_DICTIONARY(&l_datamodel);

    QS_FUN_DICTIONARY(&Datamodel_initial);
    QS_FUN_DICTIONARY(&Datamodel_read_dsm_configuration);
    QS_FUN_DICTIONARY(&Datamodel_read_dsm_info);
    QS_FUN_DICTIONARY(&Datamodel_parse_zone);
    QS_FUN_DICTIONARY(&Datamodel_parse_device);
    QS_FUN_DICTIONARY(&Datamodel_parse_user_groups);

    QS_FUN_DICTIONARY(&Datamodel_idle);

    QS_FUN_DICTIONARY(&Datamodel_busy);
    QS_FUN_DICTIONARY(&Datamodel_device_set_my_click);
    QS_FUN_DICTIONARY(&Datamodel_device_set_my_action_group);
    QS_FUN_DICTIONARY(&Datamodel_device_set_button_zone);

    QS_FUN_DICTIONARY(&Datamodel_hardware_failure);
    QS_FUN_DICTIONARY(&Datamodel_stopped);

    QS_SIG_DICTIONARY(FLASH_DMA_DONE_SIG, 0);
    QS_SIG_DICTIONARY(FLASH_OP_TIMEOUT_SIG, 0);

    /* subscribe events */
    QActive_subscribe((QActive *)me, START_SIG);
    return Q_TRAN(&Datamodel_read_dsm_configuration);
}

/*..........................................................................*/
static QState Datamodel_read_dsm_configuration(Datamodel *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* update hash */
//      me->hash = Hash_Create(me->pageData, sizeof(DatamodelDsmInfo), me->hash);
        /* set default dSM ID */
//      me->dsmid = l_defaultDsmid;
        /* set default hardware revision */
        me->hwRevision = DEFAULT_HARDWARE_REVISION;
//        QActive_postFIFO((QActive *)me, &l_flash_dma_done);
        QTimeEvt_postIn(&me->opTimeEvt, (QActive *)me, 50 / BSP_MS_PER_TICKS);
        return Q_HANDLED();
    }
    case FLASH_OP_TIMEOUT_SIG: {
        uint32_t seed;
        MMIConfigStateEvt * configStateEvt;
        /* initialize random number generator with seed from dSM ID */
        seed = me->dsmid.id[12];
        seed <<= 8;
        seed |= me->dsmid.id[13];
        seed <<= 8;
        seed |= me->dsmid.id[14];
        seed <<= 8;
        seed |= me->dsmid.id[15];
        srand(seed);
        /* send configuration event to MMI */
        configStateEvt = Q_NEW(MMIConfigStateEvt, MMI_CONFIG_STATE_SIG);
        configStateEvt->config = MMI_CONFIG_OK;
        QActive_postFIFO(AO_MMI, (QEvent *)configStateEvt);
        return Q_TRAN(&Datamodel_read_dsm_info);
    }
    }
    return Q_SUPER(&Datamodel_busy);
}

/*..........................................................................*/
static QState Datamodel_read_dsm_info(Datamodel *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        char destAddr[DSUID_STR_LEN];
        dsuid_to_string_upper(&me->dsmid, destAddr);
        Datamodel_add_vdc(destAddr); // add vdsm to vdc list
        QActive_postFIFO((QActive *)me, &l_flash_dma_done);
        return Q_HANDLED();
    }
    case FLASH_DMA_DONE_SIG: {
        return Q_TRAN(&Datamodel_parse_zone);
    }
    }
    return Q_SUPER(&Datamodel_busy);
}

/*..........................................................................*/
static QState Datamodel_parse_zone(Datamodel *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        sqlite3_stmt *statement;
        int ret;

        LOCK_ACCESS
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM zones;", -1, &statement, 0) == SQLITE_OK);
        do
        {
            ret = sqlite3_step(statement);
            if (ret == SQLITE_ROW)
            {
                /* zone is valid */
                int localZoneId = sqlite3_column_int(statement, 0);
                DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByLocalId(localZoneId);
                if (zoneInfo) {
                    /* local zone id is valid */
                    if (zoneInfo->super.magic != DATAMODEL_MAGIC) {
                        const char* name;
                        /* zone info is not yet initialized */
                        zoneInfo->super.zoneId = sqlite3_column_int(statement, 1);
                        zoneInfo->super.localZoneId = localZoneId;
                        zoneInfo->super.zoneTagId = sqlite3_column_int(statement, 2);
                        name = sqlite3_column_text(statement, 3);
                        if (name != NULL) {
                            memcpy(&zoneInfo->super.name, name, MAX_NAME_LEN);
                        }
                        zoneInfo->super.magic = DATAMODEL_MAGIC;
                        Datamodel_InitZoneInfo(zoneInfo);
                        ++me->numberOfZones;
                        Datamodel_SetLocalZoneIdUsed(localZoneId);
                        /* publish new zone */
                        Datamodel_publish_zone_added(zoneInfo->super.localZoneId, zoneInfo->super.zoneId);
                    } else {
                        Q_ERROR();
                    }
                } else {
                    Q_ERROR();
                }
            }
        } while (ret == SQLITE_ROW);
        sqlite3_finalize(statement);
        UNLOCK_ACCESS

        QActive_postFIFO((QActive *)me, &l_flash_dma_done);
        return Q_HANDLED();
    }
    case FLASH_DMA_DONE_SIG: {
        Datamodel_UpdateHash(me);

        /* find first zone for parsing groups */
        me->groupId = 1;
        for (me->localZoneId = MIN_ZONE; me->localZoneId <= MAX_ZONE; ++me->localZoneId) {
            DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByLocalId(me->localZoneId);
            Q_ASSERT(zoneInfo != 0);
            if (zoneInfo->super.magic == DATAMODEL_MAGIC) {
                break;
            }
        }
        return Q_TRAN(&Datamodel_parse_groups);
    }
    }
    return Q_SUPER(&Datamodel_busy);
}

/*..........................................................................*/
static void Datamodel_UpdateHash(Datamodel *me)
{
    uint8_t localZoneId;
    for (localZoneId = MIN_ZONE; localZoneId <= MAX_ZONE; ++localZoneId) {
        DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByLocalId(localZoneId);
        if (zoneInfo->super.magic == DATAMODEL_MAGIC) {
            /* update hash */
            me->hash = Hash_Create((const uint8_t *)&(zoneInfo->super),
                                   sizeof(zoneInfo->super),
                                   me->hash);
        }
    }
    vdsm_report(LOG_DEBUG, "Updated Hash to: %ul\n", me->hash);
}

/*..........................................................................*/
uint16_t Datamodel_create_zoneId_from_dsuid(dsuid_t const *dsuid)
{
    uint16_t zoneId = (uint16_t)(dsuid->id[14]) << 8 | (uint16_t)(dsuid->id[15]);
    return zoneId;
}

/*..........................................................................*/
static void Datamodel_fill_device_from_SQL(DatamodelDevice *device, sqlite3_stmt *statement)
{
    const char* text;
    device->deviceId = sqlite3_column_int(statement, 0);
    text = sqlite3_column_text(statement, 1);
    if (text != NULL) {
        memcpy(device->dsuid, text, DSUID_STR_LEN);
    }
    device->vid = sqlite3_column_int(statement, 2);
    device->fid = sqlite3_column_int(statement, 3);
    device->ver = sqlite3_column_int(statement, 4);
    device->locked = sqlite3_column_int(statement, 5);
    device->localZoneId = sqlite3_column_int(statement, 6);
    groups_int_to_array(sqlite3_column_int64(statement, 7), device->groupMembership);
    device->ltmode = sqlite3_column_int(statement, 8);
    device->mode = sqlite3_column_int(statement, 9);
    text = sqlite3_column_text(statement, 10);
    if (text != NULL) {
        memcpy(device->name, text, MAX_NAME_LEN);
    }
    device->active = sqlite3_column_int(statement, 11);
    /* fill device button info */
    device->buttonIdAndGroup = sqlite3_column_int(statement, 14);
    device->buttonUserGroup = sqlite3_column_int(statement, 15);
    device->flags = 0;
    device->buttonSetLocalPriority = sqlite3_column_int(statement, 16);
    device->buttonNoComingHomeCall = sqlite3_column_int(statement, 17);
    device->buttonApartment = sqlite3_column_int(statement, 18);
    device->buttonOutputChannel = sqlite3_column_int(statement, 19);
    device->buttonGlobalGroup = sqlite3_column_int(statement, 20);
}

/*..........................................................................*/
static uint64_t groups_array_to_int(uint8_t *groups)
{
    int i;
    uint64_t result = 0;
    for (i = 8; i > 0; i--) {
        result <<= 8;
        result |= groups[i - 1];
    }
    return result;
}

uint64_t Datamodel_groups_array_to_int(uint8_t *groups)
{
    return groups_array_to_int(groups);
}

/*..........................................................................*/
static void groups_int_to_array(uint64_t groups, uint8_t *array)
{
    int i;
    for (i = 0; i < 8; i++) {
        array[i] = groups & 0xFF;
        groups >>= 8;
    }
}

/*..........................................................................*/
static void Datamodel_convert_group_to_cluster(Datamodel *me)
{
    LOCK_ACCESS
    sqlite3_stmt *statementQuery;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "REPLACE INTO cluster \
               (clusterId, stateMachineGroupId, stopTimeout, remoteStopTimeout, name) \
               SELECT g_id, stateMachineGroupId, stopTimeout, remoteStopTimeout, name FROM groups WHERE g_id >= ?;",
               -1, &statementQuery, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statementQuery, 1, USER_GROUP) == SQLITE_OK);
    sqlite3_step(statementQuery);
    sqlite3_finalize(statementQuery);

    sqlite3_stmt *statementRemove;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "DELETE FROM groups WHERE g_id >= ?;",
               -1, &statementRemove, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statementRemove, 1, USER_GROUP) == SQLITE_OK);
    sqlite3_step(statementRemove);
    sqlite3_finalize(statementRemove);
    UNLOCK_ACCESS
}

/*..........................................................................*/
static uint16_t Datamodel_getClusterDeviceCount(const dsuid_t *vdc_dsuid, uint8_t clusterId)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel getClusterDeviceCount Database closed!\n");
        UNLOCK_ACCESS
        return 0;
    }
    char destAddr [DSUID_STR_LEN];
    dsuid_to_string_upper(vdc_dsuid, destAddr);

    uint64_t clusterbit = (1 << clusterId);
    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT COUNT (*) FROM devices WHERE (vdcdsuid=?) \
                                  AND (groupMembership & ?) > 0;",
                                  -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 1, destAddr, -1, SQLITE_STATIC) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int64(statement, 2, clusterbit) == SQLITE_OK);

    sqlite3_step(statement);
    uint16_t countDevice = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    return countDevice;
}

/*..........................................................................*/
static uint16_t Datamodel_getGroupDeviceCount(const dsuid_t *vdc_dsuid, uint8_t groupId, uint8_t localZoneId)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel getGroupDeviceCount Database closed!\n");
        UNLOCK_ACCESS
        return 0;
    }
    char destAddr [DSUID_STR_LEN];
    dsuid_to_string_upper(vdc_dsuid, destAddr);

    uint64_t groupbit = (1 << groupId);
    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT COUNT (*) FROM devices WHERE (vdcdsuid=?) \
                                  AND (localZoneId=?) \
                                  AND (groupMembership & ?) > 0;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement,  1, destAddr, -1, SQLITE_STATIC) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement,   2, localZoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int64(statement, 3, groupbit) == SQLITE_OK);

    sqlite3_step(statement);
    uint16_t countDevice = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return countDevice;
}

/*..........................................................................*/
static QState Datamodel_parse_device(Datamodel *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        sqlite3_stmt *statement;
        int ret;
        LOCK_ACCESS

        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM devices, deviceButtons WHERE devices.deviceId == deviceButtons.deviceId;", -1, &statement, 0) == SQLITE_OK);
        do
        {
            ret = sqlite3_step(statement);
            if (ret == SQLITE_ROW)
            {
                DatamodelDevice device;
                Datamodel_fill_device_from_SQL(&device, statement);
                /* device is valid */
                /* update hash */
//                me->hash = Hash_Create(me->pageData, sizeof(DatamodelDevice), me->hash);
                /* update zone */
                DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByLocalId(device.localZoneId);
                Q_ASSERT(zoneInfo != 0);
                /* update last device address */
                me->lastValidDeviceId = device.deviceId;
                /* mark device as valid (active is default) */
                SetBitSet8(me->validDevices, device.deviceId);
                /* update active group stop timeout */
                Datamodel_SetGroupActiveStopTimeout(me, &device);
                /* set activity */
                if (device.active) {
                    SetBitSet8(me->activeDevices, device.deviceId);
                } else {
                    ClearBitSet8(me->activeDevices, device.deviceId);
                }
            }
        } while (ret == SQLITE_ROW);
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        Datamodel_Changed();

        QActive_postFIFO((QActive *)me, &l_flash_dma_done);
        return Q_HANDLED();
    }
    case FLASH_DMA_DONE_SIG: {
        return Q_TRAN(&Datamodel_parse_device_state);
    }
    }
    return Q_SUPER(&Datamodel_busy);
}

/*..........................................................................*/
static QState Datamodel_parse_device_state(Datamodel *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        QActive_postFIFO((QActive *)me, &l_flash_dma_done);
        return Q_HANDLED();
    }
    case FLASH_DMA_DONE_SIG: {
        /* final operations */
        {
            /* subscribe events */
            QActive_subscribe((QActive *)me, SYSTEM_TERM_SIG);
            /* publish configuration initialization done */
            {
                DatamodelConfigEvt * configInitEvt = Q_NEW(DatamodelConfigEvt, CONFIG_INIT_SIG);
                configInitEvt->dsmid = me->dsmid;
                configInitEvt->apartmentId = me->dsmProperties.apartmentId;
                configInitEvt->dsmInfoFlags = me->dsmProperties.flags;
                configInitEvt->autoPingInterval = me->dsmProperties.autoPingInterval;
                configInitEvt->oneTimePingInterval = me->dsmProperties.oneTimePingInterval;
                QF_publish((QEvent *)configInitEvt);
            }
            /* go to idle */
            return Q_TRAN(&Datamodel_idle);
        }
    }
    }
    return Q_SUPER(&Datamodel_busy);
}

/*..........................................................................*/
static uint16_t Datamodel_parse_groups_next_batch(Datamodel *me)
{
    sqlite3_stmt *statement;
    int ret;
#define MAX_GROUPS_IN_BATCH 64
    uint8_t count = MAX_GROUPS_IN_BATCH;
    static uint16_t last_g_id = 0;
    bool moreGroupsToParse = false;

    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM groups WHERE g_id>?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, last_g_id) == SQLITE_OK);

    do
    {
        ret = sqlite3_step(statement);
        if (ret == SQLITE_ROW)
        {
            /* zone is valid */
            int g_id = sqlite3_column_int(statement, 0);
            uint8_t stopTimeout = sqlite3_column_int(statement, 2);
            uint8_t remoteStopTimeout = sqlite3_column_int(statement, 3);
            uint8_t localZoneId = g_id >> 8;
            uint8_t groupId = g_id & 0xFF;
            last_g_id = g_id;

            /* cache stop timeout */
            Datamodel_UpdateCachedStopTimeout(me, localZoneId, groupId, max(stopTimeout, remoteStopTimeout));

            if (groupId >= FREE_USER_GROUP) {
                /* update user group in zone */
                DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByLocalId(localZoneId);
                SetBitSet8(zoneInfo->groupMembership, groupId);
            }
        }
    } while ((ret == SQLITE_ROW) && (count-- > 0 ));
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    if (ret == SQLITE_ROW) {
        moreGroupsToParse = true;
    }
    return moreGroupsToParse;
}

/*..........................................................................*/
static QState Datamodel_parse_groups(Datamodel *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Datamodel_convert_group_to_cluster(me);
        QTimeEvt_postIn(&me->opTimeEvt, (QActive *)me, 50 / BSP_MS_PER_TICKS);
        return Q_HANDLED();
    }
    case FLASH_OP_TIMEOUT_SIG: {
        if (Datamodel_parse_groups_next_batch(me)) {
            QTimeEvt_postIn(&me->opTimeEvt, (QActive *)me, 50 / BSP_MS_PER_TICKS);
            return Q_HANDLED();
        }
        return Q_TRAN(&Datamodel_parse_clusters);
    }
    }
    return Q_SUPER(&Datamodel_busy);
}

/*..........................................................................*/
static QState Datamodel_parse_clusters(Datamodel *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        me->groupId = FIRST_CLUSTER;
        QActive_postFIFO((QActive *)me, &l_flash_dma_done);
        return Q_HANDLED();
    }
    case FLASH_DMA_DONE_SIG: {
        DatamodelCluster cluster;
        if (Datamodel_getCluster(me->groupId, &cluster)) {
            /* update hash */
            me->hash = Hash_Create((const uint8_t *)&(cluster),
                                   sizeof(DatamodelCluster),
                                   me->hash);
            Datamodel_NotifyUserGroup(cluster.id, cluster.stateMachineGroupId);
            /* cache stop timeout */
            Datamodel_UpdateCachedStopTimeout(me, 0, cluster.id, max(cluster.stopTimeout, cluster.remoteStopTimeout));
        }
        /* increment group ID */
        if (++me->groupId > LAST_CLUSTER) {
            return Q_TRAN(&Datamodel_parse_device);
        }
        QActive_postFIFO((QActive *)me, &l_flash_dma_done);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Datamodel_busy);
}

/*..........................................................................*/
/* For the defer/recall mechanism to work correctly Q_HANDLED() must never be called
 * in this state (idle) except for Q_ENTRY_SIG and Q_EXIT_SIG. Always use
 * Q_TRAN(&Datamodel_idle) instead.
 */
static QState Datamodel_idle(Datamodel *me, QEvent const *e)
{
    switch (e->sig) {
    /* ------------------------------------ */
    case Q_ENTRY_SIG: {
        /* recall deferred event (request) */
        QActive_recall((QActive *)me, &me->deferredQueue);
        return Q_HANDLED();
    }
    /* ------------------------------------ */
    case SYSTEM_TERM_SIG: {
        return Q_TRAN(&Datamodel_stopped);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DEVICE_INFO_OF_INDEX_SIG: {
        uint8_t error;
        DatamodelReqDeviceInfoOfIndexEvt const * reqEvt = (DatamodelReqDeviceInfoOfIndexEvt const *)e;
        /* save request */
        me->requester = reqEvt->basic.requester;
        if ((reqEvt->modifier == DEVICE_INFO_BY_INDEX_ONLY_ACTIVE) ||
            (reqEvt->modifier == DEVICE_INFO_BY_INDEX_ONLY_INACTIVE) ||
            (reqEvt->modifier == DEVICE_INFO_BY_INDEX)) {
            /* modifier is valid */
            DatamodelDeviceModifier modifier = (DatamodelDeviceModifier)reqEvt->modifier;

            uint16_t deviceIndex = reqEvt->deviceIndex;
            if (reqEvt->zoneId == 0) {
                uint16_t deviceId = 0;
                if (Datamodel_get_device_by_index_from_vdc(me,
                                                           &reqEvt->dsuid,
                                                           modifier,
                                                           deviceIndex,
                                                           &deviceId)) {
                    Datamodel_get_device_id_entry(me, deviceId, reqEvt->privateId);
                    return Q_TRAN(&Datamodel_idle);
                }
                /* device not found */
                error = ERROR_DEVICE_NOT_FOUND;
            } else {
                /* specific zone */
                DatamodelZoneInfo* zoneInfo;
                if (reqEvt->zoneIdModifier == ZONE_INFO_BY_INDEX) {
                    zoneInfo = Datamodel_GetZoneInfoByZoneId(reqEvt->zoneId);
                } else {
                    zoneInfo = Datamodel_GetZoneInfoByLocalId(reqEvt->zoneId);
                }
                if ((zoneInfo != 0) && (zoneInfo->super.magic == DATAMODEL_MAGIC)) {
                    /* zone is valid and available */
                    DatamodelDevice output_device;
                    if (Datamodel_get_device_entry_by_index(me,
                                                            zoneInfo->super.localZoneId,
                                                            deviceIndex,
                                                            modifier,
                                                            &reqEvt->dsuid,
                                                            &output_device)) {
                        DatamodelResDeviceEntryEvt * resEvt = Q_NEW(DatamodelResDeviceEntryEvt, DATAMODEL_RES_DEVICE_ENTRY_SIG);
                        resEvt->state = (IsBitSet8Set(me->activeDevices, output_device.deviceId) ? TRUE : FALSE);
                        resEvt->privateId = reqEvt->privateId;
                        resEvt->device = output_device;
                        QActive_postFIFO(me->requester, (QEvent *)resEvt);
                        return Q_TRAN(&Datamodel_idle);
                    }
                    /* device not found */
                    error = ERROR_DEVICE_NOT_FOUND;
                } else {
                    /* zone not found */
                    error = ERROR_ZONE_NOT_FOUND;
                }
            }
        } else {
            error = ERROR_WRONG_MODIFIER;
        }
        /* post error */
        Datamodel_PostLongError(me->requester, error, reqEvt->privateId);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_ZONE_MODIFY_SIG: {
        DatamodelReqZoneModifyEvt const * reqEvt = (DatamodelReqZoneModifyEvt const *)e;
        me->requester  = reqEvt->basic.requester;
        uint8_t error = ERROR_WRONG_MODIFIER;
        if (reqEvt->modifier == ZONE_MODIFY_ADD) {
            /* add zone */
            uint16_t zoneId = reqEvt->zoneId;
            if (zoneId != 0) {
                DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByZoneId(zoneId);
                if (zoneInfo == 0) {
                    /* zone does not exist yet */
                    if ((Datamodel_ZoneCount(me, &reqEvt->dsuid) != 0) && (me->numberOfZones < MAX_ZONES)) {
                        Datamodel_CreateZone(me, zoneId);
                        /* send response */
                        PostFIFO_ErrorOK(me->requester);
                        return Q_TRAN(&Datamodel_idle);
                    } else {
                        /* max number of zones reached */
                        error = ERROR_NO_FURTHER_ZONES;
                    }
                } else {
                    /* zone already exists, but not for vdc */
                    Datamodel_publish_zone_added(zoneInfo->super.localZoneId, zoneInfo->super.zoneId);
                    /* send response */
                    PostFIFO_ErrorOK(me->requester);
                    return Q_TRAN(&Datamodel_idle);
                }
            } else {
                /* invalid zone ID */
                error = ERROR_WRONG_PARAMETER;
            }
        } else if (reqEvt->modifier == ZONE_MODIFY_REMOVE){
            /* remove zone */
            DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByZoneId(reqEvt->zoneId);
            if (zoneInfo != 0) {
                uint8_t localZoneId = zoneInfo->super.localZoneId;
                uint16_t deviceCount = Datamodel_get_zone_device_count(me, localZoneId, ALL, &reqEvt->dsuid);
                if (deviceCount != 0) {
                    /* last zone can not be deleted */
                    error = ERROR_ZONE_CAN_NOT_BE_DELETED;
                } else if (Heating_IsControllerRunning(localZoneId)) {
                    /* zone is running heating controller */
                    error = ERROR_ZONE_CAN_NOT_BE_DELETED;
                } else {
                    /* check if zone is used in any other vdc */
                    if (Datamodel_get_zone_usage(me, localZoneId) == 0) {
                        Datamodel_RemoveZone(me, reqEvt->zoneId);
                    }
                    return Q_TRAN(&Datamodel_idle);
                }
            } else {
                /* zone not found */
                error = ERROR_ZONE_NOT_FOUND;
            }
        } else {
            /* change zone ID */
            DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByZoneId(reqEvt->zoneId);
            if (zoneInfo != 0) {
                uint16_t newZoneId = reqEvt->newZoneId;
                if (newZoneId != 0) {
                    if (Datamodel_GetZoneInfoByZoneId(newZoneId) == 0) {
                        /* zone does not exist yet */
                        zoneInfo->super.zoneId = newZoneId;
                        /* publish changed zone */
                        Datamodel_publish_zone_added(zoneInfo->super.localZoneId, reqEvt->newZoneId);
                        Datamodel_zone_modify_zoneId(me, zoneInfo->super.localZoneId, newZoneId);
                        return Q_TRAN(&Datamodel_idle);
                    } else {
                        /* zone already exists */
                        error = ERROR_ZONE_ALREADY_EXISTS;
                    }
                } else {
                    /* invalid zone ID */
                    error = ERROR_WRONG_PARAMETER;
                }
            } else {
                // check if new zone exists
                uint16_t newZoneId = reqEvt->newZoneId;
                DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByZoneId(newZoneId);
                if (zoneInfo != 0) {
                    /* publish changed zone */
                    Datamodel_publish_zone_added(zoneInfo->super.localZoneId, reqEvt->newZoneId);
                    Datamodel_zone_modify_zoneId(me, zoneInfo->super.localZoneId, newZoneId);
                    return Q_TRAN(&Datamodel_idle);
                } else {
                    /* zone not found */
                    error = ERROR_ZONE_NOT_FOUND;
                }
            }
        }
        /* post error */
        Datamodel_PostError(me->requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_ZONE_COUNT_SIG: {

        DatamodelReqDsmEvt const * reqEvt  = (DatamodelReqDsmEvt const *)e;
        uint8_t zoneCount = Datamodel_ZoneCount(me, &reqEvt->dsuid);

        DatamodelResZoneCountEvt * resEvt =
            Q_NEW(DatamodelResZoneCountEvt, DATAMODEL_RES_ZONE_COUNT_SIG);
        resEvt->zoneCount = zoneCount;
        QActive_postFIFO(reqEvt->requester, (QEvent *)resEvt);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_ZONE_DEVICE_COUNT_SIG: {
        uint8_t error;
        DatamodelReqZoneDeviceCountEvt const * reqEvt = (DatamodelReqZoneDeviceCountEvt const *)e;
        uint8_t modifier = reqEvt->modifier;
        me->requester  = reqEvt->basic.requester;
        if ((modifier == ZONE_DEVICE_COUNT_ONLY_ACTIVE) ||
            (modifier == ZONE_DEVICE_COUNT_ONLY_INACTIVE) ||
            (modifier == ZONE_DEVICE_COUNT_ALL)) {
            /* modifier is valid */
            uint8_t localZoneId;
            bool zoneIdValid = TRUE;
            if (reqEvt->zoneIdModifier == ZONE_INFO_BY_ID) {
                /* zone ID is local ID */
                localZoneId = (uint8_t)reqEvt->zoneId;
                DatamodelZoneInfo const *zoneInfo = Datamodel_GetZoneInfoByLocalId(localZoneId);
                if ((zoneInfo == 0) || (zoneInfo->super.magic != DATAMODEL_MAGIC)) {
                    zoneIdValid = FALSE;
                }
            } else {
                /* zone ID is global ID */
                if (reqEvt->zoneId == 0) {
                    localZoneId = 0;
                } else {
                    zoneIdValid = Datamodel_GetLocalZoneId(reqEvt->zoneId, &localZoneId);
                }
            }
            if (zoneIdValid) {
                /* zone ID is valid */
                DatamodelDeviceModifier deviceModifier = l_dsmApiZoneDevCountModifiers[modifier];
                uint16_t deviceCount = Datamodel_get_zone_device_count(me, localZoneId, deviceModifier, &reqEvt->dsuid);

                /* post data to requester */
                DatamodelResZoneDeviceCountEvt * resEvt =
                    Q_NEW(DatamodelResZoneDeviceCountEvt, DATAMODEL_RES_ZONE_DEVICE_COUNT_SIG);
                resEvt->deviceCount = deviceCount;
                QActive_postFIFO(me->requester, (QEvent *)resEvt);

                return Q_TRAN(&Datamodel_idle);
            }
            /* zone not found */
            error = ERROR_ZONE_NOT_FOUND;
        } else {
            error = ERROR_WRONG_MODIFIER;
        }
        /* post error */
        Datamodel_PostError(me->requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_ZONE_INFO_OF_INDEX_SIG: {

        DatamodelReqZoneInfoOfIndexEvt const * reqEvt = (DatamodelReqZoneInfoOfIndexEvt const *)e;
        DatamodelZoneInfo *zoneInfo;

        uint8_t numberOfZones = Datamodel_ZoneCount(me, &(reqEvt->dsuid));
        bool success = false;
        if (reqEvt->zoneIndex < numberOfZones) {
            success = Datamodel_GetZoneInfoPerVdcByIndex(me, reqEvt->zoneIndex, &(reqEvt->dsuid), &zoneInfo);
        }

        if (success) {
            Datamodel_SendZoneInfo(reqEvt->basic.requester, zoneInfo);
        } else {
            /* post error */
            Datamodel_PostError(reqEvt->basic.requester, ERROR_NO_FURTHER_ZONES);
        }

        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_ZONE_INFO_BY_ZONE_ID_SIG: {
        DatamodelReqZoneInfoByIdEvt const * reqEvt = (DatamodelReqZoneInfoByIdEvt const *)e;
        DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByZoneId(reqEvt->zoneId);
        if (zoneInfo && (zoneInfo->super.magic == DATAMODEL_MAGIC)) {
            Datamodel_SendZoneInfo(reqEvt->basic.requester, zoneInfo);
            return Q_TRAN(&Datamodel_idle);
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, ERROR_ZONE_NOT_FOUND);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_ZONE_LOCAL_ID_SIG: {
        DatamodelReqZoneLocalIdEvt const * reqEvt = (DatamodelReqZoneLocalIdEvt const *)e;
        uint16_t zoneId = reqEvt->zoneId;
        DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByZoneId(zoneId);
        if (zoneInfo && (zoneInfo->super.magic == DATAMODEL_MAGIC)) {
            /* zone found */
            DatamodelResZoneLocalIdEvt * resEvt =
                Q_NEW(DatamodelResZoneLocalIdEvt, DATAMODEL_RES_ZONE_LOCAL_ID_SIG);
            resEvt->localZoneId = zoneInfo->super.localZoneId;
            QActive_postFIFO(reqEvt->basic.requester, (QEvent *)resEvt);
            return Q_TRAN(&Datamodel_idle);
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, ERROR_ZONE_NOT_FOUND);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DEVICE_STATE_SIG: {
        DatamodelReqDeviceIdEvt const * reqEvt = (DatamodelReqDeviceIdEvt const *)e;
        if ((reqEvt->deviceId <= MAX_DEVICE) && IsBitSet8Set(me->validDevices, reqEvt->deviceId)) {
            DatamodelResDeviceStateEvt * resEvt = Q_NEW(DatamodelResDeviceStateEvt, DATAMODEL_RES_DEVICE_STATE_SIG);
            resEvt->active = (IsBitSet8Set(me->activeDevices, reqEvt->deviceId) ? TRUE : FALSE);
            QActive_postFIFO(reqEvt->basic.requester, (QEvent *)resEvt);
            return Q_TRAN(&Datamodel_idle);
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, ERROR_DEVICE_NOT_FOUND);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DEVICE_ENTRY_BY_ID_SIG: {
        DatamodelReqDeviceIdEvt const * reqEvt = (DatamodelReqDeviceIdEvt const *)e;
        if ((reqEvt->deviceId <= MAX_DEVICE) && IsBitSet8Set(me->validDevices, reqEvt->deviceId)) {
            /* save request */
            me->requester = reqEvt->basic.requester;
            Datamodel_get_device_id_entry(me, reqEvt->deviceId, reqEvt->privateId);
            return Q_TRAN(&Datamodel_idle);
        }
        /* post error */
        Datamodel_PostLongError(reqEvt->basic.requester, ERROR_DEVICE_NOT_FOUND, reqEvt->privateId);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DEVICE_ENTRY_BY_SERIAL_NUMBER_SIG: {
        DatamodelReqDeviceSerialEvt const * reqEvt = (DatamodelReqDeviceSerialEvt const *)e;
        /* save request */
        me->requester = reqEvt->basic.requester;
        /* initialize first page */
        Datamodel_get_device_dsuid_entry(me, reqEvt->dsuid, reqEvt->privateId);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_GROUP_MEMBERSHIP_MODIFY_SIG: {
        uint8_t error;
        DatamodelReqGroupMembershipModifyEvt const * reqEvt = (DatamodelReqGroupMembershipModifyEvt const *)e;
        if (reqEvt->groupId <= MAX_GROUP) {
            /* group ID is valid */
            if ((reqEvt->deviceAddress <= MAX_DEVICE) && IsBitSet8Set(me->validDevices, reqEvt->deviceAddress)) {
                /* device is valid */
                /* save request */
                me->requester = reqEvt->basic.requester;
                Datamodel_group_membership_modify(me, reqEvt->deviceAddress, reqEvt->groupId, reqEvt->modifier, reqEvt->sendDownstream);
                return Q_TRAN(&Datamodel_idle);
            } else {
                /* device not found */
                error = ERROR_DEVICE_NOT_FOUND;
            }
        } else {
            error = ERROR_WRONG_PARAMETER;
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DEVICE_SET_NAME_SIG: {
        DatamodelReqDeviceNameEvt const * reqEvt = (DatamodelReqDeviceNameEvt const *)e;
        me->requester = reqEvt->basic.requester;
        if ((reqEvt->deviceAddress >= MIN_DEVICE) && (reqEvt->deviceAddress <= MAX_DEVICE)) {
            if (IsBitSet8Set(me->validDevices, reqEvt->deviceAddress)) {
                uint16_t deviceId = reqEvt->deviceAddress;
                const char *name = (const char*) reqEvt->name;
                Datamodel_set_device_name(me, deviceId, name);
                if (me->requester) {
                    /* send response */
                    PostFIFO_ErrorOK(me->requester);
                    /* propagate name to vdc */
                    vdc_api_device_set_name(deviceId, name);
                }
                return Q_TRAN(&Datamodel_idle);
            }
        }
        if (me->requester) {
            /* post error */
            Datamodel_PostError(reqEvt->basic.requester, ERROR_DEVICE_NOT_FOUND);
        }
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_GROUP_SET_NAME_SIG: {
        uint8_t error;
        DatamodelReqGroupNameEvt const * reqEvt = (DatamodelReqGroupNameEvt const *)e;
        if (Datamodel_IsCluster(reqEvt->groupId)) {
            me->requester = reqEvt->basic.requester;
            Datamodel_setClusterName(reqEvt->groupId, reqEvt->name);
            PostFIFO_ErrorOK(me->requester);
            return Q_TRAN(&Datamodel_idle);
        }
        DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByZoneId(reqEvt->zoneId);
        if (zoneInfo != 0) {
            if ((reqEvt->groupId <= MAX_GROUP) && IsBitSet8Set(zoneInfo->groupMembership, reqEvt->groupId)) {
                me->requester = reqEvt->basic.requester;
                Datamodel_set_group_name(me, zoneInfo->super.localZoneId, reqEvt->groupId, reqEvt->name);
                return Q_TRAN(&Datamodel_idle);
            }
            error = ERROR_GROUP_NOT_FOUND;
        } else {
            error = ERROR_ZONE_NOT_FOUND;
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_GROUP_SET_STOP_TIMEOUT_SIG: {
        uint8_t error;
        DatamodelReqGroupStopTimeoutEvt const * reqEvt = (DatamodelReqGroupStopTimeoutEvt const *)e;
        if (Datamodel_IsCluster(reqEvt->groupId)) {
            me->requester = reqEvt->basic.requester;
            Datamodel_setClusterStopTimeout(reqEvt->groupId, reqEvt->stopTimeout);
            /* send response */
            PostFIFO_ErrorOK(me->requester);
            return Q_TRAN(&Datamodel_idle);
        }
        DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByZoneId(reqEvt->zoneId);
        if (zoneInfo != 0) {
            if ((reqEvt->groupId <= MAX_GROUP) && IsBitSet8Set(zoneInfo->groupMembership, reqEvt->groupId)) {
                me->requester = reqEvt->basic.requester;
                Datamodel_set_group_stop_timeout(me, zoneInfo->super.localZoneId, reqEvt->groupId, reqEvt->stopTimeout);
                return Q_TRAN(&Datamodel_idle);
            }
            error = ERROR_GROUP_NOT_FOUND;
        } else {
            error = ERROR_ZONE_NOT_FOUND;
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_GROUP_SET_COLOR_SIG: {
        uint8_t error;
        DatamodelReqGroupColorEvt const * reqEvt = (DatamodelReqGroupColorEvt const *)e;
        DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByZoneId(reqEvt->zoneId);
        if (zoneInfo != 0) {
            if ((reqEvt->groupId <= MAX_GROUP) && IsBitSet8Set(zoneInfo->groupMembership, reqEvt->groupId)) {
                if ((reqEvt->groupId >= USER_GROUP) && (reqEvt->groupId <= MAX_USER_GROUP_WITH_HSM)) {
                    me->requester = reqEvt->basic.requester;
                    Datamodel_set_group_color(me, zoneInfo->super.localZoneId, reqEvt->groupId, reqEvt->color);
                    return Q_TRAN(&Datamodel_idle);
                } else {
                    error = ERROR_WRONG_PARAMETER;
                }
            } else {
                error = ERROR_GROUP_NOT_FOUND;
            }
        } else {
            error = ERROR_ZONE_NOT_FOUND;
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_ZONE_SET_NAME_SIG: {
        DatamodelReqZoneNameEvt const * reqEvt = (DatamodelReqZoneNameEvt const *)e;
        DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByZoneId(reqEvt->zoneId);
        if (zoneInfo != 0) {
            me->requester = reqEvt->basic.requester;
            Datamodel_zone_modify_name(me, zoneInfo->super.localZoneId, reqEvt->name, &reqEvt->dsuid);
            return Q_TRAN(&Datamodel_idle);
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, ERROR_ZONE_NOT_FOUND);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_GET_HASH_SIG: {
        DatamodelResHashEvt * resEvt = Q_NEW(DatamodelResHashEvt, DATAMODEL_RES_GET_HASH_SIG);
        resEvt->hash = me->hash;
        resEvt->changeCount = me->changeCount;
        QActive_postFIFO(((DatamodelReqBasicEvt const *)e)->requester, (QEvent *)resEvt);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DSM_SET_APARTMENT_ID_SIG: {
        DatamodelReqSetApartmentIdEvt const * reqEvt = (DatamodelReqSetApartmentIdEvt const *)e;
        me->requester = reqEvt->basic.requester;
        Datamodel_set_dsm_apartment_id(me, reqEvt->apartmentId);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DSM_SET_TEMPERATURE_LIMITS_SIG: {
        DatamodelReqSetTemperatureLimitsEvt const * reqEvt = (DatamodelReqSetTemperatureLimitsEvt const *)e;
        if (reqEvt->errorTemperature > reqEvt->warnTemperature) {
            me->requester = reqEvt->basic.requester;
            Datamodel_set_dsm_temperature_limits(me, reqEvt->warnTemperature, reqEvt->errorTemperature);
        } else {
            /* post error */
            Datamodel_PostError(reqEvt->basic.requester, ERROR_WRONG_PARAMETER);
        }
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DSM_SET_PING_INTERVALS_SIG: {
        DatamodelReqSetPingIntervalsEvt const * reqEvt = (DatamodelReqSetPingIntervalsEvt const *)e;
        me->requester = reqEvt->basic.requester;
        Datamodel_set_dsm_ping_intervals(me, reqEvt->autoPingInterval, reqEvt->oneTimePingInterval);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DSM_SET_DSM_FLAGS_SIG: {
        DatamodelReqSetDsmFlagsEvt const * reqEvt = (DatamodelReqSetDsmFlagsEvt const *)e;
        me->requester = reqEvt->basic.requester;
        Datamodel_set_dsm_flags(me, reqEvt->flags);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DSM_SET_UPSTREAM_FREQUENCY_SIG: {
        DatamodelReqSetDsmUpstreamFreqEvt const * reqEvt = (DatamodelReqSetDsmUpstreamFreqEvt const *)e;
        me->requester = reqEvt->basic.requester;
        Datamodel_set_dsm_upstream_frequency(me, reqEvt->upstreamFrequency);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DSM_SET_NAME_SIG: {
        DatamodelReqNameEvt const * reqEvt = (DatamodelReqNameEvt const *)e;
        me->requester = reqEvt->basic.requester;
        Datamodel_set_dsm_name(me, reqEvt->name, &reqEvt->dsuid);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DSM_PROPERTIES_SIG: {
        DatamodelReqDsmEvt const * reqEvt = (DatamodelReqDsmEvt const *)e;
        /* save request */
        me->requester = reqEvt->requester;
        Datamodel_get_dsm_properties(me, &reqEvt->dsuid);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_SCENE_INFO_SIG: {
        uint8_t error;
        DatamodelReqSceneInfoEvt const * reqEvt = (DatamodelReqSceneInfoEvt const *)e;
        DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByZoneId(reqEvt->zoneId);
        if (zoneInfo != 0) {
            if ((reqEvt->groupId <= MAX_GROUP) && IsBitSet8Set(zoneInfo->groupMembership, reqEvt->groupId) ) {
                if (((reqEvt->groupId > 0) && (reqEvt->sceneIndex >= MIN_SCENE) && (reqEvt->sceneIndex <= MAX_SCENE)) ||
                    ((reqEvt->groupId == 0) && (reqEvt->sceneIndex >= (MIN_SCENE + START_APARTMENT_SCENES)) && (reqEvt->sceneIndex <= (MAX_SCENE + START_APARTMENT_SCENES)))) {
                    me->requester = reqEvt->basic.requester;
                    Datamodel_get_scene_info(me,
                                             zoneInfo->super.localZoneId,
                                             reqEvt->groupId,
                                             reqEvt->sceneIndex % START_APARTMENT_SCENES); // shift sceneIndex from 64-127 to 0-63 for group0
                    return Q_TRAN(&Datamodel_idle);
                } else {
                    error = ERROR_WRONG_PARAMETER;
                }
            } else {
                error = ERROR_GROUP_NOT_FOUND;
            }
        } else {
            error = ERROR_ZONE_NOT_FOUND;
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_SCENE_SET_NAME_SIG: {
        uint8_t error;
        DatamodelReqSceneNameEvt const * reqEvt = (DatamodelReqSceneNameEvt const *)e;
        DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByZoneId(reqEvt->zoneId);
        if (zoneInfo != 0) {
            if ((reqEvt->groupId <= MAX_GROUP) && IsBitSet8Set(zoneInfo->groupMembership, reqEvt->groupId) ) {
                if (((reqEvt->groupId > 0) && (reqEvt->sceneIndex >= MIN_SCENE) && (reqEvt->sceneIndex <= MAX_SCENE)) ||
                    ((reqEvt->groupId == 0) && (reqEvt->sceneIndex >= (MIN_SCENE + START_APARTMENT_SCENES)) && (reqEvt->sceneIndex <= (MAX_SCENE + START_APARTMENT_SCENES)))) {
                    me->requester = reqEvt->basic.requester;
                    /* do nothing for now. TODO: implement functionality for future features when needed */
                    /* send response */
                    PostFIFO_ErrorOK(me->requester);
                    return Q_TRAN(&Datamodel_idle);
                } else {
                    error = ERROR_WRONG_PARAMETER;
                }
            } else {
                error = ERROR_GROUP_NOT_FOUND;
            }
        } else {
            error = ERROR_ZONE_NOT_FOUND;
        }
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_ADD_DEVICE_SIG: {
        DatamodelDevice device;
        DatamodelReqAddDeviceEvt const * reqEvt = (DatamodelReqAddDeviceEvt const *)e;
        me->requester = reqEvt->basic.requester;
        uint16_t deviceId;
        if (!Datamodel_SelectRandomDeviceId(me, &deviceId)) {
            /* post error */
            /* TODO Change error to resources exhausted */
            Datamodel_PostError(me->requester, ERROR_REQUEST_CAN_NOT_BE_EXECUTED);
            return Q_TRAN(&Datamodel_idle);
        }
        /* initialize device data - adds device to zone, too */
        Datamodel_InitDevice(&device,
                             reqEvt->dsuid,
                             deviceId,
                             reqEvt->vid,
                             reqEvt->fid,
                             reqEvt->ver,
                             me->dsmProperties.flags & LOCK_NEW_REGISTERED_DEVICE_MASK? 1 : 0,
                             reqEvt->vdcdsuid,
                             reqEvt->zoneId);
        Datamodel_add_device(me, &device);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_PING_DEVICE_STATE_SIG: {
        DatamodelReqPingDeviceStateEvt const * reqEvt = (DatamodelReqPingDeviceStateEvt const *)e;
        uint16_t deviceId = reqEvt->deviceId;
        if (reqEvt->basic.requester != 0) {
            /*  check device ID */
            if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                /* get next device ID */
                uint16_t oneTimePingInitialDeviceId = reqEvt->oneTimePingInitialDeviceId;
                uint16_t nextDeviceId = deviceId;
                bool oneTimePingDone = FALSE;
                do {
                    ++nextDeviceId;
                    if (nextDeviceId > me->lastValidDeviceId) {
                        nextDeviceId = MIN_DEVICE;
                    }
                    if (nextDeviceId == oneTimePingInitialDeviceId) {
                        oneTimePingDone = TRUE;
                    }
                } while (IsBitSet8Cleared(me->validDevices, nextDeviceId) && (nextDeviceId != deviceId));
                if (IsBitSet8Set(me->validDevices, nextDeviceId)) {
                    me->requester = reqEvt->basic.requester;
                    Datamodel_ping_device_state(me, deviceId, nextDeviceId, reqEvt->active, oneTimePingDone);
                    return Q_TRAN(&Datamodel_idle);
                }
            }
            /* post error */
            Datamodel_PostError(reqEvt->basic.requester, ERROR_DEVICE_NOT_FOUND);
        }
        /* check if device still valid (it could have been removed) */
        if (IsBitSet8Set(me->validDevices, deviceId)) {
            /* update device accessibility */
            if (reqEvt->active) {
                /* device is now active */
                if (IsBitSet8Cleared(me->activeDevices, deviceId)) {
                    /* change device to active */
                    SetBitSet8(me->activeDevices, deviceId);
                    Datamodel_set_device_active(me, deviceId, TRUE);
                    /* send device accessibility change */
                    Datamodel_send_device_accessibility(me, deviceId, TRUE);
                }
            } else {
                /* device is now inactive */
                if (IsBitSet8Set(me->activeDevices, deviceId)) {
                    /* change device to inactive */
                    ClearBitSet8(me->activeDevices, deviceId);
                    Datamodel_set_device_active(me, deviceId, FALSE);
                    /* send device accessibility change */
                    Datamodel_send_device_accessibility(me, deviceId, FALSE);
                }
            }
        }
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_ADD_GROUP_SIG: {
        uint8_t error;
        DatamodelReqZoneGroupAddEvt const * reqEvt = (DatamodelReqZoneGroupAddEvt const *)e;
        DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByZoneId(reqEvt->zoneId);
        if (zoneInfo != 0) {
            uint8_t groupId = reqEvt->groupId;
            uint8_t stateMachineGroupId = reqEvt->stateMachineGroupId;
            if ((groupId >= USER_GROUP) && (groupId <= MAX_GROUP) && (stateMachineGroupId < USER_GROUP)) {
                if (IsBitSet8Cleared(zoneInfo->groupMembership, groupId)) {
                    if ((groupId >= MIN_USER_GROUP_WITHOUT_HSM) && (stateMachineGroupId != 0)) {
                        error = ERROR_WRONG_PARAMETER;
                    } else {
                        SetBitSet8(zoneInfo->groupMembership, groupId);
                        me->requester = reqEvt->basic.requester;
                        Datamodel_InitGroup(me, zoneInfo->super.localZoneId, groupId, stateMachineGroupId, NULL);
                        /* update cached stop timeout */
                        uint8_t stopTimeout = (groupId == SHADE_GROUP) ? 120 : 0;
                        Datamodel_UpdateCachedStopTimeout(me, zoneInfo->super.localZoneId, groupId, stopTimeout);
                        /* notify user group */
                        Datamodel_NotifyUserGroup(groupId, stateMachineGroupId);
                        /* notify change */
                        Datamodel_NotifyChange();
                        /* send response */
                        PostFIFO_ErrorOK(me->requester);
                        return Q_TRAN(&Datamodel_idle);
                    }
                } else {
                    /* user group exists already */
                    error = ERROR_GROUP_ALREADY_EXISTS;
                }
            } else {
                error = ERROR_WRONG_PARAMETER;
            }
        } else {
            error = ERROR_ZONE_NOT_FOUND;
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_REMOVE_GROUP_SIG: {
        uint8_t error = ERROR_WRONG_PARAMETER;
        DatamodelReqZoneGroupRemoveEvt const * reqEvt = (DatamodelReqZoneGroupRemoveEvt const *)e;
        DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByZoneId(reqEvt->zoneId);
        if (zoneInfo != 0) {
            if (reqEvt->groupId > MAX_GROUP) {
                error = ERROR_WRONG_PARAMETER;
            } else if ((reqEvt->groupId < FREE_USER_GROUP) || (reqEvt->groupId == ROOM_TEMPERATURE_CONTROL_GROUP)) {
                error = ERROR_GROUP_CAN_NOT_BE_DELETED;
            } else if (IsBitSet8Cleared(zoneInfo->groupMembership, reqEvt->groupId)) {
                error = ERROR_GROUP_NOT_FOUND;
            } else {
                ClearBitSet8(zoneInfo->groupMembership, reqEvt->groupId);
                /* send remove user group from zone command to SelectA */
                RemoveGroupFromZoneEvt * remEvt = Q_NEW(RemoveGroupFromZoneEvt, REM_GROUP_FROM_ZONE_SIG);
                remEvt->groupId = reqEvt->groupId;
                remEvt->localZoneId = zoneInfo->super.localZoneId;
                QActive_postFIFO(AO_SelectA, (QEvent *)remEvt);
                /* remove user group from devices in flash */
                me->requester = reqEvt->basic.requester;
                Datamodel_zone_group_modify_remove(me, zoneInfo->super.localZoneId, reqEvt->groupId);
                return Q_TRAN(&Datamodel_idle);
            }
        } else {
            error = ERROR_ZONE_NOT_FOUND;
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_ZONE_GROUP_INFO_SIG: {
        uint8_t error;
        DatamodelReqZoneGroupInfoEvt const * reqEvt = (DatamodelReqZoneGroupInfoEvt const *)e;
        DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByZoneId(reqEvt->zoneId);
        if (zoneInfo != 0) {
            uint8_t groupId;
            if (reqEvt->modifier == ZONE_INFO_BY_INDEX) {
                groupId = Datamodel_GetGroupIdByGroupIndex(zoneInfo, reqEvt->groupId);
            } else {
                groupId = reqEvt->groupId;
            }
            if ((groupId <= MAX_GROUP) && IsBitSet8Set(zoneInfo->groupMembership, groupId)) {
                me->requester = reqEvt->basic.requester;
                Datamodel_get_group_info(me, zoneInfo->super.localZoneId, groupId, &reqEvt->vdcdsuid);
                return Q_TRAN(&Datamodel_idle);
            }
            error = ERROR_GROUP_NOT_FOUND;
        } else {
            error = ERROR_ZONE_NOT_FOUND;
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_CLUSTER_INFO_SIG: {
        DatamodelReqClusterInfoEvt const *reqEvt = (DatamodelReqClusterInfoEvt const *)e;
        me->requester = reqEvt->basic.requester;
        DatamodelCluster cluster;
        Datamodel_getCluster(reqEvt->clusterId, &cluster);
        uint16_t deviceCounts = Datamodel_getClusterDeviceCount(&reqEvt->dsuid,reqEvt->clusterId);
        Datamodel_SendClusterInfoResponse(me, &cluster, deviceCounts);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_CLUSTER_SET_NAME_SIG: {
        DatamodelReqClusterNameEvt const *reqEvt = (DatamodelReqClusterNameEvt const *)e;
        me->requester = reqEvt->basic.requester;
        Datamodel_setClusterName(reqEvt->clusterId, reqEvt->name);
        /* send response */
        PostFIFO_ErrorOK(me->requester);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_CLUSTER_SET_STOP_TIMEOUT_SIG: {
        DatamodelReqClusterStopTimeoutEvt const *reqEvt = (DatamodelReqClusterStopTimeoutEvt const *)e;
        me->requester = reqEvt->basic.requester;
        Datamodel_setClusterStopTimeout(reqEvt->clusterId, reqEvt->stopTimeout);
        /* send response */
        PostFIFO_ErrorOK(me->requester);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_CLUSTER_SET_COLOR_SIG: {
        DatamodelReqClusterColorEvt const *reqEvt = (DatamodelReqClusterColorEvt const *)e;
        me->requester = reqEvt->basic.requester;
        int remoteStopTimeout = 0;
        if (reqEvt->color == SHADE_GROUP) {
            remoteStopTimeout  = 120;
        } 
        Datamodel_setClusterRemoteStopTimeout(reqEvt->clusterId, remoteStopTimeout, reqEvt->color);

        /* notify selectA statemachine */
        Datamodel_NotifyUserGroup(reqEvt->clusterId, reqEvt->color);

        /* send response */
        PostFIFO_ErrorOK(me->requester);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_CLUSTER_SET_LOCATION_CLASS_SIG: {
        DatamodelReqClusterLocationClassEvt const *reqEvt = (DatamodelReqClusterLocationClassEvt const *)e;
        me->requester = reqEvt->basic.requester;
        Datamodel_setLocationProtection(reqEvt->clusterId, reqEvt->location, reqEvt->floor, reqEvt->protectionClass);
        /* send response */
        PostFIFO_ErrorOK(me->requester);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_CLUSTER_SET_CONFIG_LOCK_SIG: {
        DatamodelReqClusterConfigLockEvt const *reqEvt = (DatamodelReqClusterConfigLockEvt const *)e;
        me->requester = reqEvt->basic.requester;
        Datamodel_setConfigurationLock(reqEvt->clusterId, reqEvt->lock);
        /* send response */
        PostFIFO_ErrorOK(me->requester);
        return Q_TRAN(&Datamodel_idle);
    }
      /* ------------------------------------ */
    case DATAMODEL_REQ_CLUSTER_SET_SCENE_LOCK_SIG: {
        DatamodelReqClusterSceneLockEvt const *reqEvt = (DatamodelReqClusterSceneLockEvt const *)e;
        me->requester = reqEvt->basic.requester;
        Datamodel_setSceneLock(reqEvt->clusterId, reqEvt->sceneLock, sizeof(reqEvt->sceneLock));
        /* send response */
        PostFIFO_ErrorOK(me->requester);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DEVICE_SET_ZONE_SIG: {
        uint8_t error = ERROR_WRONG_PARAMETER;
        DatamodelReqDeviceSetZoneEvt const * reqEvt = (DatamodelReqDeviceSetZoneEvt const *)e;
        if ((reqEvt->deviceAddress >= MIN_DEVICE) && (reqEvt->deviceAddress <= MAX_DEVICE)) {
            if (IsBitSet8Set(me->validDevices, reqEvt->deviceAddress)) {
                DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByZoneId(reqEvt->zoneId);
                if (zoneInfo != 0) {
                    /* save requester */
                    me->requester = reqEvt->basic.requester;
                    /* save request parameters */
                    Datamodel_device_set_zone(me, reqEvt->deviceAddress, zoneInfo->super.localZoneId);
                    return Q_TRAN(&Datamodel_idle);
                } else {
                    error = ERROR_ZONE_NOT_FOUND;
                }
            } else {
                error = ERROR_DEVICE_NOT_FOUND;
            }
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DEVICE_SET_BUTTON_ACTIVE_GROUP_SIG: {
        uint8_t error = ERROR_WRONG_PARAMETER;
        DatamodelReqDeviceSetButtonGroupEvt const * reqEvt = (DatamodelReqDeviceSetButtonGroupEvt const *)e;
        if ((reqEvt->deviceAddress >= MIN_DEVICE) && (reqEvt->deviceAddress <= MAX_DEVICE)) {
            if (IsBitSet8Set(me->validDevices, reqEvt->deviceAddress)) {
                if (((reqEvt->groupId >= USER_GROUP) && (reqEvt->groupId <= MAX_GROUP)) ||
                    (reqEvt->groupId == DATAMODEL_GROUP_ID_NOT_APPLICABLE)) {
                    me->requester = reqEvt->basic.requester;
                    Datamodel_device_set_button_group(me, reqEvt->deviceAddress, reqEvt->groupId);
                    return Q_TRAN(&Datamodel_idle);
                }
            } else {
                error = ERROR_DEVICE_NOT_FOUND;
            }
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    case DATAMODEL_REQ_DEVICE_SET_SET_LOCAL_PRIO_SIG: {
        uint8_t error = ERROR_WRONG_PARAMETER;
        DatamodelReqDeviceSetButtonLocalPriorityEvt const * reqEvt = (DatamodelReqDeviceSetButtonLocalPriorityEvt const *)e;
        if ((reqEvt->deviceAddress >= MIN_DEVICE) && (reqEvt->deviceAddress <= MAX_DEVICE)) {
            if (IsBitSet8Set(me->validDevices, reqEvt->deviceAddress)) {
                me->requester = reqEvt->basic.requester;
                Datamodel_device_set_button_set_local_priority(me, reqEvt->deviceAddress, reqEvt->setLocalPriority);
                return Q_TRAN(&Datamodel_idle);
            } else {
                error = ERROR_DEVICE_NOT_FOUND;
            }
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    case DATAMODEL_REQ_DEVICE_SET_NO_COMING_HOME_SIG: {
        uint8_t error = ERROR_WRONG_PARAMETER;
        DatamodelReqDeviceSetButtonNoComingHomeEvt const * reqEvt = (DatamodelReqDeviceSetButtonNoComingHomeEvt const *)e;
        if ((reqEvt->deviceAddress >= MIN_DEVICE) && (reqEvt->deviceAddress <= MAX_DEVICE)) {
            if (IsBitSet8Set(me->validDevices, reqEvt->deviceAddress)) {
                me->requester = reqEvt->basic.requester;
                Datamodel_device_set_button_set_no_coming_home(me, reqEvt->deviceAddress, reqEvt->noComingHome);
                return Q_TRAN(&Datamodel_idle);
            } else {
                error = ERROR_DEVICE_NOT_FOUND;
            }
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DEVICE_SET_BUTTON_APARTMENT_SIG: {
        uint8_t error = ERROR_WRONG_PARAMETER;
        DatamodelReqDeviceSetButtonApartmentEvt const * reqEvt = (DatamodelReqDeviceSetButtonApartmentEvt const *)e;
        if ((reqEvt->deviceAddress >= MIN_DEVICE) && (reqEvt->deviceAddress <= MAX_DEVICE)) {
            if (IsBitSet8Set(me->validDevices, reqEvt->deviceAddress)) {
                me->requester = reqEvt->basic.requester;
                Datamodel_device_set_button_set_apartment(me, reqEvt->deviceAddress, reqEvt->apartment);
                return Q_TRAN(&Datamodel_idle);
            }
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }

    /* ------------------------------------ */
    case DATAMODEL_REQ_DEVICE_SET_BUTTON_OUTPUT_CHANNEL_SIG: {
        uint8_t error = ERROR_WRONG_PARAMETER;
        DatamodelReqDeviceSetButtonOutputChannelEvt const * reqEvt = (DatamodelReqDeviceSetButtonOutputChannelEvt const *)e;
        if ((reqEvt->deviceAddress >= MIN_DEVICE) && (reqEvt->deviceAddress <= MAX_DEVICE)) {
            if (IsBitSet8Set(me->validDevices, reqEvt->deviceAddress)) {
                me->requester = reqEvt->basic.requester;
                Datamodel_device_set_button_set_output_channel(me, reqEvt->deviceAddress, reqEvt->channelId, reqEvt->sendDownstream);
                return Q_TRAN(&Datamodel_idle);
            }
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }

    /* ------------------------------------ */
    case DATAMODEL_REQ_DEVICE_SET_BUTTON_GLOBAL_GROUP_SIG: {
        uint8_t error = ERROR_WRONG_PARAMETER;
        DatamodelReqDeviceSetButtonGlobalGroupEvt const * reqEvt = (DatamodelReqDeviceSetButtonGlobalGroupEvt const *)e;
        if ((reqEvt->deviceAddress >= MIN_DEVICE) && (reqEvt->deviceAddress <= MAX_DEVICE)) {
            if (IsBitSet8Set(me->validDevices, reqEvt->deviceAddress)) {
                me->requester = reqEvt->basic.requester;
                Datamodel_device_set_button_set_global_group(me, reqEvt->deviceAddress, reqEvt->group, reqEvt->sendDownstream);
                return Q_TRAN(&Datamodel_idle);
            }
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }

    /* ------------------------------------ */
    case DATAMODEL_REQ_DEVICE_SET_LTMODE_SIG: {
        uint8_t error = ERROR_WRONG_PARAMETER;
        DatamodelReqDeviceSetLTMODERegEvt const * reqEvt = (DatamodelReqDeviceSetLTMODERegEvt const *)e;
        if ((reqEvt->deviceAddress >= MIN_DEVICE) && (reqEvt->deviceAddress <= MAX_DEVICE)) {
            if (IsBitSet8Set(me->validDevices, reqEvt->deviceAddress)) {
                me->requester = reqEvt->basic.requester;
                Datamodel_device_set_ltmode(me, reqEvt->deviceAddress, reqEvt->LTMODERegister);
                return Q_TRAN(&Datamodel_idle);
            }
        } else {
            error = ERROR_DEVICE_NOT_FOUND;
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DEVICE_SET_OUTPUT_LOAD_SIG: {
        uint8_t error = ERROR_WRONG_PARAMETER;
        DatamodelReqDeviceSetMODERegEvt const * reqEvt = (DatamodelReqDeviceSetMODERegEvt const *)e;
        if ((reqEvt->deviceAddress >= MIN_DEVICE) && (reqEvt->deviceAddress <= MAX_DEVICE)) {
            if (IsBitSet8Set(me->validDevices, reqEvt->deviceAddress)) {
                me->requester = reqEvt->basic.requester;
                Datamodel_device_set_output_load(me, reqEvt->deviceAddress, reqEvt->MODERegister);
                return Q_TRAN(&Datamodel_idle);
            }
        } else {
            error = ERROR_DEVICE_NOT_FOUND;
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DEVICE_SET_BUTTON_ID_AND_GROUP_SIG: {
        uint8_t error = ERROR_WRONG_PARAMETER;
        DatamodelReqDeviceSetButtonIdAndGroupRegEvt const * reqEvt = (DatamodelReqDeviceSetButtonIdAndGroupRegEvt const *)e;
        if ((reqEvt->deviceAddress >= MIN_DEVICE) && (reqEvt->deviceAddress <= MAX_DEVICE)) {
            if (IsBitSet8Set(me->validDevices, reqEvt->deviceAddress)) {
                me->requester = reqEvt->basic.requester;
                Datamodel_device_set_button_id_and_group(me, reqEvt->deviceAddress, reqEvt->buttonIdAndGroupRegister);
                return Q_TRAN(&Datamodel_idle);
            }
        } else {
            error = ERROR_DEVICE_NOT_FOUND;
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DEVICE_SET_LOCKED_SIG: {
        uint8_t error = ERROR_WRONG_PARAMETER;
        DatamodelReqDeviceSetLockedEvt const * reqEvt = (DatamodelReqDeviceSetLockedEvt const *)e;
        if ((reqEvt->deviceAddress >= MIN_DEVICE) && (reqEvt->deviceAddress <= MAX_DEVICE)) {
            if (IsBitSet8Set(me->validDevices, reqEvt->deviceAddress)) {
                me->requester = reqEvt->basic.requester;
                Datamodel_device_set_locked(me, reqEvt->deviceAddress, reqEvt->locked);
                return Q_TRAN(&Datamodel_idle);
            } else {
                error = ERROR_DEVICE_NOT_FOUND;
            }
        }
        /* post error */
        Datamodel_PostError(reqEvt->basic.requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_REMOVE_INACTIVE_DEVICES_SIG: {
        DatamodelReqBasicEvt const * reqEvt = (DatamodelReqBasicEvt const *)e;
        me->requester = reqEvt->requester;
        Datamodel_remove_inactive_device(me);
        return Q_TRAN(&Datamodel_idle);
    }
    case DATAMODEL_REQ_REMOVE_DEVICE_SIG: {
        DatamodelReqRemoveDeviceEvt const * reqEvt = (DatamodelReqRemoveDeviceEvt const *)e;
        uint8_t error;
        me->requester = reqEvt->basic.requester;
        if ((reqEvt->deviceAddress >= MIN_DEVICE) && (reqEvt->deviceAddress <= MAX_DEVICE)) {
            if (IsBitSet8Set(me->validDevices, reqEvt->deviceAddress)) {
                vdc_api_device_remove(reqEvt->deviceAddress);
                Datamodel_remove_device(me, reqEvt->deviceAddress);
                return Q_TRAN(&Datamodel_idle);
            } else {
                error = ERROR_DEVICE_NOT_FOUND;
            }
        } else {
            error = ERROR_WRONG_PARAMETER;
        }
        /* post error */
        Datamodel_PostError(me->requester, error);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DSM_INFO_SIG: {
        DatamodelReqDsmEvt const * reqEvt = (DatamodelReqDsmEvt const *)e;
        me->requester = reqEvt->requester;
        Datamodel_get_dsm_info_name(me, &reqEvt->dsuid);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_FACTORY_RESET_SIG: {
        /* notify shutdown to MMI */
        static const QEvent shutdown = { MMI_SHUTDOWN_SIG, 0 };
        QActive_postFIFO(AO_MMI, &shutdown);
        Datamodel_factory_reset(me);
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_SET_DSM_CONFIG_SIG: {
        sqlite3_stmt *statement;
        DatamodelReqSetDsmConfig const * reqEvt = (DatamodelReqSetDsmConfig const *)e;
        /* save requester */
        me->requester = reqEvt->basic.requester;
        LOCK_ACCESS
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE dsm_config SET dsmid=?, hwRevision=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_blob(statement, 1, &reqEvt->dsmid.id[0], DSUID_SIZE, NULL) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, reqEvt->hwRevision) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        PostFIFO_ErrorOK(me->requester);
        Datamodel_Changed();
        return Q_TRAN(&Datamodel_idle);
    }
    /* ------------------------------------ */
    case DATAMODEL_REQ_DEVICE_PUSH_SENSOR_SIG: {
        const DatamodelReqDeviceSensorPushEvt * reqEvt = (DatamodelReqDeviceSensorPushEvt const *)e;

        dsuid_t sensorDsuid;
        dsuid_t vdcDsuid;
        uint8_t sensorType;
        uint16_t zoneId = 0;
        uint8_t localZoneId = 0;
        bool sensorPrecision = false;

        if ( Datamodel_IsZoneSensor(me, reqEvt->deviceId, reqEvt->sensorIndex, &sensorDsuid, &sensorType, &sensorPrecision) &&
             Datamodel_GetZoneSensorZoneId(me, &sensorDsuid, &zoneId, &localZoneId) &&
             Datamodel_GetVdcDsuidFromDeviceId(reqEvt->deviceId, &vdcDsuid)) {
            /* send sensor push to dsm-api */
            DsmApi_SendZoneGroupSensorPushFromVdc(&vdcDsuid,
                                           &sensorDsuid,
                                           zoneId, 0,
                                           sensorType,
                                           reqEvt->sensorValue,
                                           sensorPrecision);

            ZoneGroupSensorPushEvt *pushEvt;
            /* send sensor push to vdcs */
            pushEvt = Q_NEW(ZoneGroupSensorPushEvt, SENSOR_PUSH_VALUE_SIG);
            pushEvt->zoneId = zoneId;
            pushEvt->localZoneId = localZoneId;
            pushEvt->groupId = 0;
            pushEvt->sensorType = sensorType;
            pushEvt->sensorValue = reqEvt->sensorValue;
            pushEvt->sensorPrecision = sensorPrecision;
            pushEvt->isSourceExternal = TRUE;
            QF_publish((QEvent *)pushEvt);
        }
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState Datamodel_busy(Datamodel *me, QEvent const *e)
{
    if (e->sig >= Q_USER_SIG) {
        /* defer event (request) */
        QActive_defer((QActive *)me, &me->deferredQueue, e);
        return Q_HANDLED();
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static void Datamodel_get_device_dsuid_entry(Datamodel *me, const char* dsuid, uint8_t privateId)
{
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM devices, deviceButtons WHERE devices.deviceId == deviceButtons.deviceId AND devices.dsuid=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 1, dsuid, -1, SQLITE_STATIC) == SQLITE_OK);

    if (sqlite3_step(statement) == SQLITE_ROW) {
        DatamodelResDeviceEntryEvt * resEvt = Q_NEW(DatamodelResDeviceEntryEvt, DATAMODEL_RES_DEVICE_ENTRY_SIG);
        Datamodel_fill_device_from_SQL(&resEvt->device, statement);
        resEvt->state = (IsBitSet8Set(me->activeDevices, resEvt->device.deviceId) ? TRUE : FALSE);
        resEvt->privateId = privateId;
        QActive_postFIFO(me->requester, (QEvent *)resEvt);
    } else {
        Datamodel_PostLongError(me->requester, ERROR_DEVICE_NOT_FOUND, privateId);
        vdsm_report(LOG_DEBUG, "Datamodel get device dsuid entry: device %s not found\n", dsuid);
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
}

/*..........................................................................*/
static bool Datamodel_get_device_by_index_from_vdc(Datamodel *me,  const dsuid_t  *vdcdsuid,  uint8_t modifier, uint16_t deviceIndex, uint16_t *deviceId)
{
    char destAddr [DSUID_STR_LEN];
    dsuid_to_string_upper(vdcdsuid, destAddr);

    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT deviceId FROM devices WHERE vdcdsuid=? ORDER BY deviceId ASC;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 1, destAddr, -1, SQLITE_STATIC) == SQLITE_OK);

    int index = 0;
    uint16_t tempDeviceId = 0;
    do {
        if (sqlite3_step(statement) == SQLITE_ROW) {
            tempDeviceId = sqlite3_column_int(statement, 0);
            vdsm_report (LOG_DEBUG, "Device %d of vdc:%s\n",tempDeviceId, destAddr);
        } else {
            /* no more devices for vdc */
            break;
        }
        switch (modifier) {
            case DEVICE_INFO_BY_INDEX_ONLY_ACTIVE: {
                /* only active devices */
                if (IsBitSet8Set(me->activeDevices, tempDeviceId)) {
                    ++index;
                }
                break;
            }
            case DEVICE_INFO_BY_INDEX_ONLY_INACTIVE: {
                /* only inactive devices */
                if (IsBitSet8Cleared(me->activeDevices, tempDeviceId)) {
                    ++index;
                }
                break;
            }
            default: {
                ++index;
                break;
            }
        }
    } while (index <= deviceIndex);

    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    *deviceId = tempDeviceId;
    return (index > deviceIndex);
}

/*..........................................................................*/
static void Datamodel_get_device_id_entry(Datamodel *me, uint16_t deviceId, uint8_t privateId)
{
    sqlite3_stmt *statement;
    DatamodelResDeviceEntryEvt * resEvt = Q_NEW(DatamodelResDeviceEntryEvt, DATAMODEL_RES_DEVICE_ENTRY_SIG);

    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM devices, deviceButtons WHERE devices.deviceId == deviceButtons.deviceId AND devices.deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    Datamodel_fill_device_from_SQL(&resEvt->device, statement);
    resEvt->state = (IsBitSet8Set(me->activeDevices, deviceId) ? TRUE : FALSE);
    resEvt->privateId = privateId;
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    QActive_postFIFO(me->requester, (QEvent *)resEvt);
}

/*..........................................................................*/
static bool Datamodel_get_device_entry_by_index(Datamodel *me,
                                                uint8_t localZoneId,
                                                uint8_t deviceIndex,
                                                DatamodelDeviceModifier modifier,
                                                const dsuid_t  *vdcdsuid,
                                                DatamodelDevice *output_device)
{
    char destAddr[DSUID_STR_LEN];
    dsuid_to_string_upper(vdcdsuid, destAddr);
    bool found = false;
    int ret;

    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM devices, deviceButtons WHERE devices.deviceId == deviceButtons.deviceId AND devices.localZoneId=? AND devices.vdcdsuid=? ORDER BY devices.deviceId ASC;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 2, destAddr, -1, SQLITE_STATIC) == SQLITE_OK);
    do
    {
        ret = sqlite3_step(statement);
        if (ret == SQLITE_ROW)
        {
            Datamodel_fill_device_from_SQL(output_device, statement);
            if (deviceIndex == 0) {
                found = TRUE;
                break;
            }
            /* decrement device index */
            if ((modifier == ALL) ||
                ((modifier == ONLY_ACTIVE) && (IsBitSet8Set(me->activeDevices, output_device->deviceId))) ||
                ((modifier == ONLY_INACTIVE) && (IsBitSet8Cleared(me->activeDevices, output_device->deviceId)))) {
                --deviceIndex;
            }
        }
    } while (ret == SQLITE_ROW);
    sqlite3_finalize(statement);

    UNLOCK_ACCESS
    return found;
}

/*..........................................................................*/
static uint16_t Datamodel_get_zone_usage(Datamodel *me, uint8_t localZoneId)
{
    uint16_t countLocalZone = 0;
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT COUNT (*) FROM devices WHERE localZoneId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId) == SQLITE_OK);

    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    countLocalZone = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    return countLocalZone;
}

/*..........................................................................*/
static void Datamodel_publish_zone_added(uint8_t localZoneId, uint16_t zoneId)
{
    DatamodelZoneIdEvt * zoneIdEvt = Q_NEW(DatamodelZoneIdEvt, DATAMODEL_ZONE_ID_SIG);
    zoneIdEvt->zoneId = zoneId;
    zoneIdEvt->localZoneId = localZoneId;
    QF_publish((QEvent *)zoneIdEvt);
}

/*..........................................................................*/
static void Datamodel_add_zone_to_db(Datamodel *me, uint8_t localZoneId, uint16_t zoneId)
{
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT INTO zones (localZoneId, zoneId) VALUES (?, ?);", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, zoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_remove_zone_from_db(Datamodel *me, uint8_t localZoneId)
{
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "DELETE FROM zones WHERE localZoneId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "DELETE FROM groups WHERE g_id>=? AND g_id<=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId << 8) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, (localZoneId << 8) | 0xFF) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "DELETE FROM scenes WHERE s_id>=? AND s_id<=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId << 16) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, (localZoneId << 16) | 0xFFFF) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "DELETE FROM zoneSensors WHERE localZoneId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    /* update cache */
    uint8_t groupId;
    for (groupId = 0; groupId < USER_GROUP; groupId++) {
        Datamodel_UpdateCachedStopTimeout(me, localZoneId, groupId, 0);
    }

    PostFIFO_ErrorOK(me->requester);
    Datamodel_Changed();
}

/*..........................................................................*/
static uint16_t Datamodel_get_zone_device_count(Datamodel *me, uint8_t localZoneId, DatamodelDeviceModifier deviceModifier, const dsuid_t *dsuid)
{
    sqlite3_stmt *statement;
    uint16_t deviceCount = 0;
    char vdc_dsuid_str[DSUID_STR_LEN];
    dsuid_to_string_upper(dsuid, vdc_dsuid_str);

    LOCK_ACCESS
    if (deviceModifier == ALL) {
        if (localZoneId > 0) {
            SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT COUNT (*) FROM devices WHERE localZoneId=? AND vdcdsuid=?;", -1, &statement, 0) == SQLITE_OK);
            SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId) == SQLITE_OK);
            SQL_ASSERT(sqlite3_bind_text(statement, 2, vdc_dsuid_str, -1, SQLITE_STATIC) == SQLITE_OK);
        } else {
            SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT COUNT (*) FROM devices WHERE vdcdsuid=?;", -1, &statement, 0) == SQLITE_OK);
            SQL_ASSERT(sqlite3_bind_text(statement, 1, vdc_dsuid_str, -1, SQLITE_STATIC) == SQLITE_OK);
        }
    } else {
        if (localZoneId > 0) {
            SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT COUNT (*) FROM devices WHERE active=? AND localZoneId=? AND vdcdsuid=?;", -1, &statement, 0) == SQLITE_OK);
            SQL_ASSERT(sqlite3_bind_int(statement, 2, localZoneId) == SQLITE_OK);
            SQL_ASSERT(sqlite3_bind_text(statement, 3, vdc_dsuid_str, -1, SQLITE_STATIC) == SQLITE_OK);

        } else {
            SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT COUNT (*) FROM devices WHERE active=? AND vdcdsuid=?;", -1, &statement, 0) == SQLITE_OK);
            SQL_ASSERT(sqlite3_bind_text(statement, 2, vdc_dsuid_str, -1, SQLITE_STATIC) == SQLITE_OK);
        }
        if (deviceModifier == ONLY_ACTIVE) {
            SQL_ASSERT(sqlite3_bind_int(statement, 1, 1) == SQLITE_OK);
        } else if (deviceModifier == ONLY_INACTIVE) {
            SQL_ASSERT(sqlite3_bind_int(statement, 1, 0) == SQLITE_OK);
        }
    }

    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    deviceCount = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    return deviceCount;
}

/*..........................................................................*/
static void Datamodel_group_membership_modify(Datamodel *me, uint16_t deviceId, uint8_t groupId, uint8_t modifier, bool sendDownstream)
{
    DatamodelDevice device;
    sqlite3_stmt *statement;
    vdsm_report(LOG_DEBUG, "Datamodel modify group membership for device %d, group %d\n", deviceId, groupId);
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM devices, deviceButtons WHERE devices.deviceId == deviceButtons.deviceId AND devices.deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    if (sqlite3_step(statement) == SQLITE_ROW) {
        Datamodel_fill_device_from_SQL(&device, statement);
    } else {
        sqlite3_finalize(statement);
        Datamodel_PostError(me->requester, ERROR_DEVICE_NOT_FOUND);
        UNLOCK_ACCESS
        return;
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    if (!(Fid_IsJocker(device.fid) ||
          (groupId >= USER_GROUP) ||
          Fid_HasVariableStandardGroup(device.fid))) {
        Datamodel_PostError(me->requester, ERROR_WRONG_PARAMETER);
        return;
    }
    if (modifier == DEVICE_GROUP_MEMBERSHIP_MODIFY_ADD) {
        /* add group */
        /* check if group exists in zone */
        DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByLocalId(device.localZoneId);
        Q_ASSERT(zoneInfo != 0);
        if (IsBitSet8Cleared(zoneInfo->groupMembership, groupId)) {
            /* group does not exist */
            Datamodel_PostError(me->requester, ERROR_GROUP_NOT_FOUND);
            return;
        }
        if (IsBitSet8Cleared(device.groupMembership, groupId)) {
            /* add group to device */
            if (sendDownstream) {
                /* post add group event for specific device */
                AddDeviceToGroupEvt * addDeviceToGroupEvt = Q_NEW(AddDeviceToGroupEvt, ADD_DEVICE_TO_GROUP_SIG);
                addDeviceToGroupEvt->groupId = groupId;
                addDeviceToGroupEvt->deviceId = device.deviceId;
                QActive_postFIFO(AO_SelectA, (QEvent *)addDeviceToGroupEvt);
            }
            if (Fid_HasVariableStandardGroup(device.fid) &&
                (groupId <= LAST_STANDARD_GROUP || groupId == ROOM_TEMPERATURE_CONTROL_GROUP) &&
                 sendDownstream) {
                /* if the heating group changes, remove the device from all
                 * other standard and user groups
                 */
                unsigned int i;
                for (i = 1; i <= MAX_GROUPS; i++) {
                    if (IsBitSet8Set(device.groupMembership, i)) {
                        ClearBitSet8(device.groupMembership, i);
                        /* post add group event for specific device */
                        RemoveDeviceFromGroupEvt * removeDeviceFromGroupEvt = Q_NEW(RemoveDeviceFromGroupEvt, REM_DEVICE_FROM_GROUP_SIG);
                        removeDeviceFromGroupEvt->groupId = i;
                        removeDeviceFromGroupEvt->deviceId = device.deviceId;
                        QActive_postFIFO(AO_SelectA, (QEvent *)removeDeviceFromGroupEvt);
                    }
                    /* skip unused user groups */
                    if (i == 23) {
                        i = 47;
                    }
                }
            }
            SetBitSet8(device.groupMembership, groupId);

            {
                LOCK_ACCESS
                SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE devices SET groupMembership=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
                SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
                SQL_ASSERT(sqlite3_bind_int64(statement, 1, groups_array_to_int(device.groupMembership)) == SQLITE_OK);
                SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
                sqlite3_finalize(statement);
                UNLOCK_ACCESS
            }
            /* notify change */
            Datamodel_Changed();
            Datamodel_NotifyChange();
        }
        /* device is already in group */
        /* release possible waiting sync event */
        DsmApi_SendDsmDoneSetDeviceEvent(EVENT_DSM_DONE_ADD_DEVICE_TO_GROUP, ERROR_OK, deviceId);
    } else {
        /* remove group */
        if (JOKER_GROUP == groupId || BROADCAST_GROUP == groupId) {
            /* Joker and broadcast groups can not be deleted, even on JOKER devices. */
            Datamodel_PostError(me->requester, ERROR_GROUP_CAN_NOT_BE_DELETED);
            return;
        }
        if (IsBitSet8Set(device.groupMembership, groupId)) {
            /* remove group from device */
            ClearBitSet8(device.groupMembership, groupId);
            if (sendDownstream) {
                /* post remove group event for specific device */
                RemoveDeviceFromGroupEvt * remDeviceFromGroupEvt = Q_NEW(RemoveDeviceFromGroupEvt, REM_DEVICE_FROM_GROUP_SIG);
                remDeviceFromGroupEvt->groupId = groupId;
                remDeviceFromGroupEvt->deviceId = device.deviceId;
                QActive_postFIFO(AO_SelectA, (QEvent *)remDeviceFromGroupEvt);
            }
            {
                LOCK_ACCESS
                SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE devices SET groupMembership=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
                SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
                SQL_ASSERT(sqlite3_bind_int64(statement, 1, groups_array_to_int(device.groupMembership)) == SQLITE_OK);
                SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
                sqlite3_finalize(statement);
                UNLOCK_ACCESS
            }
            /* notify change */
            Datamodel_Changed();
            Datamodel_NotifyChange();
        }
        /* device is already not in group */
    }
    /* send response */
    PostFIFO_ErrorOK(me->requester);
}

/*..........................................................................*/
static void Datamodel_get_dsm_info_name(Datamodel *me,  const dsuid_t *dsmid)
{
    char destAddr [DSUID_STR_LEN];
    dsuid_to_string_upper(dsmid, destAddr);

    DatamodelResDsmInfoEvt *resEvt = Q_NEW(DatamodelResDsmInfoEvt, DATAMODEL_RES_DSM_INFO_SIG);

    if (Datamodel_dsm_entry_exists(me, destAddr)) {
        sqlite3_stmt *statement;
        LOCK_ACCESS
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT name FROM dsm_info WHERE dsuid=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement, 1, (const char *)destAddr, -1, SQLITE_STATIC) == SQLITE_OK);
        if (sqlite3_step(statement) != SQLITE_ROW) {
            sqlite3_finalize(statement);
            UNLOCK_ACCESS
            /* name not set */
            memset (resEvt->name, 0, sizeof(resEvt->name));
        } else {
            const char *name = sqlite3_column_text(statement, 0);
            if (name != NULL) {
                memcpy(resEvt->name, name, MAX_NAME_LEN);
            } else {
                memset(resEvt->name, 0, MAX_NAME_LEN);
            }
            sqlite3_finalize(statement);
            UNLOCK_ACCESS
        }
    } else {
        vdsm_report(LOG_ERR, "get dsm info name. Element with dsid: %s does not exist\n", destAddr);

        /* name is empty */
        memset (resEvt->name, 0, sizeof(resEvt->name));
        Datamodel_Changed();
    }

    resEvt->hwVersion = me->hwRevision;
#ifdef GIT_VERSION
    resEvt->armSwVersion = GIT_VERSION;
#else
    resEvt->armSwVersion = ((VERSION_MAJOR << 24) | (VERSION_MINOR << 16) | (VERSION_REVISION << 8) | VERSION_BUILD);
#endif
    resEvt->dsuid = *dsmid;
    QActive_postFIFO(me->requester, (QEvent *)resEvt);
}

/*..........................................................................*/
static void Datamodel_get_group_info(Datamodel *me, uint8_t localZoneId, uint8_t groupId, const dsuid_t *vdc_dsuid)
{
    DatamodelResZoneGroupInfoEvt *resEvt = Q_NEW(DatamodelResZoneGroupInfoEvt, DATAMODEL_RES_ZONE_GROUP_INFO_SIG);
    resEvt->groupId = groupId;
    resEvt->countDevices = 0;

    vdsm_report(LOG_DEBUG, "Datamodel_get_group_info: localZoneId: %d, groupId: %d\n", localZoneId, groupId);

    if (Datamodel_IsCluster(groupId)) {
        DatamodelCluster cluster;
        Datamodel_InitCluster(&cluster, groupId);
        if (ERROR_OK == Datamodel_getCluster(groupId, &cluster)) {
            resEvt->groupId = cluster.id;
            resEvt->stateMachineGroupId = cluster.stateMachineGroupId;
            resEvt->countDevices = Datamodel_getClusterDeviceCount(vdc_dsuid, groupId);
            memcpy(resEvt->name, cluster.name, sizeof(resEvt->name));
            resEvt->stopTimeout = cluster.stopTimeout;
            resEvt->remoteStopTimeout = cluster.remoteStopTimeout;
        }
    } else {
        LOCK_ACCESS
        sqlite3_stmt *statement;
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM groups WHERE g_id=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, (localZoneId << 8) | groupId) == SQLITE_OK);
        if (sqlite3_step(statement) != SQLITE_ROW) {
            if (groupId < USER_GROUP) {
                resEvt->stateMachineGroupId = groupId;
            } else {
                resEvt->stateMachineGroupId = 0;
            }
            memset(resEvt->name, 0, sizeof(resEvt->name));
            resEvt->stopTimeout = 0;
            if (resEvt->stateMachineGroupId == SHADE_GROUP) {
                resEvt->remoteStopTimeout = 120;
            } else {
                resEvt->remoteStopTimeout = 0;
            }
        } else {
            const char* name;
            resEvt->stateMachineGroupId = sqlite3_column_int(statement, 1);
            resEvt->stopTimeout = sqlite3_column_int(statement, 2);
            resEvt->remoteStopTimeout = sqlite3_column_int(statement, 3);
            name = sqlite3_column_text(statement, 4);
            if (name != NULL) {
                memcpy(resEvt->name, name, sizeof(resEvt->name));
            }
        }
        sqlite3_finalize(statement);
        UNLOCK_ACCESS

        resEvt->countDevices = Datamodel_getGroupDeviceCount(vdc_dsuid, groupId, localZoneId);
    }
    QActive_postFIFO(me->requester, (QEvent *)resEvt);
}

/*..........................................................................*/
static void Datamodel_get_scene_info(Datamodel *me, uint8_t localZoneId, uint8_t groupId, uint8_t sceneId)
{
    DatamodelResSceneInfoEvt * resEvt = Q_NEW(DatamodelResSceneInfoEvt, DATAMODEL_RES_SCENE_INFO_SIG);
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT name FROM scenes WHERE s_id=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, (localZoneId << 16) | (groupId << 8) | sceneId) == SQLITE_OK);
    if (sqlite3_step(statement) == SQLITE_ROW) {
        const char* name = sqlite3_column_text(statement, 0);
        if (name != NULL) {
            memcpy(resEvt->name, name, MAX_NAME_LEN);
        }
    } else {
        memset(resEvt->name, 0, MAX_NAME_LEN);
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    QActive_postFIFO(me->requester, (QEvent *)resEvt);
}

/*..........................................................................*/
static bool Datamodel_dsm_entry_exists(Datamodel *me, const char *dsmId)
{
    sqlite3_stmt *statement;
    /* Check if available */
    int ret = 0;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM dsm_info WHERE dsuid=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 1, dsmId, -1, SQLITE_STATIC) == SQLITE_OK);
    ret = sqlite3_step(statement);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return (ret == SQLITE_ROW);
}

/*..........................................................................*/
static void Datamodel_set_dsm_apartment_id(Datamodel *me, uint8_t apartmentId)
{
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE dsm_info SET apartmentId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, apartmentId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    /* update cached upstream frequency */
    me->dsmProperties.apartmentId = apartmentId;

    /* send response */
    PostFIFO_ErrorOK(me->requester);
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_set_dsm_temperature_limits(Datamodel *me, uint8_t warnTemperature, uint8_t errorTemperature)
{
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE dsm_info SET warnTemperature=?, errorTemperature=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, warnTemperature) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, errorTemperature) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    /* update cached temperature limits */
    me->dsmProperties.warnTemperature = warnTemperature;
    me->dsmProperties.errorTemperature = errorTemperature;

    /* send response */
    PostFIFO_ErrorOK(me->requester);
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_set_dsm_ping_intervals(Datamodel *me, uint16_t autoPingInterval, uint16_t oneTimePingInterval)
{
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE dsm_info SET autoPingInterval=?, oneTimePingInterval=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, autoPingInterval) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, oneTimePingInterval) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    /* update cached ping intervals */
    me->dsmProperties.autoPingInterval = autoPingInterval;
    me->dsmProperties.oneTimePingInterval = oneTimePingInterval;
    /* send response */
    PostFIFO_ErrorOK(me->requester);
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_set_dsm_flags(Datamodel *me, uint8_t flags)
{
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE dsm_info SET flags=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement,  1, flags) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    /* update cached flags */
    me->dsmProperties.flags = flags;
    /* send response */
    PostFIFO_ErrorOK(me->requester);
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_set_dsm_upstream_frequency(Datamodel *me, uint8_t upstreamFrequency)
{
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE dsm_info SET upstreamFrequency=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement,  1, upstreamFrequency) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    /* update cached upstream frequency */
    me->dsmProperties.upstreamFrequency = upstreamFrequency;
    /* send response */
    PostFIFO_ErrorOK(me->requester);
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_set_dsm_name(Datamodel *me, const uint8_t *name, const dsuid_t *dsmid)
{
    char destAddr[DSUID_STR_LEN];
    dsuid_to_string_upper(dsmid, destAddr);

    if (Datamodel_dsm_entry_exists(me, destAddr)) {
        /* Update entry */
        sqlite3_stmt *statement;
        LOCK_ACCESS
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE dsm_info SET name=?  WHERE dsuid=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement, 1, (const char *)name,     -1, SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement, 2, (const char *)destAddr, -1, SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        /* Update name on vdc */
        vdc_api_vdc_set_name(dsmid, (const char*)name);
    } else {
        vdsm_report(LOG_ERR, "set dsm name. Element with dsid: %s does not exist\n", destAddr);
    }

    /* send response */
    PostFIFO_ErrorOK(me->requester);
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_get_dsm_properties(Datamodel *me, const dsuid_t *dsmid)
{
    char destAddr[DSUID_STR_LEN];
    dsuid_to_string_upper(dsmid, destAddr);
    DatamodelResDsmProperties *props = &(me->dsmProperties);

    /* set default values */
    props->apartmentId = DEFAULT_APARTMENT_ID;
    props->autoPingInterval = DEFAULT_AUTO_PING_INTERVAL;
    props->currentTemperature = 0;
    props->errorTemperature = DEFAULT_ERROR_TEMPERATURE;
    props->flags = DEFAULT_DSM_INFO_FLAGS;
    props->maxTemperature = 0;
    props->oneTimePingInterval = DEFAULT_ONE_TIME_PING_INTERVAL;
    props->upstreamFrequency = DEFAULT_UPSTREAM_FREQUNECY;
    props->warnTemperature = DEFAULT_WARN_TEMPERATURE;

    if (Datamodel_dsm_entry_exists(me, destAddr)) {
        sqlite3_stmt *statement;
        LOCK_ACCESS
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT apartmentId, flags, autoPingInterval, \
                                               oneTimePingInterval, warnTemperature, errorTemperature, \
                                               upstreamFrequency FROM dsm_info where dsuid=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement, 1, (const char *)destAddr, -1, SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);

        props->apartmentId         = sqlite3_column_int(statement, 0);
        props->flags               = sqlite3_column_int(statement, 1);
        props->autoPingInterval    = sqlite3_column_int(statement, 2);
        props->oneTimePingInterval = sqlite3_column_int(statement, 3);
        props->warnTemperature     = sqlite3_column_int(statement, 4);
        props->errorTemperature    = sqlite3_column_int(statement, 5);
        props->upstreamFrequency   = sqlite3_column_int(statement, 6);

        sqlite3_finalize(statement);
        UNLOCK_ACCESS
    } else {
        vdsm_report(LOG_ERR, "get dsm properties. Element with dsid: %s does not exist\n", destAddr);
    }
    QActive_postFIFO(me->requester, (QEvent *)props);
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_set_device_name(Datamodel *me, uint16_t deviceId, const char *name)
{
    sqlite3_stmt *statement;
    vdsm_report(LOG_DEBUG, "Datamodel set device %d name %s\n", deviceId, name);
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE devices SET name=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 1, name, -1, SQLITE_STATIC) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_set_group_name(Datamodel *me, uint8_t localZoneId, uint8_t groupId, const uint8_t *name)
{
    sqlite3_stmt *statement;
    int ret;

    vdsm_report(LOG_DEBUG, "Datamodel_set_group_name: localZoneId: %d, groupId: %d, name: %s\n", localZoneId, groupId, name);
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT name FROM groups WHERE g_id=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, (localZoneId << 8) | groupId) == SQLITE_OK);
    ret = sqlite3_step(statement);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    if (ret == SQLITE_ROW) {
        /* update name */
        sqlite3_stmt *statement;
        LOCK_ACCESS
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE groups SET name=? WHERE g_id=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement, 1, (const char *)name, min(strlen(name), 20), SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, (localZoneId << 8) | groupId) == SQLITE_OK);
        ret = sqlite3_step(statement);
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        if (ret != SQLITE_DONE) {
            /* could not update */
            vdsm_report(LOG_ERR, "Datamodel_set_group_name: error updating group\n");
            Datamodel_PostError(me->requester, ERROR_REQUEST_CAN_NOT_BE_EXECUTED);
            return;
        }
    } else {
        /* group does not exist yet */
        if (groupId < USER_GROUP) {
            Datamodel_InitGroup(me, localZoneId, groupId, 1, name);
        } else {
            Datamodel_InitGroup(me, localZoneId, groupId, 0, name);
        }
    }

    /* send response */
    PostFIFO_ErrorOK(me->requester);
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_set_group_stop_timeout(Datamodel *me, uint8_t localZoneId, uint8_t groupId, uint8_t stopTimeout)
{
    sqlite3_stmt *statement;
    int ret;

    vdsm_report(LOG_DEBUG, "Datamodel_set_stop_timeout: localZoneId: %d, groupId: %d stopTimeout: %d\n", localZoneId, groupId, stopTimeout);
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT stopTimeout FROM groups WHERE g_id=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, (localZoneId << 8) | groupId) == SQLITE_OK);
    ret = sqlite3_step(statement);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    if (ret != SQLITE_ROW) {
        /* group doesn't exist */
        if (groupId < USER_GROUP) {
            Datamodel_InitGroup(me, localZoneId, groupId, groupId, NULL);
        } else if ((groupId < FREE_USER_GROUP) || (groupId == ROOM_TEMPERATURE_CONTROL_GROUP)) {
            Datamodel_InitGroup(me, localZoneId, groupId, 0, NULL);
        } else {
            Datamodel_PostError(me->requester, ERROR_GROUP_NOT_FOUND);
            return;
        }
    }
    /* update stop timeout */
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE groups SET stopTimeout=?, remoteStopTimeout=? WHERE g_id=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, stopTimeout) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 3, (localZoneId << 8) | groupId) == SQLITE_OK);
    ret = sqlite3_step(statement);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    if (ret != SQLITE_DONE) {
        /* could not update */
        vdsm_report(LOG_ERR, "Datamodel_set_group_stop_timeout: error updating group\n");
        Datamodel_PostError(me->requester, ERROR_REQUEST_CAN_NOT_BE_EXECUTED);
        return;
    }
    /* update cache */
    Datamodel_UpdateCachedStopTimeout(me, localZoneId, groupId, stopTimeout);
    /* send response */
    PostFIFO_ErrorOK(me->requester);
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_set_group_color(Datamodel *me, uint8_t localZoneId, uint8_t groupId, uint8_t color)
{
    sqlite3_stmt *statement;
    uint8_t remoteStopTimeout = 0;
    int ret;

    vdsm_report(LOG_DEBUG, "Datamodel_set_group_color: localZoneId: %d, groupId: %d color: %d\n", localZoneId, groupId, color);
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT stateMachineGroupId, remoteStopTimeout FROM groups WHERE g_id=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, (localZoneId << 8) | groupId) == SQLITE_OK);
    ret = sqlite3_step(statement);
    if (ret == SQLITE_ROW) {
        /* group already exists */
        uint8_t stateMachineGroupId = sqlite3_column_int(statement, 0);
        remoteStopTimeout = sqlite3_column_int(statement, 1);
        sqlite3_finalize(statement);
        UNLOCK_ACCESS

        if (stateMachineGroupId == color) {
            /* state machine color is already set */
            PostFIFO_ErrorOK(me->requester);
            return;
        } else {
            /* update group */
            sqlite3_stmt *statement;
            remoteStopTimeout = (color == SHADE_GROUP)? 120 : 0;
            LOCK_ACCESS
            SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE groups SET stateMachineGroupId=?, stopTimeout=?, remoteStopTimeout=? WHERE g_id=?;", -1, &statement, 0) == SQLITE_OK);
            SQL_ASSERT(sqlite3_bind_int(statement, 1, color) == SQLITE_OK);
            SQL_ASSERT(sqlite3_bind_int(statement, 2, 0) == SQLITE_OK);
            SQL_ASSERT(sqlite3_bind_int(statement, 3, remoteStopTimeout) == SQLITE_OK);
            SQL_ASSERT(sqlite3_bind_int(statement, 4, (localZoneId << 8) | groupId) == SQLITE_OK);
            ret = sqlite3_step(statement);
            sqlite3_finalize(statement);
            UNLOCK_ACCESS
            if (ret != SQLITE_DONE) {
                /* could not update */
                vdsm_report(LOG_ERR, "Datamodel_set_group_color: error updating group\n");
                Datamodel_PostError(me->requester, ERROR_REQUEST_CAN_NOT_BE_EXECUTED);
                return;
            }
        }
    } else {
        /* initialize group */
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        Datamodel_InitGroup(me, localZoneId, groupId, color, NULL);
        remoteStopTimeout = (groupId == SHADE_GROUP) ? 120 : 0;
    }
    /* update cache */
    Datamodel_UpdateCachedStopTimeout(me, localZoneId, groupId, remoteStopTimeout);
    /* notify selectA statemachine */
    Datamodel_NotifyUserGroup(groupId, color);
    Datamodel_NotifyChange();
    /* send response */
    PostFIFO_ErrorOK(me->requester);
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_add_device(Datamodel *me, DatamodelDevice *device)
{
    DatamodelZoneInfo* zoneInfo;
    uint16_t zoneId, deviceId;
    deviceId = device->deviceId;
    SetBitSet8(me->validDevices, deviceId);
    SetBitSet8(me->activeDevices, deviceId);
    /* get zone ID */
    zoneInfo = Datamodel_GetZoneInfoByLocalId(device->localZoneId);
    Q_ASSERT(zoneInfo != 0);
    zoneId = zoneInfo->super.zoneId;
    {
        sqlite3_stmt *statement;
        LOCK_ACCESS
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT INTO devices VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?);", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,   1, device->deviceId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement,  2, device->dsuid, -1, SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,   3, device->vid) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,   4, device->fid) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,   5, device->ver) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,   6, device->locked) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,   7, device->localZoneId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int64(statement, 8, groups_array_to_int(device->groupMembership)) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,   9, device->ltmode) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,  10, device->mode) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement, 11, (const char *)device->name, -1, SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,  12, device->active) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement, 13, device->vdcdsuid, -1, SQLITE_STATIC) == SQLITE_OK);

        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
        // button info
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT INTO deviceButtons VALUES(?,?,?,?,?,?,?,?);", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, device->deviceId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, device->buttonIdAndGroup) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 3, device->buttonUserGroup) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 4, device->buttonSetLocalPriority) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 5, device->buttonNoComingHomeCall) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 6, device->buttonApartment) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 7, device->buttonOutputChannel) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 8, device->buttonGlobalGroup) == SQLITE_OK);

        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
    }
    /* update active group stop timeout */
    LOCK_ACCESS
    Datamodel_SetGroupActiveStopTimeout(me, device);
    UNLOCK_ACCESS
    /* send device accessibility */
    dsuid_t dsuid;
    dsuid_from_string(device->dsuid, &dsuid);
    DsmApi_SendDeviceAccessibilityEvent(deviceId, zoneId, device->vid, 0, device->fid, device->ver, dsuid, TRUE);
    /* send response */
    DatamodelResAddDeviceEvt * resEvt = Q_NEW(DatamodelResAddDeviceEvt, DATAMODEL_RES_ADD_DEVICE_SIG);
    resEvt->deviceId = deviceId;
    resEvt->localZoneId = device->localZoneId;
    resEvt->zoneId = zoneId;
    QActive_postFIFO(me->requester, (QEvent *)resEvt);
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_set_device_active(Datamodel *me, uint16_t deviceId, bool active)
{
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE devices SET active=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, active) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_ping_device_state(Datamodel *me, uint16_t deviceId, uint16_t nextDeviceId, bool active, bool oneTimePingDone)
{
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT dsuid FROM devices WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, nextDeviceId) == SQLITE_OK);
    if (sqlite3_step(statement) == SQLITE_ROW) {
        /* allocate ping device response */
        DatamodelResPingDeviceEvt * resEvt = Q_NEW(DatamodelResPingDeviceEvt, DATAMODEL_RES_PING_DEVICE_SIG);
        resEvt->nextDeviceId = nextDeviceId;
        resEvt->oneTimePingDone = oneTimePingDone;
        QActive_postFIFO(me->requester, (QEvent *)resEvt);
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    if (IsBitSet8Set(me->validDevices, deviceId)) {
        /* update device accessibility */
        if (active) {
            /* device is now active */
            if (IsBitSet8Cleared(me->activeDevices, deviceId)) {
                /* change device to active */
                SetBitSet8(me->activeDevices, deviceId);
                Datamodel_set_device_active(me, deviceId, TRUE);
                /* send device accessibility change */
                Datamodel_send_device_accessibility(me, deviceId, TRUE);
            }
        } else {
            /* device is now inactive */
            if (IsBitSet8Set(me->activeDevices, deviceId)) {
                /* change device to inactive */
                ClearBitSet8(me->activeDevices, deviceId);
                Datamodel_set_device_active(me, deviceId, FALSE);
                /* send device accessibility change */
                Datamodel_send_device_accessibility(me, deviceId, FALSE);
            }
        }
    }
}

/*..........................................................................*/
static void Datamodel_zone_modify_zoneId(Datamodel *me, uint8_t localZoneId, uint16_t zoneId)
{
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE zones SET zoneId=? WHERE localZoneId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, zoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, localZoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    PostFIFO_ErrorOK(me->requester);

    Datamodel_Changed();
    /* notify change */
    Datamodel_NotifyChange();
}

/*..........................................................................*/
static bool Datamodel_local_zone_id_exists(Datamodel *me, uint8_t localZoneId)
{
    sqlite3_stmt *statement;
    int ret = 0;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM zones WHERE localZoneId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId) == SQLITE_OK);
    ret = sqlite3_step(statement);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return (ret == SQLITE_ROW);
}

/*..........................................................................*/
static void Datamodel_zone_modify_name(Datamodel *me, uint8_t localZoneId, const uint8_t *name,  dsuid_t const *vdcdsuid)
{
    if (Datamodel_local_zone_id_exists(me, localZoneId)) {
        sqlite3_stmt *statement;
        LOCK_ACCESS
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE zones SET name=? WHERE localZoneId=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement, 1, (const char *)name, -1, SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, localZoneId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
    } else {
        uint16_t zoneId = Datamodel_create_zoneId_from_dsuid(vdcdsuid);
        sqlite3_stmt *statement;
        LOCK_ACCESS
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT INTO zones (name, localZoneId, zoneId) VALUES (?, ?, ?);", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement, 1, (const char *)name, -1, SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,  2, localZoneId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,  3, zoneId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
    }
    PostFIFO_ErrorOK(me->requester);
    Datamodel_Changed();
    /* notify change */
    Datamodel_NotifyChange();
}

/*..........................................................................*/
static void Datamodel_zone_group_modify_remove(Datamodel *me, uint8_t localZoneId, uint8_t groupId)
{
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "DELETE FROM groups WHERE g_id=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId << 8 | groupId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "DELETE FROM scenes WHERE s_id>=? AND s_id<=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, (localZoneId << 16) | (groupId << 8)) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, (localZoneId << 16) | (groupId << 8) | 0xFF) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    Datamodel_Changed();
    /* notify user group */
    Datamodel_NotifyUserGroup(groupId, 0);

    {
        sqlite3_stmt *statement;
        int ret;
        LOCK_ACCESS
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT (groupMembership,deviceId,dsuid) FROM devices WHERE localZoneId=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId) == SQLITE_OK);

        do {
            ret = sqlite3_step(statement);
            if (ret == SQLITE_ROW) {
                uint8_t groupMembership[8];
                uint16_t deviceId = sqlite3_column_int(statement, 1);
                groups_int_to_array(sqlite3_column_int64(statement, 0), groupMembership);
                if (IsBitSet8Set(groupMembership, groupId)) {
                    ClearBitSet8(groupMembership, groupId);
                    sqlite3_stmt *statement2;
                    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE devices SET groupMembership=? WHERE deviceId=?;", -1, &statement2, 0) == SQLITE_OK);
                    SQL_ASSERT(sqlite3_bind_int64(statement2, 1, groups_array_to_int(groupMembership)) == SQLITE_OK);
                    SQL_ASSERT(sqlite3_bind_int(statement2, 2, deviceId) == SQLITE_OK);
                    SQL_ASSERT(sqlite3_step(statement2) == SQLITE_DONE);
                    sqlite3_finalize(statement2);
                    /* notify vdc device */
                    device_api_remove_group(sqlite3_column_text(statement, 2), groupId);
                }
            }
        } while (ret == SQLITE_ROW);
        sqlite3_finalize(statement);
        /* update cache */
        Datamodel_ClearGroupActiveStopTimeout(me, localZoneId, groupId);
        UNLOCK_ACCESS
    }
    /* notify change */
    Datamodel_NotifyChange();
    /* send response */
    PostFIFO_ErrorOK(me->requester);
}

/*..........................................................................*/
static void Datamodel_device_set_zone(Datamodel *me, uint16_t deviceId, uint8_t localZoneId)
{
    sqlite3_stmt *statement;
    DatamodelDevice device;
    uint8_t error = ERROR_OK;

    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM devices, deviceButtons WHERE devices.deviceId == deviceButtons.deviceId AND devices.deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    if (sqlite3_step(statement) != SQLITE_ROW) {
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        /* send error response */
        Datamodel_PostError(me->requester, ERROR_DEVICE_NOT_FOUND);
        return;
    }
    Datamodel_fill_device_from_SQL(&device, statement);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    if (device.localZoneId != localZoneId) {
        /* remove device from current zone */
        DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByLocalId(device.localZoneId);
        /* add device to new zone */
        zoneInfo = Datamodel_GetZoneInfoByLocalId(localZoneId);
        device.localZoneId = localZoneId;
        /* change zone on device */
        if (!device_api_set_zone(device.dsuid, zoneInfo->super.zoneId)) {
            error = ERROR_DEVICE_SELECT_RESPONSE_TIMEOUT;
        }
        {
            LOCK_ACCESS
            SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE devices SET localZoneId=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
            SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId) == SQLITE_OK);
            SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
            SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
            sqlite3_finalize(statement);
            UNLOCK_ACCESS
        }
        Datamodel_ResetGroupStopTimeoutActive(me);
    }

    PostFIFO_ErrorOK(me->requester);
    DsmApi_SendDsmDoneSetDeviceEvent(EVENT_DSM_DONE_SET_DEVICE_ZONE, error, deviceId);
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_device_set_button_group(Datamodel *me, uint16_t deviceId, uint8_t groupId)
{
    sqlite3_stmt *statement;
    DatamodelDevice device;

    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM devices, deviceButtons WHERE devices.deviceId == deviceButtons.deviceId AND devices.deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    if (sqlite3_step(statement) != SQLITE_ROW) {
        sqlite3_finalize(statement);
        /* send error response */
        Datamodel_PostError(me->requester, ERROR_DEVICE_NOT_FOUND);
        UNLOCK_ACCESS
        return;
    }
    Datamodel_fill_device_from_SQL(&device, statement);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    if (device.buttonUserGroup == groupId) {
        /* no change -> send OK response */
        PostFIFO_ErrorOK(me->requester);
        return;
    }
    /* is button action group a standard group? */
    if ((groupId < USER_GROUP) && (!Fid_IsJocker(device.fid))) {
        /* action group must be the standard group defined by the function id (FID) */
        if (!Fid_IsGroupMatching(device.fid, groupId)) {
            /* invalid standard group -> send error response */
            Datamodel_PostError(me->requester, ERROR_WRONG_PARAMETER);
            return;
        }
    }
    /* update button group */
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE deviceButtons SET buttonUserGroup=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, groupId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    bool btnAppartment = (groupId >= FIRST_CLUSTER) && (groupId <= LAST_CLUSTER);
    Datamodel_device_set_button_set_apartment(me, deviceId, btnAppartment);

    /* notify change */
    Datamodel_NotifyChange();
    Datamodel_ResetGroupStopTimeoutActive(me);
    PostFIFO_ErrorOK(me->requester);
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_device_set_button_set_local_priority(Datamodel *me, uint16_t deviceId, bool newSetLocalPrioFlag)
{
    sqlite3_stmt *statement;

    vdsm_report(LOG_DEBUG, "Datamodel set device button local priority: %d on device %d\n", newSetLocalPrioFlag, deviceId);
    /* check if device exists */
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT deviceId FROM devices WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    if (sqlite3_step(statement) != SQLITE_ROW) {
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        /* send error response */
        Datamodel_PostError(me->requester, ERROR_DEVICE_NOT_FOUND);
        return;
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    /* update database */
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE deviceButtons SET buttonSetLocalPriority=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, newSetLocalPrioFlag) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    /* notify change */
    Datamodel_Changed();
    Datamodel_NotifyChange();
    PostFIFO_ErrorOK(me->requester);
}

/*..........................................................................*/
static void Datamodel_device_set_button_set_no_coming_home(Datamodel *me, uint16_t deviceId, bool noComingHome)
{
    sqlite3_stmt *statement;

    vdsm_report(LOG_DEBUG, "Datamodel set device button no coming home call: %d on device %d\n", noComingHome, deviceId);
    /* check if device exists */
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT deviceId FROM devices WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    if (sqlite3_step(statement) != SQLITE_ROW) {
        sqlite3_finalize(statement);
        /* send error response */
        Datamodel_PostError(me->requester, ERROR_DEVICE_NOT_FOUND);
        UNLOCK_ACCESS
        return;
    }
    sqlite3_finalize(statement);

    /* update database */
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE deviceButtons SET buttonNoComingHomeCall=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, noComingHome) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    /* notify change */
    Datamodel_Changed();
    Datamodel_NotifyChange();
    PostFIFO_ErrorOK(me->requester);
}

/*..........................................................................*/
static void Datamodel_device_set_button_set_apartment(Datamodel *me, uint16_t deviceId, bool apartment)
{
    sqlite3_stmt *statement;

    vdsm_report(LOG_DEBUG, "Datamodel set device button apartment: %d on device %d\n", apartment, deviceId);
    /* check if device exists */
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT deviceId FROM devices WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    if (sqlite3_step(statement) != SQLITE_ROW) {
        sqlite3_finalize(statement);
        /* send error response */
        Datamodel_PostError(me->requester, ERROR_DEVICE_NOT_FOUND);
        UNLOCK_ACCESS
        return;
    }
    sqlite3_finalize(statement);

    /* update database */
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE deviceButtons SET buttonApartment=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, apartment) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    /* notify change */
    Datamodel_Changed();
    Datamodel_NotifyChange();
    PostFIFO_ErrorOK(me->requester);
}

/*..........................................................................*/
static void Datamodel_device_set_button_set_output_channel(Datamodel *me, uint16_t deviceId, uint8_t channelId, bool sendDownstream)
{
    sqlite3_stmt *statement;
    vdsm_report(LOG_DEBUG, "Datamodel set device button set output channel id: %d on device %d\n", channelId, deviceId);
    /* check if device exists */
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT deviceId FROM devices WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    if (sqlite3_step(statement) != SQLITE_ROW) {
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        /* send error response */
        Datamodel_PostError(me->requester, ERROR_DEVICE_NOT_FOUND);
        return;
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    /* update database */
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE deviceButtons SET buttonOutputChannel=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, channelId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    if (sendDownstream) {
        /* send over downstream to device and update datamodel then */
        DeviceParameterEvt * reqEvt = Q_NEW(DeviceParameterEvt, SELECT_B_SET_DEVICE_PARAMETER_SIG);
        reqEvt->deviceId =   deviceId;
        reqEvt->paramOffset = PBCHANNEL;
        reqEvt->paramBank = 3;
        reqEvt->value = channelId;
        reqEvt->sendDoneModifier = EVENT_DSM_DONE_SET_DEVICE_BUTTON_OUTPUT_CHANNEL;
        QActive_postFIFO(AO_SelectB, (QEvent *)reqEvt);
    } else {
        /* release any sync command which is waiting */
        DsmApi_SendDsmDoneSetDeviceEvent(EVENT_DSM_DONE_SET_DEVICE_BUTTON_OUTPUT_CHANNEL, ERROR_OK, deviceId);
    }

    /* notify change */
    Datamodel_Changed();
    Datamodel_NotifyChange();
    PostFIFO_ErrorOK(me->requester);
}

/*..........................................................................*/
static void Datamodel_device_set_button_set_global_group(Datamodel *me, uint16_t deviceId, uint8_t group, bool sendDownstream)
{
    sqlite3_stmt *statement;
    vdsm_report(LOG_DEBUG, "Datamodel set device button set group id: %d on device %d\n", group, deviceId);
    /* check if device exists */
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT deviceId FROM devices WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    if (sqlite3_step(statement) != SQLITE_ROW) {
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        /* send error response */
        Datamodel_PostError(me->requester, ERROR_DEVICE_NOT_FOUND);
        return;
    }
    sqlite3_finalize(statement);

    /* update database */
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE deviceButtons SET buttonGlobalGroup=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, group) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    if (sendDownstream) {
        /* send over downstream to device and update datamodel then */
        DeviceParameterEvt * reqEvt = Q_NEW(DeviceParameterEvt, SELECT_B_SET_DEVICE_PARAMETER_SIG);
        reqEvt->deviceId =   deviceId;
        reqEvt->paramOffset = PBGROUP;
        reqEvt->paramBank = 3;
        reqEvt->value = group;
        reqEvt->sendDoneModifier = EVENT_DSM_DONE_SET_DEVICE_BUTTON_GLOBAL_GROUP;
        QActive_postFIFO(AO_SelectB, (QEvent *)reqEvt);
    } else {
        /* release any sync command which is waiting */
        DsmApi_SendDsmDoneSetDeviceEvent(EVENT_DSM_DONE_SET_DEVICE_BUTTON_GLOBAL_GROUP, ERROR_OK, deviceId);
    }

    /* notify change */
    Datamodel_Changed();
    Datamodel_NotifyChange();
    PostFIFO_ErrorOK(me->requester);
}

/*..........................................................................*/
static void Datamodel_device_set_ltmode(Datamodel *me, uint16_t deviceId, uint8_t LTMODERegister)
{
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE devices SET ltmode=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, LTMODERegister) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    /* notify change */
    Datamodel_NotifyChange();
    Datamodel_ResetGroupStopTimeoutActive(me);
    PostFIFO_ErrorOK(me->requester);
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_device_set_output_load(Datamodel *me, uint16_t deviceId, uint8_t MODERegister)
{
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE devices SET mode=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, MODERegister) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS


    /* notify change */
    Datamodel_NotifyChange();
    PostFIFO_ErrorOK(me->requester);
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_device_set_button_id_and_group(Datamodel *me, uint16_t deviceId, uint8_t buttonIdAndGroupRegister)
{
    sqlite3_stmt *statement;
    DatamodelDevice device;
    uint8_t prevGroup;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM devices, deviceButtons WHERE devices.deviceId == deviceButtons.deviceId AND devices.deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    if (sqlite3_step(statement) != SQLITE_ROW) {
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        /* send error response */
        Datamodel_PostError(me->requester, ERROR_DEVICE_NOT_FOUND);
        return;
    }
    Datamodel_fill_device_from_SQL(&device, statement);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    prevGroup = Datamodel_DeviceButtonGroup(&device);
    vdsm_report(LOG_DEBUG, "Datamodel set button id and group for device %d from %d to %d\n", deviceId, device.buttonIdAndGroup, buttonIdAndGroupRegister);
    if (buttonIdAndGroupRegister == device.buttonIdAndGroup) {
        /* is identical */
        PostFIFO_ErrorOK(me->requester);
        return;
    }
    /* set */
    device.buttonIdAndGroup = buttonIdAndGroupRegister;

    if (Fid_IsJocker(device.fid)) {
        {
            DatamodelReqGroupMembershipModifyEvt * reqEvt =
                Q_NEW(DatamodelReqGroupMembershipModifyEvt, DATAMODEL_REQ_GROUP_MEMBERSHIP_MODIFY_SIG);
            reqEvt->basic.requester = (QActive *)me;
            reqEvt->modifier = DEVICE_GROUP_MEMBERSHIP_MODIFY_REMOVE;
            reqEvt->deviceAddress = device.deviceId;
            reqEvt->groupId = prevGroup;
            reqEvt->sendDownstream = FALSE;
            QActive_postFIFO((QActive*)me, (QEvent *)reqEvt);
        }
        {
            DatamodelReqGroupMembershipModifyEvt * reqEvt =
                Q_NEW(DatamodelReqGroupMembershipModifyEvt, DATAMODEL_REQ_GROUP_MEMBERSHIP_MODIFY_SIG);
            reqEvt->basic.requester = (QActive *)me;
            reqEvt->modifier = DEVICE_GROUP_MEMBERSHIP_MODIFY_ADD;
            reqEvt->deviceAddress = device.deviceId;
            reqEvt->groupId = Datamodel_DeviceButtonGroup(&device);
            reqEvt->sendDownstream = FALSE;
            QActive_postFIFO((QActive*)me, (QEvent *)reqEvt);
        }
    }
    {
        LOCK_ACCESS
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE deviceButtons SET buttonIdAndGroup=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, buttonIdAndGroupRegister) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
    }
    /* notify change */
    Datamodel_NotifyChange();
    Datamodel_ResetGroupStopTimeoutActive(me);
    PostFIFO_ErrorOK(me->requester);
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_device_set_locked(Datamodel *me, uint16_t deviceId, bool locked)
{
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE devices SET locked=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, locked) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    /* notify change */
    Datamodel_NotifyChange();
    PostFIFO_ErrorOK(me->requester);
    Datamodel_Changed();
}

/*..........................................................................*/
static void Datamodel_remove_inactive_device(Datamodel *me)
{
    sqlite3_stmt *statement;
    int ret;
    bool deviceRemoved = FALSE;

    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT deviceId FROM devices WHERE active=0;", -1, &statement, 0) == SQLITE_OK);
    do {
        ret = sqlite3_step(statement);
        if (ret == SQLITE_ROW) {
            uint16_t deviceId = sqlite3_column_int(statement, 0);
            ClearBitSet8(me->validDevices, deviceId);
            {
                sqlite3_stmt *statement2;
                SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "DELETE FROM devices WHERE deviceId=?;", -1, &statement2, 0) == SQLITE_OK);
                SQL_ASSERT(sqlite3_bind_int(statement2, 1, deviceId) == SQLITE_OK);
                SQL_ASSERT(sqlite3_step(statement2) == SQLITE_DONE);
                sqlite3_finalize(statement2);
            }
            deviceRemoved = TRUE;
        }
    } while (ret == SQLITE_ROW);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    PostFIFO_ErrorOK(me->requester);
    if (deviceRemoved) {
        Datamodel_Changed();
        Datamodel_NotifyChange();
    }
}

/*..........................................................................*/
static void Datamodel_remove_device(Datamodel *me, uint16_t deviceId)
{
    sqlite3_stmt *statement;
    DatamodelDevice device;
    vdsm_report(LOG_DEBUG, "Datamodel_remove_device: Remove device %d\n", deviceId);
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM devices, deviceButtons WHERE devices.deviceId == deviceButtons.deviceId AND devices.deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    if (sqlite3_step(statement) != SQLITE_ROW) {
        sqlite3_finalize(statement);
        UNLOCK_ACCESS

        /* send error response */
        Datamodel_PostError(me->requester, ERROR_DEVICE_NOT_FOUND);
        return;
    }
    Datamodel_fill_device_from_SQL(&device, statement);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    /* remove device from zone */
    DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByLocalId(device.localZoneId);
    /* remove device from valid devices */
    ClearBitSet8(me->validDevices, deviceId);
    /* was device active? */
    if (IsBitSet8Set(me->activeDevices, deviceId)) {
        /* send device accessibility event */
        dsuid_t dsuid;
        dsuid_from_string(device.dsuid, &dsuid);
        DsmApi_SendDeviceAccessibilityEvent(deviceId, zoneInfo->super.zoneId, device.vid, 0, device.fid, device.ver, dsuid, FALSE);
    } else {
        /* reset default state (active) */
        SetBitSet8(me->activeDevices, deviceId);
    }
    /* send response */
    PostFIFO_ErrorOK(me->requester);
    /* delete device in db */
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "DELETE FROM devices WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    /* notify change */
    Datamodel_Changed();
    Datamodel_NotifyChange();
}

/*..........................................................................*/
static void Datamodel_factory_reset(Datamodel *me __attribute__((unused)))
{
    static const QEvent sysTermEvt = { SYSTEM_TERM_SIG, 0 };

    /* TODO: clean DB */

    /* done -> send system termination signal */
    QF_publish(&sysTermEvt);
    /* go to stop state */
}

/*..........................................................................*/
static bool Datamodel_GetZoneInfoPerVdcByIndex(Datamodel *me,
                                               uint8_t zoneIndex,
                                               const dsuid_t *vdc_dsuid,
                                               DatamodelZoneInfo **zoneInfo)
{
    bool success = false;

    char vdc_dsuid_str[DSUID_STR_LEN];
    dsuid_to_string_upper(vdc_dsuid, vdc_dsuid_str);
    sqlite3_stmt *statement;

    uint8_t localZoneId = 0;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT DISTINCT localZoneId FROM devices WHERE vdcdsuid=?;",
                                  -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 1, vdc_dsuid_str, -1, SQLITE_STATIC) == SQLITE_OK);
    int dbRet = 0;
    int ctr;
    for (ctr = 0; ctr <= zoneIndex; ++ctr) {
        dbRet = sqlite3_step(statement);
    }

    if (dbRet == SQLITE_ROW) {
        localZoneId = sqlite3_column_int(statement, 0);
        success = true;
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS;

    DatamodelZoneInfo *zone = Datamodel_GetZoneInfoByLocalId(localZoneId);

    *zoneInfo = zone;
    return success;
}

/*..........................................................................*/
static QState Datamodel_stopped(Datamodel *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        vdsm_report(LOG_DEBUG, "Datamodel_stopped: close database connection\n");
        LOCK_ACCESS
        int ret = sqlite3_close(me->pDb);
        if (ret != SQLITE_OK) {
            vdsm_report(LOG_ERR, "Datamodel_stopped: Could not close database connection. Error: %d\n", ret);
        }
        me->databaseOpen = FALSE;
        UNLOCK_ACCESS

        /* send data model stopped event */
        static const QEvent datamodelStoppedEvt = { DATAMODEL_STOPPED_SIG, 0 };
        QActive_postFIFO(AO_MMI, &datamodelStoppedEvt);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Datamodel_stopped_top);
}

/*..........................................................................*/
static QState Datamodel_stopped_top(Datamodel *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_REQ_PING_DEVICE_STATE_SIG: {
        /* Let SelectB finish to shut down */
        DatamodelReqPingDeviceStateEvt const * reqEvt = (DatamodelReqPingDeviceStateEvt const *)e;
        if (reqEvt->basic.requester != 0) {
            DatamodelResPingDeviceEvt * resEvt;
            resEvt = Q_NEW(DatamodelResPingDeviceEvt, DATAMODEL_RES_PING_DEVICE_SIG);
            resEvt->oneTimePingDone = FALSE;
            resEvt->nextDeviceId = reqEvt->deviceId;
            QActive_postFIFO(reqEvt->basic.requester, (QEvent *)resEvt);
        }
        return Q_HANDLED();
    }
    /* ------------------------------------ */
    default: {
        if (DATAMODEL_REQ_MIN_SIG <= e->sig && e->sig <= DATAMODEL_REQ_MAX_SIG) {
            DatamodelReqAddDeviceEvt const * reqEvt = (DatamodelReqAddDeviceEvt const *)e;
            Datamodel_PostError(reqEvt->basic.requester, ERROR_REQUEST_CAN_NOT_BE_EXECUTED);
            return Q_HANDLED();
        }
        break;
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static void Datamodel_send_device_accessibility(Datamodel *me, uint16_t deviceId, bool on)
{
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT dsuid,localZoneId,vid,fid,ver FROM devices WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    if (sqlite3_step(statement) == SQLITE_ROW) {
        const char* dsuid_str = sqlite3_column_text(statement, 0);
        Q_ASSERT(dsuid_str != NULL);
        uint8_t localZoneId = sqlite3_column_int(statement, 1);
        uint16_t vid = sqlite3_column_int(statement, 2);
        uint16_t fid = sqlite3_column_int(statement, 3);
        uint16_t ver = sqlite3_column_int(statement, 4);

        uint16_t zoneId;
        DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByLocalId(localZoneId);
        Q_ASSERT(zoneInfo != 0);
        zoneId = zoneInfo->super.zoneId;
        dsuid_t dsuid;
        dsuid_from_string(dsuid_str, &dsuid);

        sqlite3_finalize(statement);
        UNLOCK_ACCESS

        DsmApi_SendDeviceAccessibilityEvent(deviceId, zoneId, vid, 0, fid, ver, dsuid, on);
        /* notify change */
        Datamodel_NotifyChange();
    }
    else
    {
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
    }
}

/*..........................................................................*/
static void Datamodel_InitCluster(DatamodelCluster *cluster, uint8_t clusterId)
{
    memset(cluster, 0, sizeof(*cluster));
    cluster->id = clusterId;
    /* set default values */
    cluster->stateMachineGroupId = 0;
    cluster->stopTimeout = 0;
    cluster->remoteStopTimeout = 0;
    memset(cluster->name, 0, sizeof(cluster->name));
    cluster->location = 0;
    cluster->floor = 0;
    cluster->protectionClass = 0;
    cluster->configurationLock = 0;
    memset(cluster->sceneLock, 0, sizeof(cluster->sceneLock));
}

/*..........................................................................*/
static void Datamodel_InitZoneInfo(DatamodelZoneInfo * zoneInfo)
{
    int i;
    memset(zoneInfo->groupMembership, 0, sizeof(zoneInfo->groupMembership));
    /* set initial group membership for all known groups */
    SetBitSet8(zoneInfo->groupMembership, BROADCAST_GROUP);
    SetBitSet8(zoneInfo->groupMembership, LIGHT_GROUP);
    SetBitSet8(zoneInfo->groupMembership, SHADE_GROUP);
    SetBitSet8(zoneInfo->groupMembership, HEATING_GROUP);
    SetBitSet8(zoneInfo->groupMembership, AUDIO_GROUP);
    SetBitSet8(zoneInfo->groupMembership, VIDEO_GROUP);
    SetBitSet8(zoneInfo->groupMembership, SECURITY_GROUP);
    SetBitSet8(zoneInfo->groupMembership, ACCESS_GROUP);
    SetBitSet8(zoneInfo->groupMembership, JOKER_GROUP);
    SetBitSet8(zoneInfo->groupMembership, COOLING_GROUP);
    SetBitSet8(zoneInfo->groupMembership, VENTILATION_GROUP);
    SetBitSet8(zoneInfo->groupMembership, WINDOW_GROUP);
    for (i = 0; i < (NUM_CLUSTERS); ++i) {
        SetBitSet8(zoneInfo->groupMembership, (uint8_t)(FIRST_CLUSTER + i));
    }
    SetBitSet8(zoneInfo->groupMembership, ROOM_TEMPERATURE_CONTROL_GROUP);
}

/**
 * Get zone info.
 * @param localId Local ID (1-15).
 * @return specific zone info (check magic for validity) or null (invalid zone ID).
 */
DatamodelZoneInfo * Datamodel_GetZoneInfoByLocalId(uint8_t localId)
{
    if ((localId >= MIN_ZONE) && (localId <= MAX_ZONE)) {
        return &l_datamodel.zoneInfos[localId-MIN_ZONE];
    }
    return 0;
}

/*..........................................................................*/
static void Datamodel_add_default_zone_for_vdc(Datamodel *me, const dsuid_t *dsmid)
{
    uint16_t zoneId = Datamodel_create_zoneId_from_dsuid(dsmid);
    Datamodel_CreateZone(me, zoneId);
}

/*..........................................................................*/
static uint8_t Datamodel_CreateZone(Datamodel *me, uint16_t zoneId) 
{
    /* get unused local zone id and set it to used */
    uint8_t localZoneId = Datamodel_GetUnusedLocalZoneId();
    Q_ASSERT(localZoneId != 0);

    /* increment number of zones */
    ++me->numberOfZones;

    Datamodel_SetLocalZoneIdUsed(localZoneId);
    {
        /* initialize zone info */
        DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByLocalId(localZoneId);
        Q_ASSERT(zoneInfo != 0);
        Datamodel_InitZoneInfo(zoneInfo);
        Datamodel_InitZone(&(zoneInfo->super), localZoneId, zoneId, 0);
    }
    Datamodel_add_zone_to_db(me, localZoneId, zoneId);

    /* publish new zone */
    Datamodel_publish_zone_added(localZoneId, zoneId);
    return localZoneId;
}

/*..........................................................................*/
static void Datamodel_RemoveZone(Datamodel *me, uint16_t zoneId)
{
    DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByZoneId(zoneId);
    uint16_t localZoneId = zoneInfo->super.localZoneId;
    Q_ASSERT(localZoneId != 0);

    /* decrement number of zones */
    --me->numberOfZones;

    Datamodel_SetLocalZoneIdUnused(localZoneId);
    {
        /* notify removed user groups */
        uint8_t groupId;
        for (groupId = USER_GROUP; groupId <= MAX_USER_GROUP_WITH_HSM; ++groupId) {
            if (IsBitSet8Set(zoneInfo->groupMembership, groupId)) {
                Datamodel_NotifyUserGroup(groupId, 0);
            }
        }
    }
    /* publish deleted zone */
    Datamodel_publish_zone_added(localZoneId, DATAMODEL_INVALID_ZONE_ID);
    
    /* clear magic number */
    zoneInfo->super.magic = 0;
    Datamodel_remove_zone_from_db(me, localZoneId);
}

/*..........................................................................*/
static DatamodelZoneInfo * Datamodel_CreateGetZoneInfoByDsuid(Datamodel *me,
                                                              const dsuid_t *vdcdsuid)
{
    uint16_t zoneId = 0;
    zoneId = Datamodel_create_zoneId_from_dsuid(vdcdsuid);
    DatamodelZoneInfo *zone = Datamodel_GetZoneInfoByZoneId(zoneId);
    if (!zone) {
        Datamodel_add_default_zone_for_vdc(me, vdcdsuid);
        zone = Datamodel_GetZoneInfoByZoneId(zoneId);
    }
    return zone;
}

/**
 * Get zone info.
 * @param zoneId Global zone ID.
 * @return specific zone info (check magic for validity) or null (no zone info found).
 */
static DatamodelZoneInfo * Datamodel_GetZoneInfoByZoneId(uint16_t zoneId)
{
    uint8_t id;
    for (id = MIN_ZONE; id <= MAX_ZONE; ++id) {
        DatamodelZoneInfo * zoneInfo = &l_datamodel.zoneInfos[id-MIN_ZONE];
        if ((zoneInfo->super.magic == DATAMODEL_MAGIC) && (zoneInfo->super.zoneId == zoneId)) {
            return zoneInfo;
        }
    }
    return 0;
}

/**
 * Returns the number of groups in the zone
 * @param zoneInfo
 * @return
 */
static uint8_t Datamodel_GetGroupCountOfZone(DatamodelZoneInfo const * zoneInfo)
{
    uint8_t id;
    uint8_t count = 0;
    for (id = 0; id <= MAX_GROUP; ++id) {
        if (IsBitSet8Set(zoneInfo->groupMembership, id)) {
            ++count;
        }
    }
    return count;
}

/**
 * Returns the group ID or (MAX_GROUP+1) if group index does not exists
 * @param zoneInfo
 * @param groupIndex
 * @return
 */
static uint8_t Datamodel_GetGroupIdByGroupIndex(DatamodelZoneInfo const * zoneInfo, uint8_t index)
{
    uint8_t id;
    for (id = 0; id <= MAX_GROUP; ++id) {
        if (IsBitSet8Set(zoneInfo->groupMembership, id)) {
            if (index == 0) {
                break;
            }
            --index;
        }
    }
    return id;
}

/*..........................................................................*/
static void Datamodel_SendClusterInfoResponse(Datamodel *me, DatamodelCluster const * cluster, uint16_t numOfDevices)
{
    DatamodelResClusterInfoEvt *resEvt = Q_NEW(DatamodelResClusterInfoEvt, DATAMODEL_RES_CLUSTER_INFO_SIG);
    resEvt->clusterId = cluster->id;
    resEvt->countDevices = numOfDevices;
    resEvt->stateMachineGroupId = cluster->stateMachineGroupId;
    resEvt->stopTimeout = cluster->stopTimeout;
    resEvt->remoteStopTimeout = cluster->remoteStopTimeout;
    memcpy(resEvt->name, cluster->name, sizeof(resEvt->name));
    resEvt->configLock = cluster->configurationLock;
    memcpy(resEvt->sceneLock, cluster->sceneLock, sizeof(resEvt->sceneLock));
    resEvt->location = cluster->location;
    resEvt->floor = cluster->floor;
    resEvt->protectionClass = cluster->protectionClass;
    QActive_postFIFO(me->requester, (QEvent *)resEvt);
}

/*..........................................................................*/
static uint8_t Datamodel_ZoneCount(Datamodel *me, const dsuid_t *vdc_dsuid)
{
    sqlite3_stmt *statement;

    char device_dsuid_str[DSUID_STR_LEN];
    dsuid_to_string_upper(vdc_dsuid, device_dsuid_str);

    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT COUNT(DISTINCT localZoneId) FROM devices where vdcdsuid=?;",
                                  -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 1, device_dsuid_str, -1, SQLITE_STATIC) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    uint8_t rows = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS;

    return rows;
}

/*..........................................................................*/
static uint8_t Datamodel_GetUnusedLocalZoneId(void)
{
    uint8_t localId;
    for (localId = MIN_ZONE; localId <= MAX_ZONE; ++localId) {
        if (!IsBitSet8Set(l_datamodel.validZoneInfos, localId)) {
            return localId;
        }
    }
    return 0;
}

/*..........................................................................*/
static void Datamodel_SetLocalZoneIdUsed(uint8_t localId)
{
    SetBitSet8(l_datamodel.validZoneInfos, localId);
}

/*..........................................................................*/
static void Datamodel_SetLocalZoneIdUnused(uint8_t localId)
{
    ClearBitSet8(l_datamodel.validZoneInfos, localId);
}

/*..........................................................................*/
static bool Datamodel_GetDefaultLocalPrioritySetting(uint16_t fid)
{
    if (FidGetClass(fid) == FID_CLASS_SHADE) {
        return FALSE;
    }
    return TRUE;
}

/*..........................................................................*/
static void Datamodel_InitDevice(DatamodelDevice * device,
                                 const char *dsuid_str,
                                 uint16_t deviceId,
                                 uint16_t vid,
                                 uint16_t fid,
                                 uint16_t ver,
                                 uint8_t locked,
                                 const char *vdcdsuid_str,
                                 uint16_t zoneId)
{
    /* current structure version */
    device->structId = DATAMODEL_DEVICE_STRUCT_VERSION;
    strncpy(device->dsuid, dsuid_str, sizeof(device->dsuid));
    device->deviceId = deviceId;
    device->vid = vid;
    device->fid = fid;
    device->ver = ver;
    device->locked = locked;
    device->ltmode = 0;
    device->mode = 0;
    memset(device->groupMembership, 0, sizeof(device->groupMembership));
    device->buttonIdAndGroup = (FidGetClass(fid) << 4) | BUTTON_ZONE_0;
    device->buttonUserGroup = DATAMODEL_GROUP_ID_NOT_APPLICABLE;
    device->buttonOutputChannel = 0;
    device->buttonGlobalGroup = 0;
    memset(device->name, 0, MAX_NAME_LEN);
    memset(device->binaryInputTable, 0, sizeof(device->binaryInputTable));
    memset(device->sensorTable, 0xff, sizeof(device->sensorTable));
    memset(device->opcTable, 0xff, sizeof(device->opcTable));
    device->magic = DATAMODEL_MAGIC;
    device->localZoneId = 0;
    {
        // get default vdc zone id, if not defined
        if (zoneId == 0) {
            zoneId = Datamodel_getVdcDefaultZoneID(vdcdsuid_str);
        }
        if (zoneId != 0) {
            Datamodel *me = &l_datamodel;
            device->localZoneId  = Datamodel_get_or_create_localZone_by_zoneId(me, zoneId);
        } else {
            // Assign local zone id
            dsuid_t vdcdsuid;
            dsuid_from_string(vdcdsuid_str, &vdcdsuid);
            Datamodel *me = &l_datamodel;
            DatamodelZoneInfo* zoneInfo;
            zoneInfo = Datamodel_CreateGetZoneInfoByDsuid(me, &vdcdsuid);
            if (zoneInfo) {
                device->localZoneId = zoneInfo->super.localZoneId;
            }
        }
    }
    Q_ASSERT(device->localZoneId != 0);

    /* every device is member of overall group */
    SetBitSet8(device->groupMembership, FID_CLASS_OVERALL);

    device->flags = 0;
    /* get default of local priority depending on color */
    device->buttonSetLocalPriority = Datamodel_GetDefaultLocalPrioritySetting(fid);
    device->active = TRUE;
    strncpy(device->vdcdsuid, vdcdsuid_str, sizeof(device->vdcdsuid));
}

/*..........................................................................*/
static void Datamodel_InitGroup(Datamodel *me, uint8_t localZoneId, uint8_t groupId, uint8_t stateMachineGroupId, const uint8_t * name)
{
    static const uint8_t defaultUserGroupName[] = "";
    sqlite3_stmt *statement;
    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT INTO groups VALUES (?,?,?,?,?,?);", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, (localZoneId << 8) | groupId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, stateMachineGroupId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 3, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 4, (groupId == SHADE_GROUP) ? 120 : 0) == SQLITE_OK);
    if (name != NULL) {
        SQL_ASSERT(sqlite3_bind_text(statement, 5, (const char *)name, -1, SQLITE_STATIC) == SQLITE_OK);
    } else {
        SQL_ASSERT(sqlite3_bind_text(statement, 5, (const char *)defaultUserGroupName, -1, SQLITE_STATIC) == SQLITE_OK);
    }
    SQL_ASSERT(sqlite3_bind_int(statement, 6, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
}

/*..........................................................................*/
static void Datamodel_InitZone(DatamodelZone * zone, uint8_t localZoneId, uint16_t zoneId, uint8_t zoneTagId)
{
    zone->localZoneId = localZoneId;
    zone->zoneId = zoneId;
    zone->zoneTagId = zoneTagId;
    zone->magic = DATAMODEL_MAGIC;
    memset(zone->name, 0, sizeof(zone->name));
}


/*..........................................................................*/
static bool Datamodel_GetLocalZoneId(uint16_t zoneId, uint8_t *localZoneId)
{
    uint8_t id;
    for (id = MIN_ZONE; id <= MAX_ZONE; ++id) {
        DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByLocalId(id);
        if (zoneInfo->super.magic == DATAMODEL_MAGIC) {
            if (zoneInfo->super.zoneId == zoneId) {
                *localZoneId = zoneInfo->super.localZoneId;
                return TRUE;
            }
        }
    }
    return FALSE;
}

/*..........................................................................*/
static void Datamodel_PostError(QActive *activeObj, uint8_t error)
{
    DatamodelErrorEvt * errorEvt = Q_NEW(DatamodelErrorEvt, DATAMODEL_RES_ERROR_SIG);
    errorEvt->error = error;
    QActive_postFIFO(activeObj, (QEvent *)errorEvt);
}

/*..........................................................................*/
static void Datamodel_PostLongError(QActive *activeObj, uint8_t error, uint8_t privateId)
{
    DatamodelLongErrorEvt * errorEvt = Q_NEW(DatamodelLongErrorEvt, DATAMODEL_RES_ERROR_SIG);
    errorEvt->error = error;
    errorEvt->privateId = privateId;
    QActive_postFIFO(activeObj, (QEvent *)errorEvt);
}

/*..........................................................................*/
static void Datamodel_SendZoneInfo(QActive *activeObj, DatamodelZoneInfo* zoneInfo)
{
    DatamodelResZoneInfoEvt * resEvt =
        Q_NEW(DatamodelResZoneInfoEvt, DATAMODEL_RES_ZONE_INFO_SIG);
    resEvt->zoneId = zoneInfo->super.zoneId;
    resEvt->localZoneId = zoneInfo->super.localZoneId;
    resEvt->groupCount = Datamodel_GetGroupCountOfZone(zoneInfo);
    memcpy(resEvt->name, zoneInfo->super.name, sizeof(resEvt->name));
    QActive_postFIFO(activeObj, (QEvent *)resEvt);
}

/*..........................................................................*/
static void Datamodel_NotifyChange(void)
{
    static QEvent const datamodelChangedEvt = { DATAMODEL_CHANGED_SIG, 0 };
    QActive_postFIFO(AO_SelectA, &datamodelChangedEvt);
}

/*..........................................................................*/
static void Datamodel_NotifyUserGroup(uint8_t groupId, uint8_t stateMachineGroupId)
{
    if ((groupId >= USER_GROUP) && (groupId <= MAX_USER_GROUP_WITH_HSM)) {
        NotifyUserGroupEvt * notifyEvt = Q_NEW(NotifyUserGroupEvt, NOTIFY_USER_GROUP_SIG);
        notifyEvt->groupId = groupId;
        notifyEvt->stateMachineGroupId = stateMachineGroupId;
        QActive_postFIFO(AO_SelectA, (QEvent *)notifyEvt);
    }
}

/*..........................................................................*/
static bool Datamodel_SelectRandomDeviceId(Datamodel *me, uint16_t *deviceId)
{
    /* create random number */
    uint16_t count = (rand() % (MAX_DEVICE-MIN_DEVICE)) + 1;
    /* find count free valid device */
    uint16_t id = MIN_DEVICE;
    bool freeDeviceIdFound = FALSE;
    for (;;) {
        /* get next free valid device ID */
        while (IsBitSet8Set(me->validDevices, id)) {
            if (++id > MAX_DEVICE) {
                if (!freeDeviceIdFound) {
                    return FALSE;
                }
                id = MIN_DEVICE;
            }
        }
        /* set free device ID found flag */
        freeDeviceIdFound = TRUE;
        /* decrement random number count */
        if (--count == 0) {
            break;
        }
        /* increment device ID */
        if (++id > MAX_DEVICE) {
            id = MIN_DEVICE;
        }
    }
    /* update last valid device id */
    if (id > me->lastValidDeviceId) {
        me->lastValidDeviceId = id;
    }
    /* return device id */
    *deviceId = id;
    return TRUE;
}

/*..........................................................................*/
static void Datamodel_UpdateCachedStopTimeout(Datamodel *me, uint8_t localZoneId, uint8_t groupId, uint8_t stopTimeout)
{
    if (groupId < LAST_STANDARD_GROUP) {
        me->stopTimeouts.stopTimeouts[localZoneId-MIN_ZONE][groupId] = stopTimeout;
    } else if (groupId <= MAX_USER_GROUP_WITH_HSM) {
        me->stopTimeouts.userGroupStopTimeouts[groupId-USER_GROUP] = stopTimeout;
    }
}

/*..........................................................................*/
void Datamodel_ResetGroupStopTimeoutActive(Datamodel *me)
{
    sqlite3_stmt *statement;
    int ret;

    LOCK_ACCESS
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE groups SET stopTimeoutActive=0;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM devices, deviceButtons WHERE devices.deviceId == deviceButtons.deviceId;", -1, &statement, 0) == SQLITE_OK);
    do
    {
        ret = sqlite3_step(statement);
        if (ret == SQLITE_ROW) {
            DatamodelDevice device;
            Datamodel_fill_device_from_SQL(&device, statement);
            Datamodel_SetGroupActiveStopTimeout(me, &device);
        }
    } while (ret == SQLITE_ROW);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
}

/*............................................................................*/
// LOCK_ACCESS: do not use it here. function is internally used.
// Locking in under responsibility of calling function
static void Datamodel_SetGroupActiveStopTimeout(Datamodel *me, DatamodelDevice const *device)
{
    if (Fid_HasButton(device->fid) &&
        (Datamodel_DeviceButtonId(device) > 0) &&
        (Datamodel_DeviceButtonId(device) < 14))
    {
        sqlite3_stmt *statement;
        uint8_t groupId = Datamodel_DeviceButtonActiveGroup(device);
        uint8_t localZoneId = device->localZoneId;
        // use localzoneId 0 when group doesn't exist
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM groups WHERE g_id=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, (localZoneId << 8) | groupId) == SQLITE_OK);
        if (sqlite3_step(statement) != SQLITE_ROW) {
            localZoneId = 0;
        }
        sqlite3_finalize(statement);
        // update stop timeout active flag
        if (groupId < LAST_STANDARD_GROUP) {
            SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE groups SET stopTimeoutActive=1 WHERE g_id=?;", -1, &statement, 0) == SQLITE_OK);
            SQL_ASSERT(sqlite3_bind_int(statement, 1, (localZoneId << 8) | groupId) == SQLITE_OK);
            SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
            sqlite3_finalize(statement);
            // update cache
            SetBitSet8(me->stopTimeouts.activeGroupStopTimeout[device->localZoneId-MIN_ZONE], groupId);
        } else if (groupId <= MAX_USER_GROUP_WITH_HSM) {
            sqlite3_stmt *statement;
            for (localZoneId = MIN_ZONE; localZoneId <= MAX_ZONE; ++localZoneId) {
                SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE groups SET stopTimeoutActive=1 WHERE g_id=?;", -1, &statement, 0) == SQLITE_OK);
                SQL_ASSERT(sqlite3_bind_int(statement, 1, (localZoneId << 8) | groupId) == SQLITE_OK);
                SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
                sqlite3_finalize(statement);
            }
            // update cache
            SetBitSet8(me->stopTimeouts.activeGroupStopTimeout[0], (uint8_t)(groupId-USER_GROUP));
        }
    }
}

/*............................................................................*/
// LOCK_ACCESS: do not use it here. function is internally used.
// Locking in under responsibility of calling function
static void Datamodel_ClearGroupActiveStopTimeout(Datamodel *me, uint8_t localZoneId, uint8_t groupId)
{
    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM groups WHERE g_id=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, (localZoneId << 8) | groupId) == SQLITE_OK);
    int ret = sqlite3_step(statement);
    sqlite3_finalize(statement);
    if (ret != SQLITE_ROW) {
        return;
    }
    // update stop timeout active flag
    if (groupId < USER_GROUP) {
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE groups SET stopTimeoutActive=0 WHERE g_id=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, (localZoneId << 8) | groupId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
        // update cache
        ClearBitSet8(me->stopTimeouts.activeGroupStopTimeout[localZoneId-MIN_ZONE], groupId);
    } else if (groupId <= MAX_USER_GROUP_WITH_HSM) {
        sqlite3_stmt *statement;
        for (localZoneId = MIN_ZONE; localZoneId <= MAX_ZONE; ++localZoneId) {
            SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE groups SET stopTimeoutActive=0 WHERE g_id=?;", -1, &statement, 0) == SQLITE_OK);
            SQL_ASSERT(sqlite3_bind_int(statement, 1, (localZoneId << 8) | groupId) == SQLITE_OK);
            SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
            sqlite3_finalize(statement);
        }
        // update cache
        ClearBitSet8(me->stopTimeouts.activeGroupStopTimeout[0], (uint8_t)(groupId-USER_GROUP));
    }
}

/*..........................................................................*/
static bool Datamodel_IsZoneSensor(Datamodel *me, uint16_t deviceId, uint8_t sensorIndex, dsuid_t *dsuid, uint8_t *type, bool *sensorPrecision)
{
    if (!me || !dsuid || !type || !sensorPrecision) {
        vdsm_report (LOG_DEBUG, "Datamodel IsZoneSensor parameter is NULL\n");
        return false;
    }

    LOCK_ACCESS
    sqlite3_stmt *statement;

    // Get dsuid and type
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT dsuid, type, pushConvert FROM \
               devices, deviceSensors WHERE  \
               devices.deviceId == deviceSensors.deviceId AND devices.deviceId=? AND deviceSensors.sensorIndex=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, sensorIndex) == SQLITE_OK);
    if (sqlite3_step(statement) != SQLITE_ROW) {
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        return false;
    }
    char dsuid_str[DSUID_STR_LEN];
    strncpy(dsuid_str, sqlite3_column_text(statement, 0), DSUID_STR_LEN);
    dsuid_from_string(dsuid_str, dsuid);
    *type = sqlite3_column_int(statement, 1);
    *sensorPrecision = (sqlite3_column_int(statement, 2) > 0);
    sqlite3_finalize(statement);

    // check if device with dsuid is part of zone sensors
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT COUNT (*) FROM \
               zoneSensors WHERE dsuid=? AND type=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 1, dsuid_str, -1, SQLITE_STATIC) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement,  2, *type) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    uint8_t count = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    return (count > 0);
}

/*..........................................................................*/
static bool Datamodel_GetZoneSensorZoneId(Datamodel *me, dsuid_t *sensorDsuid, uint16_t *zoneId, uint8_t *localZoneId)
{
    if (!me || !sensorDsuid || !zoneId || !localZoneId) {
        vdsm_report (LOG_DEBUG, "Datamodel GetZoneSensorZoneId parameter is NULL\n");
        return false;
    }

    char dsuid_str[DSUID_STR_LEN];
    dsuid_to_string_upper(sensorDsuid, dsuid_str);

    LOCK_ACCESS
    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb,"SELECT zones.zoneId, zones.localZoneId FROM zones, zoneSensors WHERE \
               zoneSensors.localZoneId == zones.localZoneId AND zoneSensors.dsuid=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 1, dsuid_str, -1, SQLITE_STATIC) == SQLITE_OK);
    if (sqlite3_step(statement) != SQLITE_ROW) {
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        return false;
    }

    *zoneId = sqlite3_column_int(statement, 0);
    *localZoneId = sqlite3_column_int(statement, 1);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return true;
}

/*..........................................................................*/
void Datamodel_add_vdc(const char *vdcDsuid)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel Add vDC Database closed!\n");
        UNLOCK_ACCESS
        return;
    }
    sqlite3_stmt *statement;
    int ret;
    int datamodel_changed_trigger = 0;

    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT * FROM dsm_info WHERE dsuid=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 1, vdcDsuid, -1, SQLITE_STATIC) == SQLITE_OK);
    ret = sqlite3_step(statement);
    sqlite3_finalize(statement);

    if (ret != SQLITE_ROW) {
        /* insert new element: */
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT INTO dsm_info  \
                                                (dsuid, apartmentId, flags, autoPingInterval, \
                                                oneTimePingInterval, warnTemperature, \
                                                errorTemperature, upstreamFrequency) \
                                                VALUES (?,?,?,?,?,?,?,?);", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement, 1, (const char *)vdcDsuid, -1, SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,  2, DEFAULT_APARTMENT_ID) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,  3, DEFAULT_DSM_INFO_FLAGS) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,  4, DEFAULT_AUTO_PING_INTERVAL) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,  5, DEFAULT_ONE_TIME_PING_INTERVAL) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,  6, DEFAULT_WARN_TEMPERATURE) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,  7, DEFAULT_ERROR_TEMPERATURE) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement,  8, DEFAULT_UPSTREAM_FREQUNECY) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
	datamodel_changed_trigger = 1;
    }
    UNLOCK_ACCESS
    if (datamodel_changed_trigger) {
      Datamodel_Changed();
    }
}

/*..........................................................................*/
uint8_t Datamodel_GetStopTimeoutZoneGroup(uint8_t localZoneId, uint8_t groupId)
{
    Datamodel *me = &l_datamodel;
    if ((groupId < USER_GROUP) && IsBitSet8Set(me->stopTimeouts.activeGroupStopTimeout[localZoneId-MIN_ZONE], groupId)) {
        return me->stopTimeouts.stopTimeouts[localZoneId-MIN_ZONE][groupId];
    }
    uint8_t index = groupId - USER_GROUP;
    if ((groupId <= MAX_USER_GROUP_WITH_HSM) && IsBitSet8Set(me->stopTimeouts.activeGroupStopTimeout[0], index)) {
        return me->stopTimeouts.userGroupStopTimeouts[index];
    }
    return 0;
}

/*..........................................................................*/
uint8_t calculateCircuitId(uint8_t busAddress __attribute__((unused)),
                           uint8_t apartmentId  __attribute__((unused)))
{
    return 0;
}

/*..........................................................................*/
bool Datamodel_GetDsuidFromDeviceId(uint16_t deviceId, char *dsuid)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    sqlite3_stmt *statement;

    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel GetDsuidFromDeviceId Database closed!\n");
        UNLOCK_ACCESS
        return FALSE;
    }

    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT dsuid FROM devices WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    if (sqlite3_step(statement) == SQLITE_ROW) {
        strncpy(dsuid, sqlite3_column_text(statement, 0), DSUID_STR_LEN);
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        return TRUE;
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return FALSE;
}

/*..........................................................................*/
bool Datamodel_GetVdcDsuidFromDeviceId(uint16_t deviceId, dsuid_t *dsuid)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    sqlite3_stmt *statement;

    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel GetVdcDsuidFromDeviceId Database closed!\n");
        UNLOCK_ACCESS
        return FALSE;
    }

    char dsuid_str[DSUID_STR_LEN];
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT vdcdsuid FROM devices WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    if (sqlite3_step(statement) == SQLITE_ROW) {
        strncpy(dsuid_str, sqlite3_column_text(statement, 0), DSUID_STR_LEN);
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        if (dsuid_from_string(dsuid_str, dsuid) < 0) {
            return FALSE;
        }
        return TRUE;
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return FALSE;
}

/*..........................................................................*/
uint16_t Datamodel_GetDeviceIdFromDsuid(const char *dsuid)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel GetDeviceIdFromDsuid Database closed!\n");
        UNLOCK_ACCESS
        return 0;
    }
    uint16_t deviceId = 1023;
    sqlite3_stmt *statement;

    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT deviceId FROM devices WHERE dsuid=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 1, dsuid, -1, SQLITE_STATIC) == SQLITE_OK);
    if (sqlite3_step(statement) == SQLITE_ROW) {
        deviceId = sqlite3_column_int(statement, 0);
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return deviceId;
}

/*..........................................................................*/
uint8_t Datamodel_GetDeviceLtmode(uint16_t deviceId)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel GetDeviceLtmode Database closed!\n");
        UNLOCK_ACCESS
        return 0;
    }
    uint8_t ltmode = 0;
    sqlite3_stmt *statement;

    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT ltmode FROM devices WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    if (sqlite3_step(statement) == SQLITE_ROW) {
        ltmode = sqlite3_column_int(statement, 0);
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return ltmode;
}

/*..........................................................................*/
void Datamodel_SetDeviceSensor(uint16_t deviceId, uint8_t index, uint8_t type, uint32_t pollInterval, uint8_t broadcast, uint8_t pushConvert)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    sqlite3_stmt *statement;
    int ret;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel SetDeviceSensor Database closed!\n");
        UNLOCK_ACCESS
        return;
    }

    vdsm_report(LOG_DEBUG, "Datamodel set sensor for device %d, index: %d, type: %d\n", deviceId, index, type);
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT deviceId FROM deviceSensors WHERE deviceId=? AND sensorIndex=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, index) == SQLITE_OK);
    ret = sqlite3_step(statement);
    sqlite3_finalize(statement);
    if (ret == SQLITE_ROW) {
        /* sensor entry already exists */
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE deviceSensors SET type=?, pollInterval=?, broadcast=?, pushConvert=? WHERE deviceId=? AND sensorIndex=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, type) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, pollInterval) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 3, broadcast) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 4, pushConvert) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 5, deviceId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 6, index) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    } else {
        /* insert sensor entry */
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT INTO deviceSensors(deviceId, sensorIndex, type, pollInterval, broadcast, pushConvert) VALUES (?,?,?,?,?,?);", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, index) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 3, type) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 4, pollInterval) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 5, broadcast) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 6, pushConvert) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    Datamodel_Changed();
}

/*..........................................................................*/
void Datamodel_RemoveSpareSensorEntries(uint16_t deviceId, uint8_t index)
{
  LOCK_ACCESS
  Datamodel *me = &l_datamodel;
  if (!me->databaseOpen) {
      vdsm_report(LOG_ERR, "Datamodel RemoveSpareSensorEntries Database closed!\n");
      UNLOCK_ACCESS
      return;
  }
  sqlite3_stmt *statement;
  vdsm_report(LOG_DEBUG, "Datamodel remove spare sensor entries for device %d from deviceSensors\n", deviceId);
  SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "DELETE FROM deviceSensors WHERE deviceId=? AND sensorIndex>=?;", -1, &statement, 0) == SQLITE_OK);
  SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
  SQL_ASSERT(sqlite3_bind_int(statement, 2, index) == SQLITE_OK);

  sqlite3_step(statement);
  sqlite3_finalize(statement);

  UNLOCK_ACCESS
  Datamodel_Changed();
}

/*..........................................................................*/
void Datamodel_SetDeviceSensorName(uint16_t deviceId, uint8_t index, const char *name)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel SetDeviceSensorName Database closed!\n");
        UNLOCK_ACCESS
        return;
    }
    sqlite3_stmt *statement;
    int ret;

    vdsm_report(LOG_DEBUG, "Datamodel set sensor for device %d, index: %d, name: %s\n", deviceId, index, name);
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT deviceId FROM deviceSensors WHERE deviceId=? AND sensorIndex=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, index) == SQLITE_OK);
    ret = sqlite3_step(statement);
    sqlite3_finalize(statement);
    if (ret == SQLITE_ROW) {
        /* sensor entry already exists */
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE deviceSensors SET name=? WHERE deviceId=? AND sensorIndex=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement, 1, name, -1, SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 3, index) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    } else {
        /* insert sensor entry */
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT INTO deviceSensors(deviceId, sensorIndex, name) VALUES (?,?,?);", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, index) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement, 3, name, -1, SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    Datamodel_Changed();
}

/*..........................................................................*/
uint8_t Datamodel_GetDeviceSensorCount(uint16_t deviceId)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel GetDeviceSensorCount Database closed!\n");
        UNLOCK_ACCESS
        return 0;
    }
    uint8_t count = 0;
    sqlite3_stmt *statement;

    vdsm_report(LOG_DEBUG, "Datamodel get sensor count of device %d\n", deviceId);
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT deviceId FROM deviceSensors WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    while (sqlite3_step(statement) == SQLITE_ROW) {
        ++count;
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return count;
}

/*..........................................................................*/
bool Datamodel_GetDeviceSensor(uint16_t deviceId, uint8_t index, uint8_t *type, uint32_t *pollInterval, uint8_t *broadcast, uint8_t *pushConvert)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel GetDeviceSensor Database closed!\n");
        UNLOCK_ACCESS
        return FALSE;
    }
    sqlite3_stmt *statement;
    bool success = FALSE;

    vdsm_report(LOG_DEBUG, "Datamodel get sensor of device %d, index %d\n", deviceId, index);
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT type, pollInterval, broadcast, pushConvert FROM deviceSensors WHERE deviceId=? AND sensorIndex=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, index) == SQLITE_OK);
    if (sqlite3_step(statement) == SQLITE_ROW) {
        *type = sqlite3_column_int(statement, 0);
        *pollInterval = sqlite3_column_int(statement, 1);
        *broadcast = sqlite3_column_int(statement, 2);
        *pushConvert = sqlite3_column_int(statement, 3);
        success = TRUE;
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return success;
}

/*..........................................................................*/
bool Datamodel_GetDeviceSensorType(uint16_t deviceId, uint8_t index, uint8_t *type)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel GetDeviceSensorType Database closed!\n");
        UNLOCK_ACCESS
        return FALSE;
    }
    sqlite3_stmt *statement;
    bool success = FALSE;

    vdsm_report(LOG_DEBUG, "Datamodel get sensor type of device %d, index %d\n", deviceId, index);
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT type FROM deviceSensors WHERE deviceId=? AND sensorIndex=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, index) == SQLITE_OK);
    if (sqlite3_step(statement) == SQLITE_ROW) {
        *type = sqlite3_column_int(statement, 0);
        success = TRUE;
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return success;
}

/*..........................................................................*/
void Datamodel_RemoveSpareBinaryInputs(uint16_t deviceId, uint8_t index)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel RemoveSpareBinaryInputs Database closed!\n");
        UNLOCK_ACCESS
        return;
    }
    sqlite3_stmt *statement;
    vdsm_report(LOG_DEBUG, "Datamodel remove spare binary inputs for device %d from deviceSensors\n", deviceId);
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "DELETE FROM deviceBinaryInputs WHERE deviceId=? AND inputIndex>=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, index) == SQLITE_OK);
    sqlite3_step(statement);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    Datamodel_Changed();
}

/*..........................................................................*/
void Datamodel_SetDeviceBinaryInput(uint16_t deviceId, uint8_t index, uint8_t group, uint8_t type, uint8_t buttonId)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel SetDeviceBinaryInput Database closed!\n");
        UNLOCK_ACCESS
        return;
    }
    sqlite3_stmt *statement;
    int ret;

    vdsm_report(LOG_DEBUG, "Datamodel set binary input for device %d and index %d, group: %d, type: %d, buttonId: %d\n", deviceId, index, group, type, buttonId);
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT deviceId FROM deviceBinaryInputs WHERE deviceId=? AND inputIndex=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, index) == SQLITE_OK);
    ret = sqlite3_step(statement);
    sqlite3_finalize(statement);
    if (ret == SQLITE_ROW) {
        /* binary input entry already exists */
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE deviceBinaryInputs SET targetGroup=?, type=?, buttonId=? WHERE deviceId=? AND inputIndex=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, group) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, type) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 3, buttonId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 4, deviceId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 5, index) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    } else {
        /* insert binary input entry */
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT INTO deviceBinaryInputs(deviceId, inputIndex, targetGroup, type, buttonId) VALUES (?,?,?,?,?);", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, index) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 3, group) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 4, type) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 5, buttonId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    Datamodel_Changed();
}

/*..........................................................................*/
void Datamodel_SetDeviceBinaryInputName(uint16_t deviceId, uint8_t index, const char *name)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel SetDeviceBinaryInputName Database closed!\n");
        UNLOCK_ACCESS
        return;
    }
    sqlite3_stmt *statement;
    int ret;

    vdsm_report(LOG_DEBUG, "Datamodel set binary input name for device %d, index: %d, name: %s\n", deviceId, index, name);
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT deviceId FROM deviceBinaryInputs WHERE deviceId=? AND inputIndex=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, index) == SQLITE_OK);
    ret = sqlite3_step(statement);
    sqlite3_finalize(statement);
    if (ret == SQLITE_ROW) {
        /* binary input entry already exists */
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE deviceBinaryInputs SET name=? WHERE deviceId=? AND inputIndex=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement, 1, name, -1, SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 3, index) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    } else {
        /* insert binary input entry */
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT INTO deviceBinaryInputs(deviceId, inputIndex, name) VALUES (?,?,?);", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, index) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement, 3, name, -1, SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    Datamodel_Changed();
}

/*..........................................................................*/
uint8_t Datamodel_GetDeviceBinaryInputCount(uint16_t deviceId)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel GetDeviceBinaryInputCount Database closed!\n");
        UNLOCK_ACCESS
        return 0;
    }
    uint8_t count = 0;
    sqlite3_stmt *statement;

    vdsm_report(LOG_DEBUG, "Datamodel get binary input count of device %d\n", deviceId);
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT deviceId FROM deviceBinaryInputs WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    while (sqlite3_step(statement) == SQLITE_ROW) {
        ++count;
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return count;
}

/*..........................................................................*/
bool Datamodel_GetDeviceBinaryInput(uint16_t deviceId, uint8_t index, uint8_t *targetGroup, uint8_t *type, uint8_t *buttonId)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel GetDeviceBinaryInput Database closed!\n");
        UNLOCK_ACCESS
        return FALSE;
    }

    sqlite3_stmt *statement;
    bool success = FALSE;

    vdsm_report(LOG_DEBUG, "Datamodel get binary input of device %d, index %d\n", deviceId, index);
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT targetGroup, type, buttonId FROM deviceBinaryInputs WHERE deviceId=? AND inputIndex=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, index) == SQLITE_OK);
    if (sqlite3_step(statement) == SQLITE_ROW) {
        *targetGroup = sqlite3_column_int(statement, 0);
        *type = sqlite3_column_int(statement, 1);
        *buttonId = sqlite3_column_int(statement, 2);
        success = TRUE;
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return success;
}

/*..........................................................................*/
bool Datamodel_GetDeviceBinaryInputType(uint16_t deviceId, uint8_t index, uint8_t *type)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel GetDeviceBinaryInputType Database closed!\n");
        UNLOCK_ACCESS
        return FALSE;
    }

    sqlite3_stmt *statement;
    bool success = FALSE;

    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT type FROM deviceBinaryInputs WHERE deviceId=? AND inputIndex=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, index) == SQLITE_OK);
    if (sqlite3_step(statement) == SQLITE_ROW) {
        *type = sqlite3_column_int(statement, 0);
        success = TRUE;
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return success;
}

/*..........................................................................*/
void Datamodel_SetDeviceGroupMembershipAddGroup(uint16_t deviceId, uint8_t groupId)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel SetDeviceGroupMembershipAddGroup Database closed!\n");
        UNLOCK_ACCESS
        return;
    }

    uint8_t groupMembership[8];
    sqlite3_stmt *statement;
    int datamodel_changed_trigger = 0;

    vdsm_report(LOG_DEBUG, "Datamodel add group to groupMembership for device %d, groupId: %d\n", deviceId, groupId);

    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT groupMembership FROM devices WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    if (sqlite3_step(statement) == SQLITE_ROW) {
        groups_int_to_array(sqlite3_column_int64(statement, 0), groupMembership);
        sqlite3_finalize(statement);
        if (IsBitSet8Cleared(groupMembership, groupId)) {
            // add group
            SetBitSet8(groupMembership, groupId);
            SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE devices SET groupMembership=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
            SQL_ASSERT(sqlite3_bind_int64(statement, 1, groups_array_to_int(groupMembership)) == SQLITE_OK);
            SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
            SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
            sqlite3_finalize(statement);
	    datamodel_changed_trigger = 1;
        }
    } else {
        sqlite3_finalize(statement);
        vdsm_report(LOG_WARNING, "Datamodel add group to groupMembership: device %d not found\n", deviceId);
    }
    UNLOCK_ACCESS
    if (datamodel_changed_trigger) {
        Datamodel_Changed();
    }
}

/*..........................................................................*/
void Datamodel_ClearDeviceGroupMembership(uint16_t deviceId)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
      vdsm_report(LOG_ERR, "Datamodel ClearDeviceGroupMembership Database closed!\n");
      UNLOCK_ACCESS
      return;
    }

    /* every device is member of overall group */
    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE devices SET groupMembership=1 WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
}

/*............................................................................*/
// LOCK_ACCESS: do not use it here. function is internally used.
// Locking in under responsibility of calling function
static bool Datamodel_DeviceHasButton(Datamodel *me, uint16_t deviceId)
{
    sqlite3_stmt *statement;
    uint64_t fid = 0;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT fid FROM devices WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    fid = sqlite3_column_int64(statement, 0);
    sqlite3_finalize(statement);

    #define BUTTON_MASK 0x03
    return (fid & BUTTON_MASK);
}

/*............................................................................*/
// LOCK_ACCESS: do not use it here. function is internally used.
// Locking in under responsibility of calling function
static uint8_t Datamodel_DeviceGetPrimaryGroup(Datamodel *me, uint16_t deviceId)
{
    sqlite3_stmt *statement;
    uint64_t fid = 0;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT fid FROM devices WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    fid = sqlite3_column_int64(statement, 0);
    sqlite3_finalize(statement);

    return (uint8_t)(fid >> 12 & 0x0f);
}

/*..........................................................................*/
void Datamodel_DeriveGroupMembership(uint16_t deviceId)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel SetGroupMembership Database closed!\n");
        UNLOCK_ACCESS
        return;
    }
    // 1. check if device has buttons -> take the group from buttonId.
    if (Datamodel_DeviceHasButton(me, deviceId)) {
        uint8_t group = 0;
        uint8_t primaryGroup = Datamodel_DeviceGetPrimaryGroup(me, deviceId);
        if (JOKER_GROUP == primaryGroup) {
            sqlite3_stmt *statement;
            SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
                "SELECT buttonIdAndGroup FROM devices, deviceButtons WHERE devices.deviceId == deviceButtons.deviceId AND devices.deviceId=?;", -1, &statement, 0) == SQLITE_OK);
            SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
            SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
            group = sqlite3_column_int(statement, 0);
            group >>= 4;
            group  &= 0x0f;
            sqlite3_finalize(statement);
        } else {
            group = primaryGroup;
        }
        UNLOCK_ACCESS

        /* remove wrong assigned groups */
        Datamodel_ClearDeviceGroupMembership(deviceId);
        Datamodel_SetDeviceGroupMembershipAddGroup(deviceId, group);
        return;
    }
    UNLOCK_ACCESS

    Datamodel_ClearDeviceGroupMembership(deviceId);
    Datamodel_SetDeviceGroupMembershipAddGroup(deviceId, JOKER_GROUP);
}

/*..........................................................................*/
static uint8_t Datamodel_get_or_create_localZone_by_zoneId(Datamodel *me, uint16_t zoneId)
{
    uint8_t localZoneId;
    if (Datamodel_GetLocalZoneId(zoneId, &localZoneId)) {
        return localZoneId;
    }

    if (me->numberOfZones >= MAX_ZONES) {
        vdsm_report(LOG_ERR, "Datamodel create local zone: max zone limit reached\n");
        return 0;
    }

    localZoneId  = Datamodel_CreateZone(me, zoneId);
    return localZoneId;
}

/*..........................................................................*/
void Datamodel_ZoneGroupGetDeviceDsuids(uint8_t localZoneId, uint8_t groupId, device_dsuid_t **dsuid_list)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel ZoneGroupGetDeviceDsuids Database closed!\n");
        UNLOCK_ACCESS
        return;
    }

    sqlite3_stmt *statement;
    int ret;

    if (localZoneId == 0) {
        /* broadcast */
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT dsuid,groupMembership FROM devices WHERE mode!=0;", -1, &statement, 0) == SQLITE_OK);
    } else {
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT dsuid,groupMembership FROM devices WHERE localZoneId=? AND mode!=0;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId) == SQLITE_OK);
    }
    do {
        ret = sqlite3_step(statement);
        if (ret == SQLITE_ROW) {
            uint8_t groupMembership[8];
            groups_int_to_array(sqlite3_column_int64(statement, 1), groupMembership);
            if ((groupId == 0) || IsBitSet8Set(groupMembership, groupId)) {
                device_dsuid_t *dsuid = (device_dsuid_t*) malloc(sizeof(device_dsuid_t));
                strncpy(dsuid->dsuid, sqlite3_column_text(statement, 0), sizeof(dsuid->dsuid));
                LL_PREPEND(*dsuid_list, dsuid);
            }
        }
    } while (ret == SQLITE_ROW);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
}

/*..........................................................................*/
void Datamodel_SetOpcTableValue(uint16_t deviceId, uint8_t index, uint8_t channelId)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel SetOpcTableValue Database closed!\n");
        UNLOCK_ACCESS
        return;
    }

    sqlite3_stmt *statement;
    vdsm_report(LOG_DEBUG, "Datamodel set opc table value for device %d and index %d,  channelId: %d\n", deviceId, index, channelId);

    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT COUNT (*) FROM deviceOpcs WHERE deviceId=? AND opcIndex=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, index) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    int counts = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);

    if (counts > 0) {
        /* insert sensor entry */
        sqlite3_stmt *statementInsert;
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE deviceOpcs SET channelId = ? WHERE deviceId = ? AND opcIndex = ?;", -1, &statementInsert, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statementInsert, 1, channelId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statementInsert, 2, deviceId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statementInsert, 3, index) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statementInsert) == SQLITE_DONE);
        sqlite3_finalize(statementInsert);
    } else {
        /* insert sensor entry */
        sqlite3_stmt *statementInsert;
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT INTO deviceOpcs(deviceId, opcIndex, channelId) VALUES (?,?,?);", -1, &statementInsert, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statementInsert, 1, deviceId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statementInsert, 2, index) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statementInsert, 3, channelId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statementInsert) == SQLITE_DONE);
        sqlite3_finalize(statementInsert);
    }
    UNLOCK_ACCESS
    Datamodel_Changed();
}

/*..........................................................................*/
void Datamodel_RemoveSpareOpcTableEntries(uint16_t deviceId, uint8_t index)
{
  LOCK_ACCESS
  Datamodel *me = &l_datamodel;
  if (!me->databaseOpen) {
      vdsm_report(LOG_ERR, "Datamodel RemoveSpareOpcTableEntries Database closed!\n");
      UNLOCK_ACCESS
      return;
  }
  sqlite3_stmt *statement;
  vdsm_report(LOG_DEBUG, "Datamodel remove spare opc table entries for device %d from deviceOpcs\n", deviceId);
  SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "DELETE FROM deviceOpcs WHERE deviceId=? AND opcIndex>=?;", -1, &statement, 0) == SQLITE_OK);
  SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
  SQL_ASSERT(sqlite3_bind_int(statement, 2, index) == SQLITE_OK);
  sqlite3_step(statement);
  sqlite3_finalize(statement);
  UNLOCK_ACCESS
  Datamodel_Changed();
}

/*..........................................................................*/
uint8_t Datamodel_GetOpcChannelCount(uint16_t deviceId)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel GetOpcChannelCount Database closed!\n");
        UNLOCK_ACCESS
        return 0;
    }

    uint8_t count = 0;
    sqlite3_stmt *statement;

    vdsm_report(LOG_DEBUG, "Datamodel get Opc count of device %d\n", deviceId);
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT deviceId FROM deviceOpcs WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    while (sqlite3_step(statement) == SQLITE_ROW) {
        ++count;
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return count;
}

/*..........................................................................*/
bool Datamodel_GetOpcChannelFromIndex(uint16_t deviceId, uint8_t index, uint8_t *channelId)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel GetOpcChannelFromIndex Database closed!\n");
        UNLOCK_ACCESS
        return FALSE;
    }
    sqlite3_stmt *statement;
    bool success = FALSE;

    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT channelId FROM deviceOpcs WHERE deviceId=? AND opcIndex=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, index) == SQLITE_OK);
    if (sqlite3_step(statement) == SQLITE_ROW) {
        *channelId = sqlite3_column_int(statement, 0);
        success = TRUE;
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    if (success) {
        vdsm_report(LOG_DEBUG, "Datamodel get opc value of device %d, index %d channel:%d\n", deviceId, index, *channelId);
    } else {
        vdsm_report(LOG_DEBUG, "Datamodel get opc value of device not found %d, index %d\n", deviceId, index);
    }
    return success;
}

/*..........................................................................*/
void Datamodel_SetDefaultChannel(uint16_t deviceId)
{
    // Default channel: Table is used for debugging support
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel SetDefaultChannel Database closed!\n");
        UNLOCK_ACCESS
        return;
    }
    sqlite3_stmt *statement;
    int datamodel_changed_trigger = 0;
    vdsm_report(LOG_DEBUG, "Datamodel set default channel available for device %d\n", deviceId);

    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT COUNT (*) FROM deviceDefaultChannel WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);

    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    int counts = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);

    if (counts == 0) {
        /* insert sensor entry */
        sqlite3_stmt *statementInsert;
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT INTO deviceDefaultChannel(deviceId) VALUES (?);", -1, &statementInsert, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statementInsert, 1, deviceId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statementInsert) == SQLITE_DONE);
        sqlite3_finalize(statementInsert);
        datamodel_changed_trigger = 1;
    }
    UNLOCK_ACCESS
    if (datamodel_changed_trigger) {
        Datamodel_Changed();
    }
}

/*..........................................................................*/
void Datamodel_RemoveDefaultChannel(uint16_t deviceId)
{
    // Default channel: Table is used for debugging support
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel RemoveDefaultChannel Database closed!\n");
        UNLOCK_ACCESS
        return;
    }
    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT COUNT (*) FROM deviceDefaultChannel WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, deviceId) == SQLITE_OK);

    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    int counts = sqlite3_column_int(statement, 0);
    int datamodel_changed_trigger = 0;
    sqlite3_finalize(statement);

    if (counts != 0) {
        /* insert sensor entry */
        sqlite3_stmt *statementRemove;
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "DELETE FROM deviceDefaultChannel WHERE deviceId=?;", -1, &statementRemove, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statementRemove, 1, deviceId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statementRemove) == SQLITE_DONE);
        sqlite3_finalize(statementRemove);
        datamodel_changed_trigger = 1;
    }
    UNLOCK_ACCESS
    if (datamodel_changed_trigger) {
        Datamodel_Changed();
    }
}

/*..........................................................................*/
bool Datamodel_GetVdcDsuidFromDeviceDsuid(dsuid_t *deviceDsuid, dsuid_t *vdcDsuid)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel GetVdcdsuidFromDeviceDsuid Database closed!\n");
        UNLOCK_ACCESS
        return FALSE;
    }

    if (vdcDsuid == NULL) {
        vdsm_report(LOG_ERR, "return container is null\n");
        UNLOCK_ACCESS
        return FALSE;
    }

    char device_dsuid_str[DSUID_STR_LEN];
    dsuid_to_string_upper(deviceDsuid, device_dsuid_str);

    sqlite3_stmt *statement;
    bool success = FALSE;

    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT vdcdsuid FROM devices WHERE dsuid=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 1, device_dsuid_str, -1, SQLITE_STATIC) == SQLITE_OK);

    if (sqlite3_step(statement) == SQLITE_ROW) {
        const char *vdc_dsuid_str = sqlite3_column_text(statement, 0);
        if (vdc_dsuid_str) {
            success =(dsuid_from_string(vdc_dsuid_str, vdcDsuid) == 0);
        } else {
            vdsm_report(LOG_WARNING, "Datamodel device with dsuid:%s has no vdcdsuid\n", device_dsuid_str);
        }
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return success;
}

uint8_t Datamodel_GetBusMemberType(const dsuid_t *vdcDsuid)
{
    Datamodel *me = &l_datamodel;
    if (dsuid_equal(vdcDsuid, &me->dsmid) ) {
        return BusMember_vDSM;
    }
    return BusMember_vDC;
}

/*............................................................................*/
// LOCK_ACCESS: do not use it here. function is internally used.
// Locking in under responsibility of calling function
static int Datamodel_GetZoneDeviceSensorCount(Datamodel *me, uint8_t localZoneId, uint8_t sensorType, bool checkSensorType)
{
    sqlite3_stmt *statement;
    int counts = 0;
    if (checkSensorType) {
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT COUNT (*) FROM zoneSensors WHERE localZoneId=? AND type=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, sensorType) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
        counts = sqlite3_column_int(statement, 0);
        sqlite3_finalize(statement);
    } else {
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT COUNT (*) FROM zoneSensors WHERE localZoneId=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
        counts = sqlite3_column_int(statement, 0);
        sqlite3_finalize(statement);
    }
    return counts;
}

/*............................................................................*/
// LOCK_ACCESS: do not use it here. function is internally used.
// Locking in under responsibility of calling function
static bool Datamodel_ResetZoneDeviceSensor(Datamodel *me, uint8_t localZoneId, uint8_t sensorType)
{
    // Reset sensor device of type for zone
    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
               "DELETE FROM zoneSensors WHERE localZoneId=? AND  type=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, sensorType) == SQLITE_OK);
    int ret = sqlite3_step(statement);
    sqlite3_finalize(statement);
    return (ret == SQLITE_DONE);
}

/*............................................................................*/
// LOCK_ACCESS: do not use it here. function is internally used.
// Locking in under responsibility of calling function
static bool Datamodel_ZoneDeviceSensorExists(Datamodel *me, uint8_t localZoneId, dsuid_t *sensorDsuid, uint8_t sensorType)
{
    char dsuid_str[DSUID_STR_LEN];
    dsuid_to_string_upper(sensorDsuid, dsuid_str);

    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
               "SELECT COUNT (*) FROM zoneSensors WHERE localZoneId = ? AND type=? AND dsuid = ?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, sensorType) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 3, dsuid_str, -1, SQLITE_STATIC) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    int deviceCount = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return (deviceCount>0);
}

/*............................................................................*/
// LOCK_ACCESS: do not use it here. function is internally used.
// Locking in under responsibility of calling function
static bool Datamodel_AddZoneDeviceSensor(Datamodel *me, uint8_t localZoneId, dsuid_t *sensorDsuid, uint8_t sensorType)
{
    char dsuid_str[DSUID_STR_LEN];
    dsuid_to_string_upper(sensorDsuid, dsuid_str);

    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
               "INSERT INTO zoneSensors (localZoneId, type, dsuid) VALUES(?,?,?);", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, sensorType) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 3, dsuid_str, -1, SQLITE_STATIC) == SQLITE_OK);
    int ret = sqlite3_step(statement);
    sqlite3_finalize(statement);
    return (ret == SQLITE_DONE);
}

/*..........................................................................*/
uint8_t Datamodel_SetSensorPushDevice(uint8_t localZoneId, uint8_t sensorType, dsuid_t *sensorDsuid)
{
    if (sensorDsuid == NULL) {
        vdsm_report(LOG_ERR,"dsuid of sensor is null\n");
        return ERROR_WRONG_PARAMETER;
    }

    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel SetSensorPushDevice Database closed!\n");
        UNLOCK_ACCESS
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }

    // check if the according entry exists in the database
    if (Datamodel_ZoneDeviceSensorExists(me, localZoneId, sensorDsuid, sensorType)) {
        UNLOCK_ACCESS
        return ERROR_OK;
    }

    uint8_t retVal = ERROR_OK;

    // a type for the zone exists -> replace it
    int typeZoneSensorCount = Datamodel_GetZoneDeviceSensorCount(me, localZoneId, sensorType, true);
    if (typeZoneSensorCount > 0) {
        if (Datamodel_ResetZoneDeviceSensor(me, localZoneId, sensorType) &&
            Datamodel_AddZoneDeviceSensor(me, localZoneId, sensorDsuid, sensorType)) {
            retVal = ERROR_OK;
        } else {
            retVal = ERROR_DEVICE_NOT_FOUND;
        }
    } else {
        // Get total push sensors for zone
        int allZoneSensorCount = Datamodel_GetZoneDeviceSensorCount(me, localZoneId, 0, false);
        if (allZoneSensorCount < MAX_PUSH_SENSORS_PER_ZONE) {

            if (Datamodel_AddZoneDeviceSensor(me, localZoneId, sensorDsuid, sensorType)){
                retVal = ERROR_OK;
            } else {
                retVal = ERROR_DEVICE_NOT_FOUND;
            }
        } else {
            retVal = ERROR_OUT_OF_RESOURCES;
        }
    }
    UNLOCK_ACCESS
    return retVal;
}

/*..........................................................................*/
uint8_t Datamodel_GetSensorPushDevice(uint8_t localZoneId, uint8_t sensorType, dsuid_t *sensorDsuid)
{
    if (sensorDsuid == NULL) {
        vdsm_report(LOG_ERR,"GetSensorPushDevice dsuid of sensor is null\n");
        return ERROR_WRONG_PARAMETER;
    }

    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel GetSensorPushDevice Database closed!\n");
        UNLOCK_ACCESS
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }

    // Reset sensor device of type for zone
    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
               "SELECT dsuid FROM zoneSensors WHERE \
                localZoneId=? AND type=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, sensorType) == SQLITE_OK);
    int ret = sqlite3_step(statement);

    if (ret != SQLITE_ROW) {
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        vdsm_report(LOG_DEBUG, "GetSensorPushDevice no sensor Device in DB for localZone: %d, with SensorType: %d\n", localZoneId, sensorType);
        return ERROR_DEVICE_NOT_FOUND;
    }

    const char *dsuid_str = sqlite3_column_text(statement, 0);
    bool success = false;
    if (dsuid_str) {
        success = (dsuid_from_string(dsuid_str, sensorDsuid) == 0);
    } else {
        vdsm_report(LOG_WARNING, "GetSensorPushDevice no sensor Device for localZone: %d, with SensorType: %d\n", localZoneId, sensorType);
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    if (success) {
        return ERROR_OK;
    }

    vdsm_report(LOG_ERR, "Datamodel GetSensorPushDevice could not convert string %s to dsuid\n", dsuid_str);
    return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
}

/*..........................................................................*/
uint8_t Datamodel_ResetSensorPushDevice(uint8_t localZoneId, uint8_t sensorType)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel ResetSensorPushDevice Database closed!\n");
        UNLOCK_ACCESS
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }

    int counts = Datamodel_GetZoneDeviceSensorCount(me, localZoneId, sensorType, true);
    if (counts == 0) {
        UNLOCK_ACCESS
        return ERROR_OK;
    }

    Datamodel_ResetZoneDeviceSensor(me, localZoneId, sensorType);
    UNLOCK_ACCESS
    return ERROR_OK;
}

/*..........................................................................*/
// internal use only
static bool Datamodel_ZoneExists(Datamodel *me, uint16_t zoneId)
{
    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
               "SELECT COUNT (*) from zones where zoneId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, zoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    int zoneCount = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);
    return (zoneCount > 0);
}

/*..........................................................................*/
// internal use only
static void Datamodel_recoverHeatingTable(const char *dsuid_str, uint16_t zoneId)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel recoverHeatingTable Database closed!\n");
        UNLOCK_ACCESS
        return;
    }

    // check if zone exists on vdc
    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT COUNT (*) FROM devices, zones WHERE devices.vdcdsuid=? AND \
                                  devices.localZoneId == zones.localZoneId AND zones.zoneId=?;",
                                  -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 1, dsuid_str, -1, SQLITE_STATIC) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, zoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    int count = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);

    // zone exists on vdc
    if (count > 0) {
        // recover entries with vdcdsuid null
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
                 "UPDATE heating SET vdcdsuid=? \
                 WHERE zoneId=? AND vdcdsuid is null;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement, 1, dsuid_str, -1, SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, zoneId) == SQLITE_OK);
        sqlite3_step(statement);
        sqlite3_finalize(statement);
    }
    UNLOCK_ACCESS
}

/*..........................................................................*/
// internal use only
static int Datamodel_vdcHasZoneHeatingControl(const dsuid_t *vdcDsuid, uint16_t zoneId)
{
    char dsuid_str[DSUID_STR_LEN];
    dsuid_to_string_upper(vdcDsuid, dsuid_str);

    Datamodel_recoverHeatingTable(dsuid_str, zoneId);

    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel vdcHasZoneHeatingControl Database closed!\n");
        UNLOCK_ACCESS
        return 0;
    }

    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
               "SELECT COUNT (*) FROM heating WHERE \
                zoneId=? AND vdcdsuid=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, zoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 2, dsuid_str, -1, SQLITE_STATIC) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    int configCount = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    return configCount;
}

/*..........................................................................*/
bool Datamodel_GetDsuidFromHeatingZone(uint16_t zoneId, dsuid_t* dsuid)
{
    if (dsuid == NULL) {
        return false;
    }

    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel GetDsuidFromHeatingZone Database closed!\n");
        UNLOCK_ACCESS
        return false;
    }

    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
               "SELECT vdcdsuid FROM heating WHERE \
               zoneId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, zoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    const char *name = sqlite3_column_text(statement, 0);
    bool success = false;
    if (name != NULL) {
        success = (dsuid_from_string(name, dsuid) == 0);
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    return success;
}

/*..........................................................................*/
uint8_t Datamodel_setHeatingConfigOfvDC(const dsuid_t *vdcDsuid, uint16_t zoneId, HeatingConfig *heating)
{
    if (!heating) {
        vdsm_report(LOG_ERR, "Datamodel setHeatingConfig heating is NULL!\n");
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }

    // check if vdc is owner
    int count = Datamodel_vdcHasZoneHeatingControl(vdcDsuid, zoneId);
    if (count == 1) {
        return Datamodel_setHeatingConfig(vdcDsuid, zoneId, heating);
    }

    // not own vdc. Check if modification allowed.
    HeatingConfig tempHeating;
    uint8_t retVal = Datamodel_getHeatingConfig(zoneId, &tempHeating);
    if (retVal == ERROR_OK) {
        if (Heating_IsControllerModeOff(&tempHeating)) {
            return Datamodel_setHeatingConfig(vdcDsuid, zoneId, heating);
        }
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }
    return retVal;
}

/*..........................................................................*/
uint8_t Datamodel_getHeatingConfigOfvDC(const dsuid_t *vdcDsuid, uint16_t zoneId, HeatingConfig *heating)
{
    if (!heating) {
        vdsm_report(LOG_ERR, "Datamodel getHeatingConfigOfvDC heating is NULL!\n");
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }

    int configCount = Datamodel_vdcHasZoneHeatingControl(vdcDsuid, zoneId);
    if (configCount == 1) {
        return Datamodel_getHeatingConfig(zoneId, heating);
    }

    if (Heating_initializeConfig(heating)) {
        return ERROR_OK;
    }

    return ERROR_NO_CONFIG;
}

/*..........................................................................*/
uint8_t Datamodel_getHeatingOperationModesOfvDC(const dsuid_t *vdcDsuid, uint16_t zoneId, uint16_t *operationMode, uint16_t length)
{
    if (!operationMode) {
        vdsm_report(LOG_ERR, "Datamodel getOperationModesOfvDC operationMode is NULL!\n");
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }

    int configCount = Datamodel_vdcHasZoneHeatingControl(vdcDsuid, zoneId);
    if (configCount == 1) {
        return Datamodel_getHeatingOperationModes(zoneId, operationMode, length);
    }

    if (Heating_initializeOperationModes(operationMode, length)) {
        return ERROR_OK;
    }

    return ERROR_NO_CONFIG;
}

/*..........................................................................*/
uint8_t Datamodel_setHeatingConfig(dsuid_t const *vdcDsuid, uint16_t zoneId, HeatingConfig *heating)
{
    if (!heating || !vdcDsuid) {
        vdsm_report(LOG_ERR, "Datamodel setHeatingConfig heating is NULL!\n");
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }

    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel setHeatingConfig Database closed!\n");
        UNLOCK_ACCESS
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }

    if (!Datamodel_ZoneExists(me, zoneId)) {
        UNLOCK_ACCESS
        return ERROR_ZONE_NOT_FOUND;
    }

    // check if configuration valid
    if (!Heating_IsConfigurationValid(heating, zoneId)) {
        UNLOCK_ACCESS
        return ERROR_WRONG_PARAMETER;
    }

    sqlite3_stmt *statement;
    // check if zone already available
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
               "SELECT COUNT (*) FROM heating WHERE \
                zoneId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, zoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    int itemsFound = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);

    char dsuid_str[DSUID_STR_LEN];
    dsuid_to_string_upper(vdcDsuid, dsuid_str);

    if (itemsFound > 0) {
        /* update */
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
                   "UPDATE heating SET \
                    Kp=?, Ti=?, \
                    Ts=?,  Kd=?, \
                    imin=?, imax=?, \
                    ymin=?, ymax=?, \
                    offset=?, sourceZoneId=?, \
                    emergencyValue=?, antiWindUp=?, \
                    controllerMode=?, keepFloorWarm=?, manualValue=?, vdcdsuid=?  WHERE zoneId=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, heating->Kp) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, heating->Ti) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 3, heating->Ts) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 4, heating->Kd) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 5, heating->imin) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 6, heating->imax) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 7, heating->ymin) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 8, heating->ymax) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 9, heating->offset) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 10, heating->sourceZoneId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 11, heating->emergencyValue) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 12, heating->antiWindUp) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 13, heating->controllerMode) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 14, heating->keepFloorWarm) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 15, heating->manualValue) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 15, heating->manualValue) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement,16, dsuid_str, -1, SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 17, zoneId) == SQLITE_OK);
    } else {
        /* insert */
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
                   "INSERT INTO heating \
                    (zoneId, Kp, Ti, Ts, \
                    Kd, imin, imax, ymin, \
                    ymax, offset, sourceZoneId, emergencyValue, \
                    antiWindUp, controllerMode, keepFloorWarm, manualValue,vdcdsuid) \
                    VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, zoneId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, heating->Kp) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 3, heating->Ti) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 4, heating->Ts) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 5, heating->Kd) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 6, heating->imin) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 7, heating->imax) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 8, heating->ymin) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 9, heating->ymax) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 10, heating->offset) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 11, heating->sourceZoneId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 12, heating->emergencyValue) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 13, heating->antiWindUp) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 14, heating->controllerMode) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 15, heating->keepFloorWarm) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 16, heating->manualValue) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement,17, dsuid_str, -1, SQLITE_STATIC) == SQLITE_OK);

    }
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);

    // get localZoneId
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT localZoneId FROM zones WHERE zoneId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, zoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    uint8_t localZoneId = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    /* notify heating FSM about updated configuration */
    Heating_NotifyConfigChange(localZoneId);
    return ERROR_OK;
}

/*..........................................................................*/
uint8_t Datamodel_getHeatingConfig(uint16_t zoneId, HeatingConfig *heating)
{
    if (!heating) {
        vdsm_report(LOG_ERR, "Datamodel getHeatingConfig heating is NULL!\n");
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }

    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel getHeatingConfig Database closed!\n");
        UNLOCK_ACCESS
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }

    if (!Datamodel_ZoneExists(me, zoneId)) {
        UNLOCK_ACCESS
        return ERROR_ZONE_NOT_FOUND;
    }

    sqlite3_stmt *statement;
    /* check if zone already available */
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
               "SELECT * FROM heating WHERE \
                zoneId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, zoneId) == SQLITE_OK);
    if (sqlite3_step(statement) != SQLITE_ROW) {
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        memset(heating, 0, sizeof(HeatingConfig));
        return ERROR_OK;
    }

    heating->Kp = sqlite3_column_int(statement, 1);
    heating->Kd = sqlite3_column_int(statement, 2);
    heating->Ti = sqlite3_column_int(statement, 3);
    heating->imin = sqlite3_column_int(statement, 4);
    heating->imax = sqlite3_column_int(statement, 5);
    heating->ymin = sqlite3_column_int(statement, 6);
    heating->ymax = sqlite3_column_int(statement, 7);
    heating->Ts = sqlite3_column_int(statement, 8);
    heating->antiWindUp = sqlite3_column_int(statement, 9);
    heating->sourceZoneId = sqlite3_column_int(statement, 10);
    heating->offset = sqlite3_column_int(statement, 11);
    heating->emergencyValue = sqlite3_column_int(statement, 12);
    heating->manualValue = sqlite3_column_int(statement, 13);
    heating->keepFloorWarm = sqlite3_column_int(statement, 14);
    heating->controllerMode = sqlite3_column_int(statement, 15);

    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return ERROR_OK;
}

uint8_t Datamodel_setHeatingOperationModes(uint16_t zoneId, uint16_t *operationMode, uint16_t length)
{
    if (!operationMode) {
        vdsm_report(LOG_ERR, "Datamodel setOperationModes operationMode is NULL!\n");
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }

    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel setOperationModes Database closed!\n");
        UNLOCK_ACCESS
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }

    if (!Datamodel_ZoneExists(me, zoneId)) {
      UNLOCK_ACCESS
      return ERROR_ZONE_NOT_FOUND;
    }

    sqlite3_stmt *statement;

    // check if zone already available
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
               "SELECT COUNT (*) FROM heatingOperationMode WHERE \
                zoneId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, zoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    int itemsFound = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);

    if (itemsFound > 0) {
        /* update */
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
                   "UPDATE heatingOperationMode SET operationMode=? WHERE zoneId=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_blob(statement, 1, operationMode, length*sizeof(uint16_t), SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, zoneId) == SQLITE_OK);
    } else {
        /* insert */
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
                   "INSERT INTO heatingOperationMode \
                    (zoneId, operationMode) VALUES(?, ?);", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, zoneId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_blob(statement, 2, operationMode, length*sizeof(uint16_t), SQLITE_STATIC) == SQLITE_OK);
    }

    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);

    // get localZoneId
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT localZoneId FROM zones WHERE zoneId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, zoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    uint8_t localZoneId = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    /* notify heating FSM about updated operation modes */
    Heating_NotifyOperationModesChange(localZoneId);
    return ERROR_OK;
}

uint8_t Datamodel_getHeatingOperationModes(uint16_t zoneId, uint16_t *operationMode, uint16_t length)
{
    if (!operationMode) {
        vdsm_report(LOG_ERR, "Datamodel getOperationModes operationMode is NULL!\n");
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }

    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel getOperationModes Database closed!\n");
        UNLOCK_ACCESS
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }

    sqlite3_stmt *statement;
    /* check if zone available */

    if (!Datamodel_ZoneExists(me, zoneId)) {
        UNLOCK_ACCESS
        return ERROR_ZONE_NOT_FOUND;
    }

    SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
               "SELECT operationMode FROM heatingOperationMode WHERE \
                zoneId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, zoneId) == SQLITE_OK);
    if (sqlite3_step(statement) != SQLITE_ROW) {
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        memset(operationMode, 0, length * sizeof(uint16_t));
        return ERROR_OK;
    }

    uint16_t blobSize;
    blobSize = sqlite3_column_bytes(statement, 0);
    if (blobSize != length*sizeof(uint16_t)) {
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        return ERROR_WRONG_SIZE;
    }

    memcpy(operationMode, sqlite3_column_blob(statement, 0), blobSize);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return ERROR_OK;
}

uint8_t Datamodel_updateVdcName(const char *vdc_dsuid, const char *name)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel updateVdcName Database closed!\n");
        UNLOCK_ACCESS
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }

    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE dsm_info SET name=? WHERE dsuid=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 1, (const char *)name, -1, SQLITE_STATIC) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 2, (const char *)vdc_dsuid, -1, SQLITE_STATIC) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return ERROR_OK;
}

uint8_t Datamodel_updateDeviceZone(uint16_t* zoneId, uint16_t deviceId, const char* vdcdsuid_str)
{
    uint8_t localZoneId = 0;

    // get default vdc zone id, if not defined
    if (*zoneId == 0) {
        *zoneId = Datamodel_getVdcDefaultZoneID(vdcdsuid_str);
    }
    if (*zoneId != 0) {
        Datamodel *me = &l_datamodel;
        localZoneId  = Datamodel_get_or_create_localZone_by_zoneId(me, *zoneId);
    } else {
        // Assign local zone id
        dsuid_t vdcdsuid;
        dsuid_from_string(vdcdsuid_str, &vdcdsuid);
        Datamodel *me = &l_datamodel;
        DatamodelZoneInfo* zoneInfo;
        zoneInfo = Datamodel_CreateGetZoneInfoByDsuid(me, &vdcdsuid);
        if (zoneInfo) {
            localZoneId = zoneInfo->super.localZoneId;
        } else {
          vdsm_report(LOG_ERR, "Datamodel updateDeviceZone zone Info not found.\n");
          return ERROR_ZONE_NOT_FOUND;
        }
    }

    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel updateDeviceZone Database closed!\n");
        UNLOCK_ACCESS
        return ERROR_OUT_OF_RESOURCES;
    }

    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb,
        "UPDATE devices SET localZoneId=? WHERE deviceId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, localZoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 2, deviceId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS

    return ERROR_OK;
}

/*............................................................................*/
// LOCK_ACCESS: do not use it here. function is internally used.
// Locking in under responsibility of calling function
static void Datamodel_setClusterSceneLock(Datamodel *me, uint8_t clusterId, uint64_t data)
{
    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT OR REPLACE INTO sceneLockCluster (clusterId, value) VALUES(?,?);", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, clusterId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int64(statement, 2, data) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
    sqlite3_finalize(statement);
}

/*............................................................................*/
// LOCK_ACCESS: do not use it here. function is internally used.
// Locking in under responsibility of calling function
static bool Datamodel_getClusterSceneLock(uint8_t clusterId, uint64_t *data)
{
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel getClusterSceneLock Database closed!\n");
        return false;
    }
    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT value FROM sceneLockCluster WHERE clusterId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, clusterId) == SQLITE_OK);
    int ret = sqlite3_step(statement);
    if (ret != SQLITE_ROW) {
        sqlite3_finalize(statement);
        return false;
    }
    *data = sqlite3_column_int64(statement, 0);
    sqlite3_finalize(statement);
    return true;
}

/*............................................................................*/
// LOCK_ACCESS: do not use it here. function is internally used.
// Locking in under responsibility of calling function
static bool Datamodel_clusterExists(Datamodel *me, uint8_t clusterId)
{
    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT COUNT (*) FROM cluster \
               WHERE clusterId=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, clusterId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_step(statement) == SQLITE_ROW);
    uint8_t count = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);
    return (count > 0);
}

/*............................................................................*/
uint8_t Datamodel_getCluster(uint8_t clusterId, DatamodelCluster* cluster)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel getCluster Database closed!\n");
        UNLOCK_ACCESS
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }

    Datamodel_InitCluster(cluster, clusterId);

    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT clusterId, stateMachineGroupId, stopTimeout, \
                                remoteStopTimeout, name, location, floor, protectionClass, \
                                configurationLock from cluster WHERE clusterId = ?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement, 1, cluster->id) == SQLITE_OK);
    int ret = sqlite3_step(statement);
    if (ret != SQLITE_ROW)
    {
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        return ERROR_GROUP_NOT_FOUND;
    }

    cluster->stateMachineGroupId = sqlite3_column_int(statement, 1);
    cluster->stopTimeout = sqlite3_column_int(statement, 2);
    cluster->remoteStopTimeout = sqlite3_column_int(statement, 3);
    const char *text = sqlite3_column_text(statement, 4);
    if (text != NULL) {
        memcpy(cluster->name, text, MAX_NAME_LEN);
    }
    cluster->location = sqlite3_column_int(statement, 5);
    cluster->floor = sqlite3_column_int(statement, 6);
    cluster->protectionClass = sqlite3_column_int(statement, 7);
    cluster->configurationLock = sqlite3_column_int(statement, 8);

    uint64_t data = 0;
    if (Datamodel_getClusterSceneLock(cluster->id, &data)) {
        int i = 0;
        for (i = 0; i < 16; ++i) {
            cluster->sceneLock[i] = (data &0x01);
            data >>= 1;
            // TODO(someday): move `<<=` before `|=` to make this correct inverse of serialization code?
        }
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return ERROR_OK;
}

void Datamodel_setVdcDefaultZoneID(const char *vdcDsuid, uint16_t defaultZoneId)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel setVdcDefaultZoneID Database closed!\n");
        UNLOCK_ACCESS
    }
    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE dsm_info SET defaultZoneId=? WHERE dsuid=?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_int(statement,  1, defaultZoneId) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 2, (const char *)vdcDsuid, -1, SQLITE_STATIC) == SQLITE_OK);

    if (sqlite3_step(statement) != SQLITE_DONE) {
        vdsm_report(LOG_ERR, "Datamodel setVdcDefaultZoneID Database failed!\n");
    }
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
}

uint16_t Datamodel_getVdcDefaultZoneID(const char *vdcDsuid)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel getVdcDefaultZoneID Database closed!\n");
        UNLOCK_ACCESS
        return 0;
    }
    sqlite3_stmt *statement;
    SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "SELECT defaultZoneId from dsm_info WHERE dsuid = ?;", -1, &statement, 0) == SQLITE_OK);
    SQL_ASSERT(sqlite3_bind_text(statement, 1, (const char *)vdcDsuid, -1, SQLITE_STATIC) == SQLITE_OK);
    int ret = sqlite3_step(statement);
    if (ret != SQLITE_ROW) {
        sqlite3_finalize(statement);
        UNLOCK_ACCESS
        return 0; // error not found
    }
    uint16_t defaultZoneId = sqlite3_column_int(statement, 0);
    sqlite3_finalize(statement);
    UNLOCK_ACCESS
    return defaultZoneId;
}

static uint8_t Datamodel_setClusterName(uint8_t groupId, const uint8_t* name)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel setClusterName Database closed!\n");
        UNLOCK_ACCESS
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }
    sqlite3_stmt *statement;
    if (Datamodel_clusterExists(me, groupId)) {
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE cluster SET  name=? WHERE clusterId=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement, 1, (const char *)name, -1, SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, groupId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
    } else {
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT INTO cluster (clusterId,  name) VALUES(?,?);", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, groupId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_text(statement, 2, (const char *)name, -1, SQLITE_STATIC) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
    }
    UNLOCK_ACCESS
    return ERROR_OK;
}

static uint8_t Datamodel_setClusterStopTimeout(uint8_t groupId, uint8_t stopTimeout)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel setClusterName Database closed!\n");
        UNLOCK_ACCESS
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }
    sqlite3_stmt *statement;
    if (Datamodel_clusterExists(me, groupId)) {
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE cluster SET stopTimeout=? WHERE clusterId=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, stopTimeout) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, groupId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
    } else {
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT INTO cluster (clusterId,  stopTimeout) VALUES(?,?);", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, groupId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, stopTimeout) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
    }
    UNLOCK_ACCESS
    return ERROR_OK;
}

static uint8_t Datamodel_setClusterRemoteStopTimeout(uint8_t groupId, uint8_t remoteStopTimeout, uint8_t color)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel setClusterName Database closed!\n");
        UNLOCK_ACCESS
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }
    sqlite3_stmt *statement;
    if (Datamodel_clusterExists(me, groupId)) {
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE cluster SET remoteStopTimeout=?, stateMachineGroupId=? WHERE clusterId=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, remoteStopTimeout) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, color) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 3, groupId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
    } else {
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT INTO cluster (clusterId,  remoteStopTimeout, stateMachineGroupId) VALUES(?,?,?);", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, groupId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, remoteStopTimeout) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 3, color) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
    }
    UNLOCK_ACCESS
    return ERROR_OK;
}

static uint8_t Datamodel_setLocationProtection(uint8_t groupId, uint16_t location, uint16_t floor, uint16_t protectionClass)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel setClusterName Database closed!\n");
        UNLOCK_ACCESS
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }
    sqlite3_stmt *statement;
    if (Datamodel_clusterExists(me, groupId)) {
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE cluster SET location=?, floor=?, protectionClass=? WHERE clusterId=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, location) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, floor) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 3, protectionClass) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 4, groupId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
    } else {
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT INTO cluster (clusterId, location, floor, protectionClass) VALUES(?,?,?,?);", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, groupId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, location) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 3, floor) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 4, protectionClass) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
    }
    UNLOCK_ACCESS
    return ERROR_OK;
}

static uint8_t Datamodel_setConfigurationLock(uint8_t groupId, bool lock)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel setClusterName Database closed!\n");
        UNLOCK_ACCESS
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }
    sqlite3_stmt *statement;
    if (Datamodel_clusterExists(me, groupId)) {
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "UPDATE cluster SET configurationLock=? WHERE clusterId=?;", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, lock) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, groupId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
    } else {
        SQL_ASSERT(sqlite3_prepare_v2(me->pDb, "INSERT INTO cluster (clusterId,  configurationLock) VALUES(?,?);", -1, &statement, 0) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 1, groupId) == SQLITE_OK);
        SQL_ASSERT(sqlite3_bind_int(statement, 2, lock) == SQLITE_OK);
        SQL_ASSERT(sqlite3_step(statement) == SQLITE_DONE);
        sqlite3_finalize(statement);
    }
    UNLOCK_ACCESS
    return ERROR_OK;
}

uint64_t Datamodel_sceneLock_array_to_int(const uint8_t *sceneLock, uint8_t sceneLockSize)
{
    uint64_t data = 0;
    int i = 0;
    for (i = sceneLockSize; i > 0; --i) {
        data |= (0x01 & sceneLock[i]);
        data <<= 1;
    }
    return data;
}

static uint8_t Datamodel_setSceneLock(uint8_t groupId, const uint8_t *sceneLock, uint8_t sceneLockSize)
{
    LOCK_ACCESS
    Datamodel *me = &l_datamodel;
    if (!me->databaseOpen) {
        vdsm_report(LOG_ERR, "Datamodel setClusterName Database closed!\n");
        UNLOCK_ACCESS
        return ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
    }

    uint64_t data = Datamodel_sceneLock_array_to_int(sceneLock, sceneLockSize);
    Datamodel_setClusterSceneLock(me, groupId, data);

    UNLOCK_ACCESS
    return ERROR_OK;
}
